#####################################################################################
#
# Module: m25ai
# Author: m25
# 
# Description: custom artificial intelligence for the game Freedom Force
#
#
# 
#
#####################################################################################

# please update as needed
M25AI_VERSION ={
'majorVersion': 3, # official major version, ex.: 3 for 3.0.4
'minorVersion': 3, # official update version, ex.: 0 for 3.0.4
'patchVersion': 0, # official patch version, ex.: 4 for 3.0.4
'buildVersion': 3, # for each new main branch, new build number; for someone's branch, name of the main build from which this was branched
'buildBranch': "Gold v.3.3 with Patch" # either 'main' (followed by some identification for public releases, such as 'beta 1' or 'final') for merged code or name of the author of the branch, optionally followed by the branch's numbering
}
print "Loading m25ai.py v. %i.%i.%i build %i; branch = %s" %  (M25AI_VERSION['majorVersion'], M25AI_VERSION['minorVersion'], M25AI_VERSION['patchVersion'], M25AI_VERSION['buildVersion'], M25AI_VERSION['buildBranch'])

#June 25, 2008
#m25: removed obselete testing functions
#m25: fixed bug in SetupAI()
#m25: changed inactivity timer in AISupport_DoPower to correct odd behaviour

#June 6, 2008
#m25: CLONE CHECK TIMER set to 1 to grab clones more quickly
#m25: AI_AUTOTIMER now uses a constant from m25config (m25_AI_DELAY)
#m25: removed auto-calc of AI_SPEED_DELAY based on timing a dynamic function (didn't provide much benefit other than overhead)
# ------------   use AI_GetDelay(char), and AI_SetDelay(char) to set the AI delay for individual characters dynamically (only works if AI_SPEED_DELAY is turned off)


#May 14, 2008
#m25: Added GetAIDelay() to read a parameter called 'delay' from the AI definition
#m25: updated GetAutoAIDelay() and AIRestart() to use either the value of GetAIDelay() or a value calculated from the character's speed

# Update the version string in version_update and uncomment this
# to update the version string on all of the m25 files
#import version_update
#version_update.Do_VersionUpdate()


import ff
import js

import ai
import cshelper
import math
import types
import time
from time import *
from random import randint

import string

import m25event
import m25timestamp
import m25distance
import m25aidata
import m25template
import m25string
import m25report
import m25missionover
import m25config
import m25energy

import missionobjvar
import datfiles
import chardata

import skXAngles

import m25generateai

from m25distance import *

from m25report import *
from m25getset import *
from m25string import *
from m25missionover import *
from m25energy import *
from m25timestamp import *

from m25config import *
    
import m25team
from m25team import *

from mlogreader import *

try:
	import ffx
except:
	print 'AI cannot import FFX!'

import m25custompowers
from m25custompowers import *
import ffxcustom
import ffxcustom2
from ffxcustom2 import *


import m25cutscene
from m25cutscene import *

import skXMapInfo
from skXMapInfo import *

if M25_USE_SKXCAMERA:
	import skXCamera2
	from skXCamera2 import *

import m25aiopt




#-------------------------------------------------------------------------------------
#
# For testing purposes
#
#-------------------------------------------------------------------------------------

debugpowerlist = ['none']
AI_DEBUG_CHANNEL = 'AIDEBUG'
m25report.Report_SetChannel(AI_DEBUG_CHANNEL,1)

ALLOW_ERRORS = m25config.M25_USE_REPORT

AI_CHANNEL_GOAL = 'GOAL'
GOAL_CHANNEL = 'GOAL'
AI_CHANNEL = 'AI'
BRAWL_CHANNEL = 'BRAWL'
ERROR_CHANNEL = 'ERROR'
WARNING_CHANNEL = 'WARNING'
GROUP_CHANNEL = 'GROUP'
TIME_CHANNEL = 'TIME'

TARGET_CHANNEL = 'TARGET'
m25report.Report_SetChannel(GROUP_CHANNEL,1)
m25report.Report_SetChannel(ERROR_CHANNEL,2)
m25report.Report_SetChannel(WARNING_CHANNEL,0)
m25report.Report_SetChannel(AI_CHANNEL,1)
m25report.Report_SetChannel(AI_CHANNEL_GOAL,1)
m25report.Report_SetChannel(BRAWL_CHANNEL,0)
m25report.Report_SetChannel(TEAM_CHANNEL,0)
m25report.Report_SetChannel(TARGET_CHANNEL,0)
m25report.Report_SetChannel(TIME_CHANNEL,0)

#-------------------------------------------------------------------------------------
#
# Startup
#
#-------------------------------------------------------------------------------------

# Call this once at the start of a mission to get all of the background functions set up
def InitScripts():
	# if the AI has already been initialized, don't do it again
	if m25getset.GetMissionVar('m25ai_initialised',default=0,integer=1):
		return
	m25getset.SetMissionVar('m25ai_initialised',1)
	
	#parse the AI code to find all subtypes, so that subtype code can be dynamically generated		
	generate_st_switch()		
	
##	mlogreader.MLOG_Init()  # stumpy 2005-10-13: bumped this ahead in the initialization queue

	if m25config.M25_USE_AI:
		datfiles.Campaign_ReadCharacterData()
		datfiles.Campaign_ReadPowers()
		m25timestamp.Timestamp_Setup()
		js.Event_CancelSinks('CloneCheck')
		m25event.SetTimer(3,'CloneCheck')
		AI_SetupMapInfo()
				
	skXMapInfo.GetMapInfo()
	#js.Event_CancelSinks('AIMapInfoTick')
	#m25event.SetTimer(1,'AIMapInfoTick')
		
	if M25_USE_SKXCAMERA:
		skXCamera2.fSkXCameraInitialise(1)
		fSkXWatchMode()
		
	SetDefaultTeamAlliances()
	#m25custompowers.InitCustomPowers()


def AIMapInfoTick(event):
	#print 'AIMapInfoTick'
	start = ff.clock()
	try:
		ff.MapInfo_UpdateFighters()
	except:
		print 'AIMapInfo Update_Fighter did not work.'
		ff.GetMapInfo()

	total = ff.clock() - start
	m25report.Report('TIME', 'skXMapInfo.UpdateFighters took: %f'%(total))
	m25event.SetTimer(1,'AIMapInfoTick')
	
	
# Call this for each character in a mission after they are spawned to start up the custom AI for that character
def SetupOneChar(char):
	SetupAI(char)

# Call this after respawning a character of the same name as one that has previously died, or when reviving a character
def ResetAI(char):
	if IsAISet(char):
		SetObjectVar(char,'ai_flailing',0,nosave=1)
		AISetupTimer(char)
	else:
		SetupOneChar(char)
	
#DMC moved from m25skirmish - initialises a character in Skirmish mode with all the basics
def SetupCharacter(c):
	if not IsAISet(c) and js.Object_GetAttr(c,'physical'):
		SetDefaultTeam(c)
		if GetTeam(c) != TEAM_NONCOMBATANT:
			SetupAI(c)			
			SetBrawler(c)
			AI_SetAlerted(c, 1)
			ai = cshelper.findAI(c)
			ai.removeNaturalKillGoal()

def SetDefaultTactics(char):
	#no longer used
	return
			
#####################################################################################
#	Contents of the rest of this file:
#
#	AI Initialization - setting up the ai
#	AI Main Loop - where all the action is
#	AI Reader - reads ai descriptions
#	AI Variables and Functions - important functions such as enabling and disabling the ai
#	Brawler - getting non-custom AI's to fight according to their assigned teams
#	Dynamic Values - values within tactics descriptions that can change on the fly
#	Goals - dynamic tactics
#	Roles - dynamic subtypes
#	Subtypes - finding characters and objects based on specific conditions
#	Tactic Conditions - checks that have to pass before a tactic will execute
#	Tactics Generic - most of the work in tactics is done here
#	Tactics Main Game - tactics for powers present in the FFvTTR game
#	Tactics FFX - tactics for the FFX addon
#	Targeting - finding a target for tactics
#	Teams - assigning teams and alliances
#	Throwing - such a beast that it needs its own category (tactic for throwing things)
#####################################################################################



#####################################################################################
#
# AI - Initialization
# 
# Description: declare constants and setup the timers and triggers for the AI
#
#####################################################################################

#-------------------------------------------------------------------------------------
#
# Constants
#
#-------------------------------------------------------------------------------------

_aicache = {}

# ai constants
AI_DEFAULTTIMER = 1
AI_AUTOTIMER = 0.5
try:
	AI_AUTOTIMER = m25config.M25_AI_DELAY
except:
	pass

AI_SPEED_DELAY = max(AI_AUTOTIMER/5.0, 0.1)
AI_CALC_SPEED_DELAY = 1
try:
	AI_CALC_SPEED_DELAY = m25config.M25_USE_SPEED_AI_DELAY
except:
	pass

Global_RegisterAttr('ai_delay', AI_AUTOTIMER)


# ai report channels


BRAWL_SKIRMISH = 'skirmish'
BRAWL_OLDTARGET = 'brawloldtarget'
BRAWL_TIMER = 0.5

CLONE_CHECK_TIMER = 1

OBJECT_BRAWLER = 1
OBJECT_NOT_BRAWLER = 0

from m25aiconstants import *

# tactic priority
# TACTIC_PRIORITY_LOW = 20
# TACTIC_PRIORITY_MEDIUM = 50
# TACTIC_PRIORITY_HIGH = 100
# TACTIC_PRIORITY_CRITICAL = 1000

# tactic constants
AI_STANDARD_MOVE_TIME = 1.0
AI_CACHE = 3

# targeting constants
# TARGET_ALLY = 'friend'
# TARGET_ENEMY = 'enemy'
TARGET_PRIORITY_LOW = 2
TARGET_PRIORITY_MEDIUM = 5
TARGET_PRIORITY_HIGH = 10
TARGET_CACHED = 'targetcached'
TARGET_CACHED_TIMER = 1
TARGET_LIST = 'targetlist'



ROLE_AREA_ATTACK = 'using_area_attack'
ROLE_THROWING = 'throwing'

#-------------------------------------------------------------------------------------
#
# Define how the AI statements will be seen (if at all)
#
#-------------------------------------------------------------------------------------

def SetupAIChannel(visibility=REPORT_NONE, goalvis=REPORT_NONE):
	Report_SetChannel(AI_CHANNEL,visibility)
	Report_SetChannel(AI_CHANNEL_GOAL, goalvis)
	Report_SetChannel(BRAWL_CHANNEL, visibility)


#-------------------------------------------------------------------------------------
#
# Setup the custom AI for a character
#
#-------------------------------------------------------------------------------------	
	
def SetupAI(char, force=0):
	#start = ff.clock()
	if (not js.Object_Exists(char)) or (not js.Object_GetAttr(char, 'physical') and not force) or (not ST_char(char) and not force) or IsAISet(char):
		return
	if not IsTeamSet(char): SetDefaultTeam(char)
	if not m25config.M25_USE_AI: return
	Report(AI_CHANNEL,'SetupAI() - setting ai for ' + char + ' (' + `m25template.GetComplexTemplate(char)` + ')')
	ai = FindM25AI(char,default=None)
	if ai != None:
		m25custompowers.AddCustomPowers(char)
		for line in ai:
			if line[0] == 'timer':
				if line[1] == 'auto':
					SetAIAuto(char)
					AI_SetDelay(char, GetAIDelay(char))
					#make sure we leave the game AI ON for controllable characters (PCs)
					if (ffx.FFX_HasPortrait(char)):
						SetAIStandardOnly(char)
					else:
						SetAICustomOnly(char)
				else:
					if not ((type(time) is IntType) or (type(time) is FloatType)):
						m25report.Report(ERROR_CHANNEL, 'SetupTimer(): incorrect timer in tactics for ' + char)
						return
				SetAITimer(char,line[1])
			elif line[0] == 'special targets':
				AI_SetupSpecialTargets(char, line[1:])
			else:
				pass
	#delay starting the AI until the Attributes initialise before the AI tries to use them - TMX per M25 07/18/08
#	AISetupTimer(char)
	m25event.SetTimer(2,'AI_DelayAIStart',char)
	MarkAISet(char)
	if ffx.FFX_HasPortrait(char) and m25config.M25_PLAYERAITOGGLE:
		js.Mission_CustomAction('start ai', '', char, 'PlayerEnableAI', 1000, 1)
	#print 'setupai: total %f'%(ff.clock()-start)

	
def AI_DelayAIStart(event):
	AISetupTimer(event.object)

def AI_SetupSpecialTargets(char, targets):
	if 'buildings' in targets: AI_SetAttacksBuildings(char)
	if 'objects' in targets: AI_SetAttacksObjects(char)
	if 'civilians' in targets: AI_SetAttacksCivilians(char)
	

def PlayerEnableAI(char, hero):
	if AIInWatch(char): 
		pass
	else:
		ClearAIStandardOnly(char)
		AIEnableWatch(char)
		AIEnable(char)
		AIEnableCustom(char)
		SetBrawler(char)
		js.Mission_StatusText('AI started for ' + char + ' (' + m25template.GetComplexTemplate(char) + ')')
	js.Mission_CustomAction('stop ai', '', char, 'PlayerDisableAI', 1000, 1)
		
def PlayerDisableAI(char, hero):
	if not AIInWatch(char): 
		pass
	else:
		SetAIStandardOnly(char)
		AIDisableWatch(char)
		AIDisableCustom(char)
		js.Mission_StatusText('AI stopped for ' + char + ' (' + m25template.GetComplexTemplate(char) + ')')
	js.Mission_CustomAction('start ai', '', char, 'PlayerEnableAI', 1000, 1)
			
def AISetupTimer(char,auto=AI_DEFAULTTIMER):
	if m25missionover.GetMissionOver() or (not Object_Exists(char)):
		return
	if not m25config.M25_USE_AI: return
	timer = GetAITimer(char,default=AI_DEFAULTTIMER)
	if not timer: timer = auto
	m25event.SetTimer(timer,'AICheckTimer',char)
			

#####################################################################################
#
# AI - Main Loop
# 
# Description: 
#	1 get the list of tactics for the type of trigger
#	2 try those tactics one at a time
#	3 enact an ability from a tactic if it is valid
#	4 perform clean up or special actions once the ability is done
#	5 restart the AI for the character
#
#####################################################################################
		
		
#-------------------------------------------------------------------------------------
#
# Triggers
#
#-------------------------------------------------------------------------------------



def AICheckTimer(event):
	#print '----- ' + m25template.GetComplexTemplate(event.object)
	start = ff.clock()
	char = event.object
	
	#if the character has been destroyed, don't do anything
	if not js.Object_Exists(char): return
	
	#stamp the character to let the recovery functions know that the AI is still working
	SetAITimeStamp(char)
	
	#if the AI is disabled, do nothing for now
	#AIStardardEnable and AICustomEnable will restart the AI loop
	if IsAIDisabled(char):
		AISetRunning(char,0)
		js.AI_SetEnabled(char, 0)
		return	
		
	if (ffx.FFX_HasPortrait(char) and not AIInWatch(char)):
		AISetRunning(char, 0)
		return
		
	# if the character is ko'd, but still in the game, leave the AI running in case they get revived
	if js.Object_Exists(char) and not js.Object_IsAlive(char):
		AISetupTimer(char,auto=5)
		return
		
	#if the AI is in a state of knockback, and the MLOG_READER is in use, wait for a callback
	#indicating the character is no longer being knocked back
	if mlogreader.MLOG_IsInKnockBack(char):
		mlogreader.regKnockBackStop(char, 'm25ai.AIGetupStop')
		#mlogreader.regIdleStart(char, 'm25ai.AIGetupStop')
		return

	#if the AI is disabled in any way, check when released from that state		
 	if not AICharacterStateCheck(char):
 		AISetupTimer(char, auto=2)
 		return		
		
	#if the character is panicked or blanked, apply special movement rules to them	
	if AISpecialState(char):
		#AISetupTimer(char)
		return
		
	mark = ff.clock()
	#print 'prelim ' + `mark-start`
	
	tactic_id = 0
	Target_ClearTargets(char, TARGET_ENEMY)
	try:
		target = FindTarget(char, TARGET_ENEMY)
		target_ally = FindTarget(char, TARGET_ALLY)
	except:
		m25report.Report('AI', 'Unable to find targets for %s'%char)
		AISetupTimer(char, auto=2)
 		return

	marka = ff.clock()
	#print 'targets ' + `marka-mark`
	
 	try:
 		AIDoGoals(char)
 	except:
 		Report(AI_CHANNEL_GOAL,'Unable to perform goals for ' + char)
 		print sys.exc_type, sys.exc_value
 		#print sys.exc_info()[0], sys.exc_info()[1], str(sys.exc_info()[2])
 		if ALLOW_ERRORS:
 			raise
 	
	mark2 = ff.clock()
	#print 'goals ' + `mark2-marka`
			
	#try trash talk
	#if M25_USE_BANTER:
	#	DoBanter(char)	

	#do the actual tactics
	if AIHasTimerTactics(char) and not IsAICustomDisabled(char):
		try: tactic_id = AIExecTactics(event)
		except:
			Report(AI_CHANNEL, 'Unable to do tactics for ' + char)
			print sys.exc_type, sys.exc_value
			#print sys.exc_info()[0], sys.exc_info()[1], str(sys.exc_info()[2])
 			if ALLOW_ERRORS:
 				raise
 	#if not a custom AI, try using standard brawl rules
 	else:
 		if not IsAIStandardDisabled(char):
 			DoBrawl(char)	

	if IsAIAuto(char):
 		if not IsPerformingTactic(char):
			Report(AI_CHANNEL,char + ' no tactic chosen this cycle.')
			time = GetAutoAIDelay(char)
			AISetupTimer(char,auto=time)
		else:
			m25event.SetTimer(10.0, 'AIRecoverBroken', char, user=GetAITimeStamp(char))		
	else:
		AISetupTimer(char)
	
	end = ff.clock()
	ai_exec_time = end-start
	m25report.ReportTime('ai for %s'%(m25template.GetComplexTemplate(event.object)), ai_exec_time)


def AIRecoverBroken(event):
	stamp = event.user
	char = event.object
	if not js.Object_Exists(char): return
	if not js.Object_IsAlive(char): return
	old_stamp = GetAITimeStamp(char)
	if stamp == old_stamp:  #nothing happened for 15 or so seconds
		mlogreader.MLOG_ResetKnockBack(char)
		if AI_CHANNEL: m25report.Report(AI_CHANNEL, 'Had to recover AI for %s'%char)
		AIRestart(char,100000)

# Determine how long to wait before invoking the AI
# I'm using longer timers when the AI isn't likely to need to take action
# so as to increase performance	
def GetAutoAIDelay(char):
	if not js.Object_Exists(char): return AI_AUTOTIMER
 	if not AICharacterStateCheck(char):  #character is in stasis, stunned or generally cannot act at this time
 		return 4 + AI_AUTOTIMER + randint(0,2)
 	target = Target_GetCurrentTarget(char, TARGET_ENEMY)
 	if (target == None):
 		return 4 + AI_AUTOTIMER + randint(0,2)
 	if (m25distance.GetDistance(char,target) > m25distance.DISTANCE_VISIBLE):
 		return 1 + AI_AUTOTIMER + randint(0,2)
	if AI_CALC_SPEED_DELAY:
		tm = (11-js.Object_GetAttr(char, 'speed')) * AI_SPEED_DELAY
	else:
		tm = AI_GetDelay(char, AI_AUTOTIMER)
	if tm <= 0.1: tm = 0.1
	return tm
# 	#if AIInKnockBack(char):
# 	#	return 2 + AI_AUTOTIMER
# 	return AI_AUTOTIMER
	

#-------------------------------------------------------------------------------------
#
# Support for the event sinks from mlogreader
# Makes the AI more natural in its actions and may increase performance
#
#-------------------------------------------------------------------------------------

def AIFlailStart(event):
	print 'AIFLAILSTART ' + event.object

def AIFlailStop(event):
	char = event.object
	print 'AIFLAILSTOP for ' + char
	regLandStop(char,'AILandStop')

def AILandStart(event):
	char = event.object
	print 'AILANDSTART for ' + char	
	regLandStop(char,'AILandStop')

def AIGetupStop(event):
	#resume the AI
	char = event.object
	#print 'AILANDSTOP for ' + char
	AISetupTimer(char)

	
#-------------------------------------------------------------------------------------
#
# Do Tactics
#
#-------------------------------------------------------------------------------------

def AISpecialState(char):
	if ST_blanked(char):
		AISupport_DoMove(char, Target_GetCurrentTarget(char), 'walk', 1, FindRandomPoint(char,mindist=100), js.Get_ObjectPos(char), TACTIC_PRIORITY_CRITICAL, 2, desc='blanked walk')
		return 1
	if ST_panicked(char):
		AISupport_DoMove(char, Target_GetCurrentTarget(char), 'move', 0, FindRandomPoint(char,mindist=100), js.Get_ObjectPos(char), TACTIC_PRIORITY_CRITICAL, 2, desc='panicked run')
		return 1
	return 0

ai_compiled_tactics = {}
								
def AIExecTactics(event, tactics=''):
	global ai_compiled_tactics
	char = event.object
	template = m25template.GetComplexTemplate(event.object)
	#template = js.Object_GetTemplate(char)
	current_priority = GetAIPriority(char)
	tactic_id = 0
	tac_string = ''
	if ai_compiled_tactics.has_key(template):
		tac_string = ai_compiled_tactics[template]
	else:
		tac_string = m25aiopt.Optimize_GenerateOptimizedTactics(char,GetAITimerTactics(char)[1:],GetAIEnergyReserve(char))
		ai_compiled_tactics[template] = tac_string
	exec tac_string


#-------------------------------------------------------------------------------------
#
# Enact Abilities
#
#-------------------------------------------------------------------------------------


def AIDoAnimation(char, anim='', fx='', priority=TACTIC_PRIORITY_LOW, time=3, function='', description=''):
	AIDisableStandard(char)
	if m25string.IsString(priority): priority = AIGetCurrentTacticPriority(char)
	SetAIPriority(char,priority)
	if anim != '':
		AI_Animate(char,anim)
	if fx != '':
		try:    #stumpy 2006-08-26 - mod because lack of certain FX would kill AI
			js.Object_PlayEffect(char,fx, '', js.FX_TRACK_OBJECT_POS)
		except RuntimeError:
			print 'm25ai.AIDoAnimation: error playing <%s> FX for <%s>\n\t%s\t%s' % (fx,char,sys.exc_type, sys.exc_value)
	m25event.SetTimer(time,'AITacticDone',char,user=priority)
	SetPerformingTactic(char)
	Report(AI_CHANNEL,char + ':' + `m25template.GetComplexTemplate(char)` + ' ' + anim + ' ' + fx + ' ' + description)

def AIDoGeneric(char,priority=TACTIC_PRIORITY_LOW,time=3,function='',description=''):
	AIDisableStandard(char)
	if m25string.IsString(priority): priority = AIGetCurrentTacticPriority(char)
	SetAIPriority(char,priority)
	m25event.SetTimer(time,'AITacticDone',char,user=priority)
	SetPerformingTactic(char)
	Report(AI_CHANNEL,char + ':' + `m25template.GetComplexTemplate(char)` + ' ' + description)	
	
# do movement	
def AIDoMove(char, mode=None, target=None, priority=TACTIC_PRIORITY_LOW, time=AI_STANDARD_MOVE_TIME, speed=0, description='', mindist=30, flee=0):
	AIDisableStandard(char)
	if m25string.IsString(priority): priority = AIGetCurrentTacticPriority(char)
	SetAIPriority(char,priority)
	SetPerformingTactic(char)
	if mode == None:
		mode = DetermineMoveType(char,target,flee)
	if target == None:
		Trigger_Move(char,mode)
		target = ''
	else:
		pos = 0
		dist = mindist
		if type(target) == types.TupleType:
			pos = target
		else:
			try:
				pos = Get_ObjectPos(target)
			except:
				try:
					pos = Get_MarkerPos(target)
					dist=5
				except:
					pos = 0
		if pos != 0:
			#print 'AIDoMove ' + `pos`
			Trigger_Move(char,mode,pos,'',dist,speed)
			cpos = js.Get_ObjectPos(char)
			#print 'move distance ' + `m25distance.GetTupleDistance(pos,cpos)`
			multiplier = 20
			if mode == 'fly': multiplier = 40			
			speed = js.Object_GetAttr(char, 'speed')+1
			if speed == 0: speed = 1
			time = m25distance.GetTupleDistance(pos,cpos)/(speed*multiplier)
		try:
			cpos = js.Get_ObjectPos(char)
			SetObjectVar(char,'ai_last_x',cpos[0],nosave=1)
			SetObjectVar(char,'ai_last_y',cpos[1],nosave=1)
			#print 'setting aichecknotmoving ' + `priority`
			timetarget = target
			if type(target) == types.TupleType:
				timetarget = `(int(target[0]),int(target[1]),int(target[2]))`
			m25event.SetTimer(0.5,'AICheckNotMoving',char,timetarget,user=1024*priority+GetAITimeStamp(char))
		except:
			raise
			pass
	#print 'move time = ' + `time` 
	if time < 0.5: time = 0.5
	else: 
		if time > 5.0: time = 5.0
	m25event.SetTimer(time-0.3,'AITacticDone',char,user=priority)
	Report(AI_CHANNEL,char + ':' + `m25template.GetComplexTemplate(char)` + ' ' + description + ' ' + `target`)
	

	

def ai_destring(val):
	return val			
			
# trigger powers	
	

	
def AIDoPower(char, power='', target='', priority=TACTIC_PRIORITY_MEDIUM, level=0, time=3, function='', description='', end_function=None, end_function_time=5, use_timer=0):
	AIDisableStandard(char)
	if m25string.IsString(priority): priority = int(AIGetCurrentTacticPriority(char))
	SetAIPriority(char,priority)
	if type(target) == types.TupleType:
		if js.Marker_Exists('ai_travel_marker'):
			js.Marker_SetPos('ai_travel_marker', target)
			js.Object_MoveToMarker('ai_travel_object','ai_travel_marker')
		else:
			js.Marker_Spawn('ai_travel_marker', target)
			js.Object_Spawn('ai_travel_object', 'ffx_tiggot', 'ai_travel_marker')
			js.Object_SetSecondaryState('ai_travel_object', js.SCSTATE_INVISIBLE, 1, js.PERMANENT_STATE)
			js.Object_SetAttr('ai_travel_object', 'minHealth', 1)
	index = m25custompowers.GetCarrierIndex(char,power)
	#print 'carrier index ' + power + ' ' + `index`
	#if (function == '') and (not m25custompowers.IsCarrierInit(char,power)) and (index > -1):
	if (function == '') and (index > -1):
		user=index+1024*int(js.Object_GetAttr(char,'ffxID'))
		if type(target) == types.TupleType:
			js.Trigger_Power(char,'ai_travel_object',power,'SetupCustomCarrier',level,target,user)
		else:
			js.Trigger_Power(char,target,power,'SetupCustomCarrier',level,js.Get_ObjectPos(char),user)
		#m25custompowers.SetCarrierInit(char,power)
	else:
		if type(target) == types.TupleType:
			js.Trigger_Power(char,'ai_travel_object',power,function,level,target)
		else:
			if target != None and js.Object_Exists(target):
				js.Trigger_Power(char,target,power,function,level)
			else:
				#something went wrong with the target, resart the AI quickly
				use_timer = 1
				time = 0.1
	#if type(target) == types.TupleType: target = `target`
	#if not use_timer and m25config.M25_USE_MLOG: #Ep. 2007-03-31: MLOG is now mandatory
	if not use_timer:
		regIdleStart(char, 'AITacticDone', user=priority)
		m25event.SetTimer(0.3, 'AICheckNotActing', char, `priority`,user=GetAITimeStamp(char))
	else:
		m25event.SetTimer(time,'AITacticDone',char,user=priority)
	if end_function != None:
		m25event.SetTimer(end_function_time,end_function,char,str(target),user=priority)
	SetPerformingTactic(char)
	Report(AI_CHANNEL,char + ':' + `m25template.GetComplexTemplate(char)` + ' ' + power + ' ' + description + ' ' + `target`)


	
	
#-------------------------------------------------------------------------------------
#
# Recovery Functions
# -ensure that the AI doesn't get blocked
#
#-------------------------------------------------------------------------------------	
	

def AICheckNotActing(event):
	if not js.Object_Exists(event.object) or not mlogreader.MLOG_IsIdle(event.object): return
	if IsAIDisabled(event.object): return
	char = event.object
	stamp = event.user
	if stamp < GetAITimeStamp(char): return
	if int(event.string) < GetAIPriority(event.object): return
	if AI_CHANNEL:
		Report(AI_CHANNEL, "%s: (%s) random move on inability to act."%(char,`m25template.GetComplexTemplate(char)`))
	object = Target_GetCurrentTarget(char)
	if object == None: return
	mode = 'jump'
	if AIIsFlying(char): mode = 'move'
	pos = FindRandomPoint(char,maxdist=m25distance.DISTANCE_SHORT,default=None)
	js.Trigger_Move(char,mode,pos,'',0,0)
#     try:		
#         Trigger_Move(char,mode,js.Get_ObjectPos(object),'',0,0)
#     except:
#         Report(AI_CHANNEL,char + ':' + `m25template.GetComplexTemplate(char)` + ' trying random jump.')
#         pos = FindRandomPoint(char,maxdist=m25distance.DISTANCE_SHORT,default=None)
#         Trigger_Move(char,mode,pos,'',0,0)


def AICheckNotMoving(event):
	if not js.Object_Exists(event.object): return
	if IsAIDisabled(event.object): return
	#don't do anything if the AI has already restarted
	priority = int(event.user)&1023
	timestamp = int(event.user/1024)
	if event.user < int(GetAITimeStamp(event.object)) or (priority < GetAIPriority(event.object)): return
	if not mlogreader.MLOG_IsIdle(event.object): return
	else: return
	if Target_GetCurrentTarget(event.object) == None: return
	try:
		mode = 'jump'
		if AIIsFlying(event.object): mode = 'move'
		m25report.Report(AI_CHANNEL, event.object + ' bad move. Now trying to jump/move.')
		if ST_isranged(event.object):
			pos = FindRandomPoint(event.object,maxdist=m25distance.DISTANCE_SHORT,default=None)
			js.Trigger_Move(event.object,mode,pos,'',0,0)
		else:
			Trigger_Move(event.object,mode,js.Get_ObjectPos(Target_GetCurrentTarget(event.object)),'',0,0)
	except:
		m25report.Report(AI_CHANNEL, event.object + ' unable to move at all!')


#-------------------------------------------------------------------------------------
#
# Abilities Completed
#
#-------------------------------------------------------------------------------------

def AIAreaDone(event):
	char = event.object
	RemoveRole(char,ROLE_AREA_ATTACK)

def AICloneDone(event):
	char = event.object
	allchars = js.Mission_GetDynamicObjects()
	for c in allchars:
		if (c[:5] == 'mobj_') and (js.Object_GetTemplate(c) == js.Object_GetTemplate(char)) and (GetObjectVar(c,'_clonefound',default=0,integer=1)==0):
			Report(AI_CHANNEL,'CloneDone(): ' + char + ' has clone ' + c)
			SetObjectVar(c,'_clonefound',1)
			AISetClone(c)
			SetupOneChar(c)
			print 'setting ' + c + ' to team ' + `GetTeam(char)`
			SetTeam(c,GetTeam(char))
			SetBrawler(c)

def AICloneTargetDone(event):
	char = event.object
	target = event.string
	allchars = js.Mission_GetDynamicObjects()
	for c in allchars:
		if (c[:5] == 'mobj_') and (js.Object_GetTemplate(c) == js.Object_GetTemplate(target)) and (not GetObjectVar(c,'_clonefound',default=0,integer=1)):
			SetObjectVar(c,'_clonefound',1)
			AISetClone(c)
			SetupOneChar(c)
			SetTeam(c,GetTeam(char))
			SetBrawler(c)

def AIDecoyDone(event):
	char = event.object
	allchars = js.Mission_GetDynamicObjects()
	for c in allchars:
		if (js.Object_GetTemplate(c) == 'standard_image_projection') and (not GetObjectVar(c,'_decoyfound',default=0,integer=1)):
			SetObjectVar(c,'_decoyfound',1)
			AISetDecoy(c)
			SetTeam(c,GetTeam(char))			
			
def AILivingFiresDone(event):
	print 'living fires done'
	char = event.object
	allchars = js.Mission_GetDynamicObjects()
	for c in allchars:
		if (js.Object_GetTemplate(c) == 'fire_elemental') and (not GetObjectVar(c,'_livingfiresfound',default=0,integer=1)):
			print c + ' living fire found'
			SetObjectVar(c,'_livingfiresfound',1)
			SetTeam(c,GetTeam(char))
			SetBrawler(c)
			
def AIPolymorphDone(event):
	char = event.object
	target = event.string
	#print 'polymorph done ' + char + ' ' + target
	allchars = js.Mission_GetDynamicObjects()
	for c in allchars:
		if (c[:5] == 'mobj_') and (not IsAISet(c)) and (js.Object_GetAttr(c,'physical')): # and (js.Object_GetPrimaryState(c) == js.PCSTATE_POLYMORPHED):
			#print 'polymorph found ' + c
			SetupOneChar(c)
			SetTeam(c,GetTeam(char))
			SetBrawler(c)
	
def AITacticDone(event):
	char = event.object
	if not js.Object_Exists(char): return
	priority = int(event.user)
	if IsAIDisabled(char) or IsAIAllDisabled():
		m25event.SetTimer(1,'AITacticDone',char,user=priority)
		return
	# VERY IMPORTANT - restart the AI
	AIRestart(char,priority)

	
#-------------------------------------------------------------------------------------
#
# Restart the AI
# put it back in a state where it will kick in again
#
#-------------------------------------------------------------------------------------

def AIRestart(char,priority):
	if not js.Object_Exists(char): return
	if priority < GetAIPriority(char):
		return  # a higher priority tactic has overriden this one
	if not IsAICustomOnly(char) and not IsAIDisabled(char) and not InPeace():
		AIEnableStandard(char)
	ClearPerformingTactic(char)
	SetAIPriority(char, 0)
	if IsAIAuto(char):
		#print 'AIRestart: ' + char + ' ' + `AI_AUTOTIMER`
		#time = AI_AUTOTIMER
		if AI_CALC_SPEED_DELAY:
			tm = (11-js.Object_GetAttr(char, 'speed')) * AI_SPEED_DELAY
		else:
			tm = AI_GetDelay(char, AI_AUTOTIMER)
		if tm <= 0.1: tm = 0.1
		AISetupTimer(char,auto=tm) 


#####################################################################################
#
# AI Reader
#
# Description: read the contents of the m25aidata file
#
#####################################################################################

def FindM25AI(char,default=[],istemplate=0):
	global AIExtensions_Dictionary
	template = ''
	if 'AIExtensions_Dictionary' not in globals().keys():
		AIExtensions_Dictionary = {}
	if not istemplate:
		template = m25template.GetComplexTemplate(char)
		if template == None: return default
	else:
		template = char
	#try dictionary first
	if AIExtensions_Dictionary.has_key(template):
		return AIExtensions_Dictionary[template]
	#print 'Findm25ai ' + `char`
	#see if we have a hero file
	if m25custompowers.TemplateIsHeroFile(template):
		try:
			filename = os.path.join(datfiles.GetHeroFilePath(),template+'.m25ai')
			f = open(filename,'r')
			cust = f.read()
			f.close()
			try:
				ai = eval(cust)
				AIExtensions_Dictionary[template] = ai
				#print 'Findm25ai ' + `char` + ' in hero file '
				return ai
			except:
				m25report.Report(ERROR_CHANNEL,'There is an error in ' + template + '.m25ai')
				AIExtensions_Dictionary[template] = []
				return default
		except:
			pass
		#check the m25aidata file
	for ai in m25aidata.AIExtensions:
		if ai[0] == template:
			#print 'FindM25AI ' + template + ' found in aiextensions file.'
			AIExtensions_Dictionary[template] = ai
			return ai
	#print 'Findm25ai ' + `template` + ' was not found '
	AIExtensions_Dictionary[template] = []
	return default
	
def FindTacticType(char,tactic,default=[],istemplate=0):
	details = FindM25AI(char,istemplate=istemplate)
	if details != []:
		for m in details[1:]:
			if m[0] == tactic:
				return m[1:]
	return default

def GetAISubtypes(char,default=[],istemplate=0):
	return FindTacticType(char,'subtype',default,istemplate=istemplate)

def GetAITargetTypes(char,reaction='enemy',default=[]):
	val = m25getset.GetObjectVar(char, 'ai_target_'+reaction, default=None, nosave=1)
	if val == None:
		val = FindTacticType(char,'target '+reaction,default)	
		m25getset.SetObjectVar(char, 'ai_target_'+reaction, val, nosave=1)
	return val

def GetAITimerTactics(char,default=[]):
	return FindTacticType(char,'timer',default)

	
Global_RegisterAttr('ai_timer_delay',AI_DEFAULTTIMER)	
def GetAITimer(char, default=1):
	return js.Object_GetAttr(char, 'ai_timer_delay')
	#return m25getset.GetObjectVar(char,'ai_timer_delay',default=default)

def SetAITimer(char, value=1, default=1):
	if value == 'auto': value = 0
	js.Object_SetAttr(char, 'ai_timer_delay', value)
	#m25getset.SetObjectVar(char,'ai_timer_delay', value)
	
def AIHasTimerTactics(char):
	val = m25getset.GetObjectVar(char, 'ai_has_timer_tactics', default=None, nosave=1)
	if val != None: return val
	val = 0
	t = GetAITimerTactics(char, default=None)
	if t != None:
		val = 1
	m25getset.SetObjectVar(char, 'ai_has_timer_tactics', val, nosave=1)
	return val
	
def GetAIEnergyReserve(char,default=0):
	val = m25getset.GetObjectVar(char, 'ai_energy_reserve', default=None, nosave=1)
	if val != None:
		return val
	val = default
	vallist = FindTacticType(char,'energy reserve',default)
	#if vallist == default: 
	#	return default
	if type(vallist) == types.ListType:
		val = vallist[0]
		#return vallist[0]
	m25getset.SetObjectVar(char, 'ai_energy_reserve', val, nosave=1)
	return val
	
def GetAIDelay(char, default=1.0):
    val = m25getset.GetObjectVar(char, 'ai_inter_delay', default=None, nosave=1)
    if val != None:
        return val
    val = default
    vallist = FindTacticType(char, 'delay', default)
    if type(vallist) == types.ListType: val = vallist[0]
    m25getset.SetObjectVar(char, 'ai_inter_delay', val, nosave=1)
    return val
    
def GetAIRunAnimTime(char,default=1):
	vallist = FindTacticType(char,'run anim time',default)
	if vallist == default: 
		return default
	if type(vallist) == types.ListType:
		return vallist[0]
	return default	
	
def AddM25AI(char,ai):
	if Object_Exists(char):
		template = []
		template = m25template.GetComplexTemplate(char)
		for ai in m25aidata.AIExtensions:
			if ai[0] == template:
				m25report.Report(ERROR_CHANNEL, 'AddM25AI() - ai definition for ' + template + ' already exists.')
				return
		m25aidata.AIExtensions.append(ai)
		AIReaderSetAI(char,ai)

def AIReaderSetAI(char, ai):
	global AIExtensions_Dictionary
	if 'AIExtensions_Dictionary' not in globals().keys():
		AIExtensions_Dictionary = {}
	template = m25template.GetComplexTemplate(char)
	AIExtensions_Dictionary[template] = ai

#####################################################################################
#
# AI - Variables and Functions
# 
# Description: Important properties of the AI - includes enabling and disabling
#
#####################################################################################

#AI delay is used to allow processor time for other activities between ai decisions

def AI_GetDelay(char, default=0):
	return js.Object_GetAttr(char, 'ai_delay')
	
def AI_SetDelay(char, value):
	js.Object_SetAttr(char, 'ai_delay', value)

		
#enable/disable both custom and standard AI

Global_RegisterAttr('ai_running', 1)
Global_RegisterAttr('ai_disabled', 0)
Global_RegisterAttr('ai_standard_disabled', 0)
Global_RegisterAttr('ai_standard_can_enable', 0)
Global_RegisterAttr('ai_custom_disabled', 0)
Global_RegisterAttr('ai_custom_can_enable', 0)
Global_RegisterAttr('ai_can_enable', 0)
Global_RegisterAttr('ai_custom_busy', 0)
Global_RegisterAttr('ai_custom_only', 0)


	

def IsAIRunning(char):
	return js.Object_GetAttr(char, 'ai_running')
	
def AISetRunning(char, value=1):
	m25report.Report(AI_CHANNEL, 'AISetRunning %s to %i'%(char,value))
	js.Object_SetAttr(char, 'ai_running', value)

	
def AIRun(char):
	if ffx.FFX_HasPortrait(char) and not AIInWatch(char): return
	if not IsAIRunning(char) and IsAISet(char):		
		m25report.Report(AI_CHANNEL, 'AIRun is restarting %s'%(char))
		AISetRunning(char,1)
		AISetupTimer(char,auto=1)
	
def IsAIDisabled(char):
	return js.Object_GetAttr(char,'ai_disabled')
	
def IsAIStandardDisabled(char):
	return js.Object_GetAttr(char,'ai_standard_disabled')
	
	
def AIDisable(char, peacecall=0):
	if InPeace(): 
		js.Object_SetAttr(char, 'ai_can_enable',0)
	js.Object_SetAttr(char, 'ai_disabled', 1)
	AIDisableStandard(char)
	AIDisableCustom(char)
	
def AIEnable(char, peacecall=0): 
	if InPeace(): js.Object_SetAttr(char, 'ai_can_enable',1)
	else:
		js.Object_SetAttr(char, 'ai_disabled', 0)
	if not IsAICustomOnly(char): AIEnableStandard(char)
	if not IsAIStandardOnly(char): AIEnableCustom(char)	

# testing for using goals only	
	
def AIEnableGoalsOnly(char):
	SetObjectVar(char,'goals_only',1)
	
def AIDisableGoalsOnly(char):
	SetObjectVar(char,'goals_only',0)
	
def IsAIGoalsOnly(char):
	return GetObjectVar(char,'goals_only',default=0,integer=1)	
	
#Standard (default game) AI	enable/disable
	
def AIDisableStandard(char, peacecall=0):
	if InPeace():
		js.Object_SetAttr(char, 'ai_standard_can_enable', 0)
	#else:
	js.Object_SetAttr(char, 'ai_standard_disabled', 1)
	js.AI_SetEnabled(char,0)
	
def AIEnableStandard(char, peacecall=0):
	if InPeace():
		if not IsAICustomOnly(char):
			js.Object_SetAttr(char, 'ai_standard_can_enable', 1)
	else:
		if not (IsAIDisabled(char) or IsAICustomOnly(char)):
			js.Object_SetAttr(char, 'ai_standard_disabled', 0)
			AIRun(char)
			js.AI_SetEnabled(char,1)
		
#Custom AI enable/disable
		
def IsAICustomDisabled(char):
	return js.Object_GetAttr(char,'ai_custom_disabled')
	
def AIDisableCustom(char, peacecall=0):
	if InPeace():
		js.Object_SetAttr(char, 'ai_custom_can_enable',0)
#	else:
	js.Object_SetAttr(char,'ai_custom_disabled',1)
	
def AIEnableCustom(char, peacecall=0):
	if InPeace():
		if not IsAIStandardOnly(char):
			js.Object_SetAttr(char, 'ai_custom_can_enable',1)
	else:	
		if not (IsAIDisabled(char) or IsAIStandardOnly(char)):
			js.Object_SetAttr(char,'ai_custom_disabled',0)
			AIRun(char)

	
# Entire mission AI Disable/enable	
	
def IsAIAllDisabled():
	return js.Mission_TestAttr('peace')
	
def InPeace():	
	return IsAIAllDisabled()
	
def AIPeace():
	if IsAIAllDisabled():
		return
	AIDisableAll()
	
def AIWar():
	if not InPeace(): return
	AIEnableAll()	
	
def AIDisableAll():
	if not InPeace():
		js.Mission_SetAttr('peace',1)
		objects = Mission_GetDynamicObjects()
		for obj in objects:
			js.Object_SetAttr(obj, 'ai_can_enable', not js.Object_GetAttr(obj, 'ai_disabled'))
			js.Object_SetAttr(obj, 'ai_standard_can_enable', not js.Object_GetAttr(obj, 'ai_standard_disabled'))
			js.Object_SetAttr(obj, 'ai_custom_can_enable', not js.Object_GetAttr(obj, 'ai_custom_disabled'))
			js.Object_SetAttr(obj, 'ai_disabled', 1)
			js.Object_SetAttr(obj, 'ai_standard_disabled',1)
			js.Object_SetAttr(obj, 'ai_custom_disabled',1)
			js.AI_SetEnabled(obj, 0)

def AIEnableAll():
	js.Mission_SetAttr('peace',0)
 	objects = Mission_GetDynamicObjects()
 	for obj in objects:
 		js.Object_SetAttr(obj, 'ai_disabled', not js.Object_GetAttr(obj, 'ai_can_enable'))
		js.Object_SetAttr(obj, 'ai_standard_disabled', not js.Object_GetAttr(obj, 'ai_standard_can_enable'))
		js.Object_SetAttr(obj, 'ai_custom_disabled', not js.Object_GetAttr(obj, 'ai_custom_can_enable'))
		js.Object_SetAttr(obj, 'ai_can_enable', 0)
		js.Object_SetAttr(obj, 'ai_standard_can_enable',0)
		js.Object_SetAttr(obj, 'ai_custom_can_enable',0)
		if not (IsAICustomDisabled(obj) and IsAIStandardDisabled(obj)): AIRun(obj)
		if not IsAIStandardDisabled(obj): js.AI_SetEnabled(obj, 1)

# Is the AI executing a tactic?				

def IsPerformingTactic(char):
	return js.Object_GetAttr(char,'ai_custom_busy')
	
def SetPerformingTactic(char):
	js.Object_SetAttr(char,'ai_custom_busy',1)
	
def ClearPerformingTactic(char):
	js.Object_SetAttr(char,'ai_custom_busy',0)


# Custom AI only (no standard AI)		
		
def IsAICustomOnly(char):
	return js.Object_GetAttr(char, 'ai_custom_only')	
	
def SetAICustomOnly(char):
	js.Object_SetAttr(char, 'ai_custom_only',1)
	cshelper.disable(char)
	
def ClearAICustomOnly(char):
	js.Object_SetAttr(char,'ai_custom_only',0)
	cshelper.enable(char)
	
Global_RegisterAttr('ai_standard_only', 0)	
	
def IsAIStandardOnly(char):
	return js.Object_GetAttr(char, 'ai_standard_only')
	
def SetAIStandardOnly(char):
	js.Object_SetAttr(char, 'ai_standard_only', 1)
	
def ClearAIStandardOnly(char):
	js.Object_SetAttr(char, 'ai_standard_only', 0)
	
# AI has been initialized
	
Global_RegisterAttr('ai_initialized', 0)
Global_RegisterAttr('ai_priority', 0)
Global_RegisterAttr('ai_actionstamp', 0)
Global_RegisterAttr('ai_auto', 0)
Global_RegisterAttr('ai_decoy', 0)
Global_RegisterAttr('ai_clone', 0)
	
def IsAISet(char):
	return js.Object_GetAttr(char,'ai_initialized')
	
def MarkAISet(char,val=1):
	js.Object_SetAttr(char, 'ai_initialized', val)
			
# Priority of the current tactic the custom AI is executing	
	
def GetAIPriority(char):
	return int(js.Object_GetAttr(char,'ai_priority'))
	
def SetAIPriority(char,value):
	js.Object_SetAttr(char, 'ai_priority',value)

def GetAITimeStamp(char):
	return js.Object_GetAttr(char, 'ai_actionstamp')	
	
def SetAITimeStamp(char):
	js.Object_SetAttr(char, 'ai_actionstamp',m25timestamp.Timestamp_GetTime())
	
# Is the timer set to 'auto'?	
		
def IsAIAuto(char):
	return js.Object_GetAttr(char, 'ai_auto')
	
def SetAIAuto(char):
	js.Object_SetAttr(char, 'ai_auto', 1)

# is this character actually a decoy object
	
def AIIsDecoy(char):
	return js.Object_GetAttr(char, 'ai_decoy')	

def AISetDecoy(char):
	js.Object_SetAttr(char, 'ai_decoy',1)

# is this character a clone

def AIIsClone(char):
	return js.Object_GetAttr(char, 'ai_clone')

def AISetClone(char):
	js.Object_SetAttr(char, 'ai_clone', 1)
	
	
# current tactic priority

Global_RegisterAttr('ai_currenttacpriority', 0)
Global_RegisterAttr('ai_watch_mode', 0)
Global_RegisterAttr('ai_attacks_buildings', 0)
Global_RegisterAttr('ai_attacks_civilians', 0)
Global_RegisterAttr('ai_attacks_objects', 0)
Global_RegisterAttr('ai_ignore_target', 0)

def AIGetCurrentTacticPriority(char):
	#print 'current tactic priority ' + `GetObjectVar(char,'current_tactic_priority',default=0,integer=1)`
	return int(js.Object_GetAttr(char, 'ai_currenttacpriority'))
	#return int(GetObjectVar(char,'current_tactic_priority',default=0,integer=1))
	
def AISetCurrentTacticPriority(char,priority):
	js.Object_SetAttr(char, 'ai_currenttacpriority',priority)
	#return SetObjectVar(char,'current_tactic_priority',priority)

# hero watch mode

def AIInWatch(char):
	return js.Object_GetAttr(char, 'ai_watch_mode')
	#return GetObjectVar(char, 'ai_watch_mode', default=0,integer=1)
	
def AIEnableWatch(char):
	js.Object_SetAttr(char, 'ai_watch_mode', 1)
	#SetObjectVar(char, 'ai_watch_mode', 1)
	
def AIDisableWatch(char):
	js.Object_SetAttr(char, 'ai_watch_mode', 0)
	#SetObjectVar(char, 'ai_watch_mode', 0)
	
	
# ignore the target

def AI_GetIgnore(char, target=None):
	if not target:
		return js.Object_GetAttr(char, 'ai_ignore_target')
	else:
		return (js.Object_GetAttr(char, 'ai_ignore_target')) or (target in m25getset.GetObjectVar(char, 'ai_ignore_list', default=[]))
	
def AI_SetIgnore(char, target=None):
	if target==None:
		js.Object_SetAttr(char, 'ai_ignore_target', 1)
	else:
		m25getset.SetObjectVar(char, 'ai_ignore_list', m25getset.GetObjectVar(char, 'ai_ignore_list',default=[]).append(target))
	
def AI_RemoveIgnore(char, target=None):
	if not target:
		js.Object_SetAttr(char, 'ai_ignore_target', 0)		
	else:
		tlist = m25getset.GetObjectVar(char, 'ai_ignore_list', default=[])
		if target in tlist: 
			tlist.remove(target)
			m25getset.SetObjectVar(char, 'ai_ignore_list', tlist)
		
		
# special settings for attacking buildings and objects	

def AI_SetAttacksCivilians(char, val=1):
	js.Object_SetAttr(char, 'ai_attacks_civilians', 1)
	
def AI_GetAttacksCivilians(char):
	return js.Object_GetAttr(char, 'ai_attacks_civilians')

def AI_SetAttacksBuildings(char, val=1):
	js.Object_SetAttr(char, 'ai_attacks_buildings', 1)
	#m25getset.SetObjectVar(char, 'ai_attacks_buildings', val)
	
def AI_GetAttacksBuildings(char):
	return js.Object_GetAttr(char, 'ai_attacks_buildings')
	#return m25getset.GetObjectVar(char, 'ai_attacks_buildings', default=0) 
	
def AI_SetAttacksObjects(char, val=1):
	js.Object_SetAttr(char, 'ai_attacks_objects', 1)
	#m25getset.SetObjectVar(char, 'ai_attacks_objects', val)
	#print m25getset.GetObjectVar(char, 'ai_attacks_objects', default=0)
	
def AI_GetAttacksObjects(char):
	return js.Object_GetAttr(char, 'ai_attacks_objects')
	#print m25getset.GetObjectVar(char, 'ai_attacks_objects', default=0)
	#return m25getset.GetObjectVar(char, 'ai_attacks_objects', default=0)
	
	
#####################################################################################
#
# Banter
# 
# Description: characters 'trash talk' while fighting
#
#####################################################################################		

def DoBanter(char):
	if not Object_Exists(char) or not Object_IsAlive(char): return
	template = m25template.GetComplexTemplate(char)
	banter = m25getset.GetObjectVar(char,'banter_enabled',default=None,nosave=1)
	if banter == 0: return
	
	# variable percent chance to say something
	last_banter = m25getset.GetMissionVar('ai_last_banter',default=0,nosave=1)
	current = m25timestamp.Timestamp_GetTime()
	adjust = current-last_banter
	if randint(1,100) > adjust/2: return
	m25getset.SetMissionVar('ai_last_banter',current,nosave=1)
	
	if banter == None:
		if m25cutscene.checkCaption(template+'_taunt_01'):
			banter = 1
			m25getset.SetObjectVar(char,'banter_enabled',1,nosave=1)
			#print 'Banter enabled for ' + char
		else:
			banter = 0
			m25getset.SetObjectVar(char,'banter_enabled',0,nosave=1)
	if banter:
		target = Target_GetCurrentTarget(char)
		if target != None and (m25distance.GetDistance(char,target) < 300):
			saying = FindBanter(char,'taunt')
			if saying != None:
				cs = ['CS_SpeakTo('+`char`+ ',' + `target` + ',' + `saying` + ')']
				if randint(1,2) == 1:
					response = FindBanter(target,'retaunt')
					if response != None:
						cs.append('CS_SpeakTo('+`target`+ ',' + `char` + ',' + `response` + ')')
						if randint(1,2) == 1:
							third = FindBanter(char,'reretaunt')
							if third != None:
								cs.append('CS_SpeakTo('+`char`+ ',' + `target` + ',' + `third` + ')')
								if randint(1,2) == 1:
									fourth = FindBanter(char,'rereretaunt')
									if fourth != None:
										cs.append('CS_SpeakTo('+`target`+ ',' + `char` + ',' + `fourth` + ')')
				cs.append('CS_Wait(2)')
				m25cutscene.PlayCutscene(cs)

def FindBanter(char, tag='taunt'):
	total = m25getset.GetObjectVar(char,'banter_'+tag+'_total',default=None,nosave=1,integer=1)
	template = m25template.GetComplexTemplate(char)
	if total == None:
		done = 0
		num = 1
		while not done:
			if num < 10: numstring = '0'+`num`
			else: numstring = `num`
			if checkCaption(template+'_'+tag+'_'+numstring):
				num = num + 1
			else:
				done = 1
		m25getset.SetObjectVar(char,'banter_'+tag+'_total',num-1,nosave=1)
		total = num-1
	#find a random char_tag line from captions.txt
	if total > 0:
		num = randint(1,total)
		if num < 10: numstring = '0' + `num`
		else: numstring = `num`
#		target = Target_GetCurrentTarget(char)
# 		if target != None:
# 			targettemplate = m25template.GetComplexTemplate(target)
# 			potential = template+'_'+tag+'_'+numstring+'_'+targettemplate
# 			if checkCaption(potential):
# 				return potential
#		return template+'_'+tag+'_'+numstring
		return tag + '_' + numstring
	return None

		
#####################################################################################
#
# Brawler
# 
# Description: have standard-ai (non custom ai) characters fight each other 
# according to their team designation
#
#####################################################################################



#-------------------------------------------------------------------------------------
#
# Brawler
#
#-------------------------------------------------------------------------------------

def SetupBrawl():
	#SetTimer(BRAWL_TIMER, 'ff.BrawlEvent')
	js.Event_CancelSinks('BrawlEvent')
	js.Event_RegisterSink(EVENT_TIMER, 'BrawlEvent', '', '', 1, 0, 0)
	Report(BRAWL_CHANNEL,'SetupBrawl() - Brawl starting up')
	
def BrawlEvent(event):
	#timeS = ff.time()
	if GetMissionOver():
		print 'mission over'
		return
	Report(BRAWL_CHANNEL,'BrawlEvent() - Checking for new targets')
	all = Mission_GetDynamicObjects()
	for c in all:
		if (not IsAIAuto(c)) and CanBrawl(c) and (not IsAIDisabled(c)):
			if ffx.FFX_HasPortrait(c):
				if AIInWatch(c):
					DoBrawl(c)
			else:
				DoBrawl(c)
	js.Event_RegisterSink(EVENT_TIMER, 'BrawlEvent', '', '', BRAWL_TIMER, 0, 0)
	#timeE = ff.time()
	#print 'Brawl event took %f seconds'%(timeE - timeS)

def DoBrawl(char):
	#if not ff.MapInfo.has_key(char): return
	maxdist = m25distance.DISTANCE_VISIBLE
	if InSkirmishMode():
		maxdist = m25distance.DISTANCE_WHOLEMAP
	old_target = m25getset.GetObjectVar(char,BRAWL_OLDTARGET,nosave=1)
	new_target = Target_GetCurrentTarget(char)
	if old_target != new_target:
		m25report.Report('BRAWL', '%s changes targets from %s to %s '%(char,`old_target`,`new_target`))
	if new_target == None:
		Report(BRAWL_CHANNEL, 'DoBrawl() - ' + char + ' cannot find a target (maxdist=' + `maxdist` + ').')
		return
	
	if m25distance.GetDistance(char, new_target) > maxdist:
		Report(BRAWL_CHANNEL, 'DoBrawl() - ' + char + ' target too far away to brawl.')
		return
 	if old_target != None:
 		if not (js.Object_Exists(old_target) and js.Object_IsAlive(old_target)):
			old_target = None
		
	if new_target == old_target:
		#if (js.Object_GetClass(char)&OC_CONTROLLABLE):
		if Timestamp_ObjectAttrOutdated(char, 'brawl_same_target', delta=30):
			Report(BRAWL_CHANNEL,'DoBrawl() - controllable ' + char + ' targeting ' + `new_target`)
			cshelper.addKillGoal(char,new_target,ai.goal.PRI_LOW,replace=1)
			Timestamp_ObjectAttrSync(char, 'brawl_same_target')
		return
	else:
		Report(BRAWL_CHANNEL,'DoBrawl() - ' + char + ' targeting ' + `new_target`)
		SetObjectVar(char, BRAWL_OLDTARGET, new_target,nosave=1)
		cshelper.addKillGoal(char,new_target,ai.goal.PRI_LOW,replace=1)
		Timestamp_ObjectAttrSync(char, 'brawl_same_target')

def SetSkirmishModeOn():
	js.Mission_SetAttr(BRAWL_SKIRMISH,1)
	#SetMissionVar(BRAWL_SKIRMISH,1)
	
def SetSkirmishModeOff():
	js.Mission_SetAttr(BRAWL_SKIRMISH,0)
	#SetMissionVar(BRAWL_SKIRMISH,0)
	
def InSkirmishMode():
	return js.Mission_TestAttr(BRAWL_SKIRMISH)
	#return GetMissionVar(BRAWL_SKIRMISH,default=0,integer=1) == 1

#-------------------------------------------------------------------------------------
#
# Brawler Status
#
#-------------------------------------------------------------------------------------	
			
Global_RegisterAttr('ai_brawler',0)

def SetBrawler(object, brawl=OBJECT_BRAWLER):
	if type(object) == types.ListType:
		for obj in object:
			js.Object_SetAttr(obj,'ai_brawler',brawl)
			#SetObjectVar(obj, 'brawler', brawl)
	else:
		js.Object_SetAttr(object,'ai_brawler',brawl)
		#SetObjectVar(object, 'brawler', brawl)
	
def GetBrawler(object):
	return js.Object_GetAttr(object, 'ai_brawler')
	#return GetObjectVar(object, 'brawler', integer=1)
	
def IsBrawler(object):
	return js.Object_GetAttr(object, 'ai_brawler') > OBJECT_NOT_BRAWLER
	#return GetObjectVar(object,'brawler',default=0, integer=1) > OBJECT_NOT_BRAWLER
		
def BrawlerDisabled(object):
	state = Object_GetPrimaryState(object)
	if state in [js.PCSTATE_STUNNED,js.PCSTATE_STATIC,js.PCSTATE_BLANK,js.PCSTATE_PANICKED]:
		return 1
	return 0
	
def CanBrawl(object):
	if Object_IsAlive(object):
		if IsBrawler(object) and (not IsAIDisabled(object)):
			if not(BrawlerDisabled(object)):
				return 1
	return 0
	
	
#####################################################################################
#
# Clone Check
# 
# Description: periodically check for clones from non-custom AIs and set them to the proper team
#
#####################################################################################

def CloneCheck(event):
	allchars = js.Mission_GetDynamicObjects()
	for char in allchars:
		actual_clones = js.Object_GetAttr(char,'cloneCount')
		current = actual_clones
		known_clones = GetObjectVar(char,'_lastcloneCount',default=0,integer=1)
		if (actual_clones > known_clones):
			for c in allchars:
				if (c[:5] == 'mobj_') and (js.Object_GetTemplate(c) == js.Object_GetTemplate(char)) and (not GetObjectVar(c,'_clonefound',default=0,integer=1)) and (known_clones < actual_clones):
					Report(AI_CHANNEL,'CloneCheck(): ' + char + ' has clone ' + c)
					SetObjectVar(c,'_clonefound',1)
					AISetClone(c)
					SetTeam(c,GetTeam(char))
					if ffx.usingM25AI():
						SetupOneChar(c)
						SetBrawler(c)
					known_clones = known_clones + 1
		# check decoys as well
		if (char[:5] == 'mobj_') and (js.Object_GetTemplate(char) in ['standard_image_projection','black_jack_image']) and (js.Object_GetClass(char) == 167779556) and (not AIIsDecoy(char)):
			Report(AI_CHANNEL,'CloneCheck(): found decoy ' + char)
			SetObjectVar(char,'_decoyfound',1)
			AISetDecoy(char)
			SetTeam(char,TEAM_GOODGUYS)
		#check for fire elementals
		if char[:5] == 'mobj_' and (js.Object_GetTemplate(char) == 'fire_elemental') and (not m25getset.GetObjectVar(char,'_livingfiresfound',default=0)):
			#find the red sun who spawned them
			for redsun in allchars:
				if m25template.GetComplexTemplate(redsun) == 'red_sun':
					m25getset.SetObjectVar(char,'_livingfiresfound',1)
					SetTeam(char, GetTeam(redsun))
					if ffx.usingM25AI():
						SetupOneChar(char)
						SetBrawler(char)
		#check for polymorph victims
		#doesn't work - all characters show up as polymorphed!
# 		if (char[:5] == 'mobj_') and (js.Object_GetPrimaryState(char) & js.PCSTATE_POLYMORPHED) and (GetTeam(char) != -1):
# 			SetTeam(char,-1)
		if current != actual_clones:
			SetObjectVar(char,'_lastcloneCount',actual_clones)
	m25event.SetTimer(CLONE_CHECK_TIMER,'CloneCheck')

#####################################################################################
#
# Dynamic values
# 
# Description: values that can change based on current conditions
#	currently used for priorities and percents
#
#####################################################################################

def DetermineDynamicTacticValue(char, tactic_id, target, value_name, value):
	if IsFunction(value):
		newvalue = 0
		current_value = GetObjectVar(char, 'tactic_'+ value_name + `tactic_id`,default=0,nosave=1)
		tac_string = 'newvalue = ' + GetLeftString(value,'(') + '(' + `char` + ',' + `tactic_id` + ',' + `target` + ',' + `current_value` + ',' + GetRightString(value,'(')
		exec tac_string
		return newvalue
	else:
		return value
				
def DOscillating(char, tactic_id, target, current, val1, val2, pct):
	if current==0:
		current=val1
	if randint(0,100) <= pct:
		if current == val1:
			return val2
		return val1
	return current
	
def DRandom(char, tactic_id, target, current, val1, val2, pct):
	if randint(0,100) <= pct:
		return randint(val1, val2) 
	return current               

def DChangeOverTime(char, tactic_id, target, current, val1, val2, pct, delta):
	if current == 0:
		current=val1
	if randint(0,100) <= pct:
		newval = current + delta
		if val1 > val2:
			if newval < val2:
				newval = val2
		else:
			if newval > val2:
				newval = val2
		return newval
	return current

def DHealthAndTime(char, tactic_id, target, current, val1, val2, pct, delta):
	if current == 0:
		current=val1
	newval = current
	healthval = (100 * (1-js.Object_GetAttr(char,'health')/js.Object_GetAttr(char,'maxHealth')))
	if healthval > newval:
		newval = healthval
	if randint(0,100) <= pct:
		newval = newval + delta
	if val1 > val2:
		if newval < val2:
			newval = val2
	else:
		if newval > val2:
			newval = val2
	return newval
	
def DHealthWatch(char, tactic_id, target, current, low=20, high=100):
	if current == 0:
		current=low
	newval = current
	healthval = (1.0-js.Object_GetAttr(char,'health')/js.Object_GetAttr(char,'maxHealth'))
	newval = low + (high-low)*healthval
	return newval	
	
def DEnergyWatch(char, tactic_id, target, current, low=20, high=100):
	if current == 0:
		current=low
	newval = current
	energyval = (1-js.Object_GetAttr(char,'energyPoints')/100.0)
	newval = low + (high-low)*energyval
	return newval	
	
def DNumberOfEnemies(char, tactic_id, target, current, radius=m25distance.DISTANCE_AREA_LARGE, subtype='none', low=50, high=100):
	if current == 0:
		current = low
	newval = current
	enemylist = FindAllOfType(target, reaction=TEAM_ALLIES, maxdist=radius, subtype=subtype, default=[])
	numenemies = len(enemylist)
	enemyval = (numenemies/3.0)  #warning: magic number here
	newval = low + (high-low)*enemyval
	return newval

def DSubtype(char, tactic_id, target, current, subtypelow='none', subtypehigh='none', low=20, normal=70, high=90):
	if IsSubtype(target, subtypelow, char):
		return low
	if IsSubtype(target, subtypehigh, char):
		return high
	return normal

#####################################################################################
#
# Goals
# 
# Description: tactics that can be applied and removed dynamically
#
#####################################################################################

# AI Goals: management functions

ai_compiled_goals = {}

def AIGetGoals(char, compiled=0):
	global ai_compiled_goals
	if not compiled:
		return GetObjectVar(char,'ai_goals',default=None)
	else:
		if ai_compiled_goals.has_key(char):
			return ai_compiled_goals[char]
		else:
			textgoals = GetObjectVar(char, 'ai_goals', default=None)
			if textgoals != None:
				AICompileGoals(char, textgoals)
				return ai_compiled_goals[char]	
	
# 		goals = GetObjectVar(char,'ai_compiled_goals',default=None,nosave=1)
# 		if goals == None:
# 			textgoals = GetObjectVar(char, 'ai_goals', default=None)
# 			if textgoals != None:
# 				AICompileGoals(char, textgoals)
# 				return GetObjectVar(char, 'ai_compiled_goals',default=None,nosave=1)
# 			else:
# 				return None
# 		return goals
	
def AISetGoals(char, goals):
	#print 'AISetGoals ' + char + ' ' + `goals`
	SetObjectVar(char,'ai_goals',goals)
	AICompileGoals(char, goals)

def AICompileGoals(char, goals):
	global ai_compiled_goals
	#create compiled goals
	compile_string = ''
	num = 0
	if goals in [None, '', [],{}]:
		ai_compiled_goals[char] = None
		#SetObjectVar(char, 'ai_compiled_goals', None, nosave=1)
		return
	for goal in goals.keys():
		g = goals[goal]
		compile_string = compile_string + 'ff.' + GetLeftString(g,'(') + '(' + `char` + ',' + GetRightString(g,'(') + '\n'
		num = num + 1
	compiled_goals = compile(compile_string, '<string>', 'exec')
	ai_compiled_goals[char] = compiled_goals
	#SetObjectVar(char, 'ai_compiled_goals',compiled_goals,nosave=1)		
		
def AIAddGoal(goal_id, char, goal):
	Report(AI_CHANNEL_GOAL,'Adding goal ' + goal + ' to character ' + char)
	goals = AIGetGoals(char)
	if goals == None:
		goals = {goal_id:goal}
		AISetGoals(char, goals)
	else:
		goals[goal_id]=goal
		AISetGoals(char, goals)

def AIUnableToDoGoals(char):
	if not js.Object_Exists(char):
		return 1
 	if cshelper.hypnotized(char):
 		return 1
	return 0		

def AIDoGoals(char):
	if AIUnableToDoGoals(char):
		return
	goals = AIGetGoals(char, compiled=1)
	if goals != None:
		exec goals
			
def AIHasGoal(char, partialgoal):
	goals = AIGetGoals(char)
	if goals != None:
		for g in goals.keys():
			if PartialGoalMatch(goals[g],partialgoal):
				return g
	return 0
	
def PartialGoalMatch(goal, partialgoal):
	num = 0
	for p in partialgoal:
		if p != goal[num]:
			return 0
		num = num + 1
	return 1
			
def AIRemoveGoalByID(goal_id,char,goal=''):
	goals = AIGetGoals(char)
	if goals != None:
		if goals.has_key(goal_id):
			del goals[goal_id]
			AISetGoals(char,goals)
			Report(AI_CHANNEL_GOAL,'Removing goal ' + goal + ' from character ' + char)
	return

def AIRemoveGoal(char, goal=''):
	goal_id = AIHasGoal(char ,goal)
	if goal_id:
		AIRemoveGoalByID(goal_id,char,'')
	
def AIGetNextGoalID():
	gid = GetMissionVar('__goal_id_count__',default=1000,integer=1)
	SetMissionVar('__goal_id_count__',gid+1)
	return gid	
	
def GoalEnactReturnFunction(char,fn):
	if (fn == None) or (fn==''):
		return
	if GetRightString(fn,'(') == ')':
		fn_string = 'ff.' + GetLeftString(fn,'(') + '(' + `char` + GetRightString(fn,'(')
	else:
		fn_string = 'ff.' + GetLeftString(fn,'(') + '(' + `char` + ',' + GetRightString(fn,'(')
	try:
		#print 'ff.' + GetLeftString(fn,'(') + '(' + `char` + ',' + GetRightString(fn,'(')
		exec fn_string
	except:
		Report('Error in GoalEnactReturnFunction(): '+fn_string)
		print sys.exc_type, sys.exc_value
		raise	
	
#----------------------------------------
#      
# Actual AI goals
#
#----------------------------------------
	
# attack	

def AddGoal_Attack(char,target=None,target_type=TARGET_ENEMY,target_priority=TARGET_PRIORITY_MEDIUM):
	if target==None:
		m25report.Report(AI_CHANNEL_GOAL, 'Cannot add attack goal for %s with target None'%(char))
		return
	if type(target) == types.StringType:
		target = [target]
	gid = AIGetNextGoalID()
	AIAddGoal(gid,char,'Goal_Attack('+ `gid` + ',' + `target` + ',' + `target_type` + ',' + `target_priority`+')')
    
def Goal_Attack(char, goal_id, target=None, target_type=TARGET_ENEMY, target_priority=TARGET_PRIORITY_MEDIUM):
	if target == None:
		return
	newtarget = []
	if target in [m25team.TEAM_ENEMIES, m25team.TEAM_ALLIES, m25team.TEAM_NEUTRAL]:
		newtarget = FindTarget(char,target, nocache=1)
		if newtarget == None:
			return
	else:
		if not type(target) == types.ListType:
			target = [target]
		done = 1
		for t in target:
			if t in [m25team.TEAM_ENEMIES,m25team.TEAM_ALLIES,m25team.TEAM_NEUTRAL]:
				newt = Target_GetCurrentTarget(char,t)
				if newt != None: 
					newtarget.append(newt)
					done = 0
			else:
				if js.Object_Exists(t): done = 0
				newtarget.append(t)
		if done:
			AIRemoveGoalByID(goal_id,char,'Goal_Attack(' + `goal_id`)
			return
		#newtarget = target
	priority = DetermineDynamicTacticValue(char, goal_id, target, target_priority, target_priority)
	Target_AddTarget(char,newtarget,target_type,priority)
	m25report.Report(GOAL_CHANNEL, 'Goal_attack: ' + `char` + ' targeting ' + `target` + ' with priority ' + `priority`)

	
# call for help

def AddGoal_CallForHelp(char, pct=100, priority=TACTIC_PRIORITY_HIGH, situation='none'):
	gid = AIGetNextGoalID()
	AIAddGoal(gid,char, 'Goal_CallForHelp(' + `gid` + ',' + `pct` + ',' + `priority` + ',' + `situation` + ')')
	
def Goal_CallForHelp(char, goal_id, pct, priority, situation):
	#only call for help once every 30 seconds
	if not Timestamp_ObjectAttrOutdated(char, 'ai_last_help_call', delta=30): 
		return
	situation = AddToList(situation,'lowerhealthpct 30')
	if AITacticCheck(goal_id,char,energy=0,pct=pct,priority=priority,role='none',situation=situation,target=None):
		groupname = GetCharGroup(char)
		if groupname != None:
			group = GetGroup(groupname)
			if group != None:
				yelp = 0
				for g in group:
					if m25distance.GetDistance(char,g) > 200:
						AddGoal_MoveToTarget(g,char,mindist=60)
						yelp = 1
				m25report.Report(GOAL_CHANNEL, 'Goal_CallForHelp: initiated for ' + `char`)
				if yelp:
					cshelper.speakNoCB(char,'M25AI_HELP')
				Timestamp_ObjectAttrSync(char, 'ai_last_help_call')
	
		
#quickly add temporary abilities				
						
def AddGoal_DynamicAbility(char, temp_attribute=None, tactic=''):
	gid = AIGetNextGoalID()
	tactic_call = m25aiopt.Optimize_CreateTacticCall(m25aiopt.Opt_CalcParameters(char,tactic))
	m25ai.AIAddGoal(gid, char,'Goal_DynamicAbility(%i, "%s", %s)'%(gid,temp_attribute,`tactic_call`))
	return gid

def Goal_DynamicAbility(char, goal_id, temp_attribute, tactic_call):
	#check to see if I still have this power
	if temp_attribute and not ffx.FFX_ObjectGetAttr(char,temp_attribute,0):
		m25ai.AIRemoveGoalByID(goal_id,char,'Goal_DynamicAbility')
		print 'temp_attribute1'
		return 0
	else:
		event = mlogreader.PseudoAnimationEvent(objectStr=char)
		tactic_id = goal_id
		return eval(tactic_call)				



# guard
	
def AddGoal_Guard(char,target,priority=TACTIC_PRIORITY_LOW-1):
	gid = AIGetNextGoalID()
	AIAddGoal(gid,char,'Goal_Guard(' + `gid` +  ',' + `target` + ',' + `priority` + ')')
	
def Goal_Guard(char, goal_id, target, priority):
	if GetAIPriority(char) >= priority: return 0
	if not Object_Exists(target): 
		AIRemoveGoalByID(goal_id,char,'Goal_Guard')
		return
	if not Object_IsAlive(target): 
		AIRemoveGoalByID(goal_id,char,'Goal_Guard')
		return
	enemytarget = Target_GetCurrentTarget(char, TARGET_ENEMY)
	#enemytarget = FindTarget(char,maxdist='map',default=None)
	if ((enemytarget != None) and Object_Exists(enemytarget)):
		if Target_GetDistance(char) < m25distance.DISTANCE_VISIBLE: return
	groupname = GetCharGroup(char)
	if groupname != None:
		group = GetGroup(groupname)
		if group != None:
			total = len(group)
			num = GetElementNum(group, char)
			miAngle = skXAngles.fSkXGetAngleOfSegment (total,num)
			target = skXAngles.fSkXGetPointFromVector (js.Get_ObjectPos(target), miAngle, 75)
	time = AI_STANDARD_MOVE_TIME + (randint(1,10) / 10.0)
	AIDoMove(char,None,target,priority,time,description='goal guard')	

def GetElementNum(tlist, telement):
	num = 0
	for t in tlist:
		if t == telement:
			return num
		num = num + 1
	return 0		
			
# hunt
def AddGoal_HuntEnemy(char,priority=TACTIC_PRIORITY_LOW):
	gid = AIGetNextGoalID()
	AIAddGoal(gid,char,'Goal_HuntEnemy(' + `gid` +  ',' + `priority` + ')')
	
def Goal_HuntEnemy(char, goal_id, priority):
	if not AITacticCheck(goal_id,char,energy=0,pct=100,priority=priority,role='none',situation='none',target=None):
		return
	time = AI_DEFAULTTIMER + (randint(1,10) / 10.0)
	if IsAIAuto(char):
		time = AI_STANDARD_MOVE_TIME + (randint(1,10) / 10.0)
	target = Target_GetCurrentTarget(char)
	if target == None:
		target = FindTarget(char, TARGET_ENEMY, 0, m25distance.DISTANCE_WHOLEMAP, force=1)
	AIGenericMoveOpt(goal_id,char,'',0,0, pct=100, time=time, mindist=m25distance.DISTANCE_LONG, maxdist=m25distance.DISTANCE_WHOLEMAP, type=TARGET_ENEMY, subtype=None, situation=None, role=None, priority=priority, desc='hunting')



# move out of the way
def AddGoal_MoveAway(char,target,priority=TACTIC_PRIORITY_HIGH, dist=m25distance.DISTANCE_MEDIUM, fn=''):
	gid = AIGetNextGoalID()
	AIAddGoal(gid,char,'Goal_MoveAway(' + `gid` +  ',' + `target` + ',' + `priority` + ',' + `dist` + ',' + `fn` + ')')
	
def Goal_MoveAway(char, goal_id, target, priority,dist, fn):
	if not AITacticCheck(goal_id,char,energy=0,pct=100,priority=priority,role='none',situation='none',target=None):
		return
	if ((target != None) and Object_Exists(target)):
		if (m25distance.GetDistance(char,target) > m25distance.DistanceConvert(dist)):
			AIRemoveGoalByID(goal_id,char,'Goal_MoveAway(' + `goal_id`)
			GoalEnactReturnFunction(char,fn)
		else:
			object = FindFleePoint(char,target,mindist=dist+50,maxdist=dist+100,default=None)
			AIDoMove(char,None,object,priority,AI_STANDARD_MOVE_TIME,description='goal move away from')	
	
			
def AddGoal_MoveToBuildings(char,priority=TACTIC_PRIORITY_LOW, dist=m25distance.DISTANCE_LONG, fn=''):
	gid = AIGetNextGoalID()
	AIAddGoal(gid,char,'Goal_MoveToBuildings(' + `gid` +  ',' + `priority` + ',' + `dist` + ',' + `fn` + ')')
	
def Goal_MoveToBuildings(char, goal_id, priority,dist, fn):
	#print 'GOAL MOVE TO BUILDINGS'
	if not AITacticCheck(goal_id,char,energy=0,pct=100,priority=priority,role='none',situation='none',target=None):
		#print 'TACTIC CHECK FAILED?'
		return
	buildings = AI_GetNearBuildings(char,10000)
	if buildings == []: 
		#print 'NO BUILDINGS!'
		return
	dist = m25distance.DISTANCE_WHOLEMAP
	building = buildings[0]
	for b in buildings:
		bdist = m25distance.GetDistance(char,b)
		if bdist < dist:
			dist = bdist
			building = b
	if ((building != None) and Object_Exists(building)):
		#if (m25distance.GetDistance(char,building) > m25distance.DistanceConvert(dist)):
		AIDoMove(char,None,building,priority,AI_STANDARD_MOVE_TIME,description='goal move to buildings')			
		#else:
		#	print 'TOO CLOSE?'
	else:
		#print 'BUILDING DOES NOT EXIST?'
		pass
			
# move to enemy	
	
def AddGoal_MoveToEnemy(char,priority=TACTIC_PRIORITY_LOW, dist=m25distance.DISTANCE_MELEE, fn=''):
	gid = AIGetNextGoalID()
	AIAddGoal(gid,char,'Goal_MoveToEnemy(' + `gid` +  ',' + `priority` + ',' + `dist` + ',' + `fn` + ')')
	
def Goal_MoveToEnemy(char, goal_id, priority,dist, fn):
	if not AITacticCheck(goal_id,char,energy=0,pct=100,priority=priority,role='none',situation='none',target=None):
		return
	enemytarget = Target_GetCurrentTarget(char, TARGET_ENEMY)
	if ((enemytarget != None) and Object_Exists(enemytarget)):
		if (m25distance.GetDistance(char,enemytarget) < m25distance.DistanceConvert(dist)):
			AIRemoveGoalByID(goal_id,char,'Goal_MoveToEnemy(' + `goal_id`)
			GoalEnactReturnFunction(char,fn)
		else:
			AIDoMove(char,None,enemytarget,priority,AI_STANDARD_MOVE_TIME,description='goal move to enemy')			
	
# move while in one of the various gaseous forms
# - try to move towards the current enemy	
	
def AddGoal_MoveGaseous(char,priority=TACTIC_PRIORITY_HIGH, dist=m25distance.DISTANCE_MELEE, fn=''):
	gid = AIGetNextGoalID()
	AIAddGoal(gid,char,'Goal_MoveGaseous(' + `gid` +  ',' + `priority` + ',' + `dist` + ',' + `fn` + ')')
	
def Goal_MoveGaseous(char, goal_id, priority,dist, fn):
	enemytarget = FindTacticTarget(goal_id, char, mindist=50, maxdist=300, energy=0, pct=100, priority=priority, role='none', situation='none', reaction=TARGET_ENEMY, target_trait='not nearenemy 50')
	if ((enemytarget != None) and Object_Exists(enemytarget)):
		if not ST_displaced(char):
			AIRemoveGoalByID(goal_id,char,'Goal_MoveGaseous(' + `goal_id`)
			GoalEnactReturnFunction(char,fn)
		else:
			AIDoMove(char,None,enemytarget,priority,AI_STANDARD_MOVE_TIME,description='goal move gaseous')	

		
# move to hero	
	
def AddGoal_MoveToHero(char,priority=TACTIC_PRIORITY_LOW,followdist=m25distance.DISTANCE_MEDIUM):
	gid = AIGetNextGoalID()
	AIAddGoal(gid,char,'Goal_MoveToHero(' + `gid` +  ',' + `priority` + ',' + `followdist` + ')')
	
def Goal_MoveToHero(char, goal_id, priority, followdist):
	if not AITacticCheck(goal_id,char,energy=0,pct=100,priority=priority,role='none',situation='none',target=None):
		return
	hero = GetClosestHero(char)
	if hero != None and m25distance.GetDistance(char,hero) > m25distance.DistanceConvert(followdist):
		AIDoMove(char,None,hero,priority,AI_STANDARD_MOVE_TIME,description='goal move to hero')
	#if not IsAIAuto(char):
	#	cshelper.addMoveGoal(char,hero,ai.goal.PRI_HI)
		#cshelper.moveTo(char,hero)
	
def GetClosestHero(char):
	heroes = ffx.GetAllHeroes()
	dist = 1000000
	found = None
	for h in heroes:	
		newdist = m25distance.GetDistance(char,h)
		if  newdist < dist:
			found = h
			dist = newdist
	return found
		
	
# move to marker

def AddGoal_MoveToMarker(char,marker='',mindist=50,priority=TACTIC_PRIORITY_MEDIUM,pct=80,patrol=0,fn=None, remove=1):
	gid = AIGetNextGoalID()
	AIAddGoal(gid,char,'Goal_MoveToMarker('+`gid` + ',' + `marker` + ',' + `mindist` + ',' + `priority` + ',' + `pct` + ',' + `patrol` + ',' + `fn`+ ',' + `remove` + ')')
            
def Goal_MoveToMarker(char, goal_id, marker='', mindist=50, priority=TACTIC_PRIORITY_LOW, pct=80, patrol=0, fn=None, remove=1):
	dist = GetDistanceToMarker(char,marker)
	if dist <= mindist and remove:
		AIRemoveGoalByID(goal_id,char,'Goal_MoveToMarker(' + `goal_id`)
		GoalEnactReturnFunction(char,fn)
	else:
		if not AITacticCheck(goal_id,char,energy=0,pct=pct,priority=priority,role='none',situation='none',target=None):
			return
		if patrol:
			enemytarget = Target_GetCurrentTarget(char, TARGET_ENEMY)
			if ((enemytarget != None) and Object_Exists(enemytarget)):
				if m25distance.GetDistance(char,enemytarget) >= m25distance.DistanceConvert(char,'visible'):
					return
		time = AI_DEFAULTTIMER
		if IsAIAuto(char):
			time = AI_STANDARD_MOVE_TIME
		else:
			#cshelper.addMoveGoal(char,marker)
			cshelper.moveTo(char,marker)
		AISupport_DoMove(char, marker, 'move', 0, js.Get_MarkerPos(marker), js.Get_ObjectPos(char), priority, time, desc='goal move to marker')
		#AIDoMove(char,None,marker,priority,time,description='goal move to marker')


# move to a specific target (object, character, etc.)
				
def AddGoal_MoveToTarget(char,target='',mindist=50,priority=TACTIC_PRIORITY_MEDIUM-1,pct=100,fn=None):
	gid = AIGetNextGoalID()
	AIAddGoal(gid,char,'Goal_MoveToTarget('+`gid` + ',' + `target` + ',' + `mindist` + ',' + `priority` + ',' + `pct` + ','  + `fn`+')')
		
def Goal_MoveToTarget(char, goal_id, target='', mindist=50, priority=TACTIC_PRIORITY_LOW, pct=100, fn=None):
	if type(target) == types.ListType:
		target = skXMapInfo.MapInfo_GetNearest(char, taObjects=target)
	dist = m25distance.GetDistance(char,target)
	if dist <= mindist:
		AIRemoveGoalByID(goal_id,char,'Goal_MoveToTarget(' + `goal_id`)
		GoalEnactReturnFunction(char,fn)
	if not AITacticCheck(goal_id,char,energy=0,pct=pct,priority=priority,role='none',situation='none',target=None):
		return
	if (not js.Object_Exists(target)) or (m25distance.GetDistance(char,target) <= mindist):
		AIRemoveGoalByID(goal_id,char,'Goal_MoveToTarget(' + `goal_id`)
		GoalEnactReturnFunction(char,fn)
		return
	time = AI_DEFAULTTIMER
	if IsAIAuto(char):
		time = AI_STANDARD_MOVE_TIME
	AIDoMove(char,None,target,priority,time,description='goal move to target')
	
	
# retrieve item

def AddGoal_RetrieveItem(char,item='',mindist=50,priority=TACTIC_PRIORITY_MEDIUM,pct=80,fn=None):
	gid = AIGetNextGoalID()
	AIAddGoal(gid,char,'Goal_RetrieveItem('+`gid` + ',' + `item` + ',' + `mindist` + ',' + `priority` + ',' + `pct` + ',' + `fn`+')')
            
def Goal_RetrieveItem(char, goal_id, item='', mindist=50, priority=TACTIC_PRIORITY_LOW, pct=20, fn=None):
	if not (Object_Exists(item) and Object_IsAlive(item)):
		AIRemoveGoalByID(goal_id,char,'Goal_RetrieveItem(' + `goal_id`)
	if not AITacticCheck(goal_id,char,energy=0,pct=pct,priority=priority,role='none',situation='none',target=None):
		return
	dist = m25distance.GetDistance(char,item)
	if dist <= mindist:
		AIRemoveGoalByID(goal_id,char,'Goal_RetrieveItem(' + `goal_id`)
		GoalEnactReturnFunction(char,fn)
	else:
		time = AI_DEFAULTTIMER
		if IsAIAuto(char):
			time = AI_STANDARD_MOVE_TIME
		AIDoMove(char,None,item,priority,time,description='goal retrieve item')		
		

# stay with group (used by group leader)
				
def AddGoal_StayWithTeam(char, priority=TACTIC_PRIORITY_LOW):
	gid = AIGetNextGoalID()
	AIAddGoal(gid,char,'Goal_StayWithTeam('+`gid` + ',' + `priority` + ')')
	
def Goal_StayWithTeam(char, goal_id, priority=TACTIC_PRIORITY_MEDIUM):
	if not AITacticCheck(goal_id,char,energy=0,pct=100,priority=priority,role='none',situation='none',target=None):
		return
	enemytarget = Target_GetCurrentTarget(char, TARGET_ENEMY)
	#enemytarget = FindTarget(char,maxdist='map',default=None)
	if ((enemytarget != None) and Object_Exists(enemytarget)):
		if (m25distance.GetDistance(char,enemytarget) < m25distance.DISTANCE_VISIBLE):
			return
	furthest = GetFurthestGroupMember(char)
	if furthest != None:
		if m25distance.GetDistance(char,furthest) > m25distance.DISTANCE_VISIBLE:
			time = AI_DEFAULTTIMER
			if IsAIAuto(char):
				time = AI_STANDARD_MOVE_TIME + 1
			AIDoMove(char,None,furthest,priority,time,mindist=m25distance.GetDistance(char,furthest)/2,description='leader move back to group')

	
def GetFurthestGroupMember(char):
	#groupname = GetCharGroup(char)
	#if groupname == None: return None
	group = CharGetGroup(char)
	if group == None: return None
	dist = 0
	furthest = None
	for g in group:
		if Object_Exists(g) and Object_IsAlive(g) and (g != char):
			gdist = m25distance.GetDistance(char,g)
			if dist < gdist:
				dist = gdist
				furthest = g
	return furthest
	
	
# surround
	
def AddGoal_SurroundTarget(char,priority=TACTIC_PRIORITY_MEDIUM):
	gid = AIGetNextGoalID()
	AIAddGoal(gid,char,'Goal_SurroundTarget(' + `gid` +  ',' + `priority` + ')')
	
def Goal_SurroundTarget(char, goal_id, priority):
	return 
	target = Target_GetCurrentTarget(char, TARGET_ENEMY)
	if (target == None) or (not Object_Exists(target)):
		return
	if not AITacticCheck(goal_id,char,energy=0,pct=50,priority=priority,role='none',situation='none',target=target):
		return
	if m25distance.GetDistance(char,target) > m25distance.DISTANCE_MELEE:
		total = GetNumTargettedBy(target,TARGET_ENEMY)
		num = GetTargettedByPosition(char,TARGET_ENEMY,target)
		dist = m25distance.DISTANCE_SHORT
		miAngle = skXAngles.fSkXGetAngleOfSegment (total,num)
		target = skXAngles.fSkXGetPointFromVector (js.Get_ObjectPos(target), miAngle, dist)
		AIDoMove(char,None,target,priority,AI_STANDARD_MOVE_TIME+1,description='goal surround')


#throw	
	
def AddGoal_Throw(char,object,mode,time,priority):
	if Object_Exists(object):
		gid = AIGetNextGoalID()
		AIAddGoal(gid,char,'Goal_Throw(' + `gid` + ',' + `object` + ',' + `mode` + ',' + `time` + ',' + `priority` + ')')

def Goal_Throw(char,goal_id,object,mode,time,priority):
	if GetAIPriority(char) >= priority: return 0
	if not js.Object_Exists(object):
		AIRemoveGoalByID(goal_id,char, 'Goal_Throw')
		RemoveRole(char,'throwing')
		return
	dist = m25distance.GetDistance(char,object)
	target = Target_GetCurrentTarget(char)
	if target != None:
		if not AIIsFlying(target):
			tdist = Target_GetDistance(char)
			if (tdist <= m25distance.DISTANCE_MELEE) or (ST_nearenemy(target,param=80)):
				AIRemoveGoalByID(goal_id,char, 'Goal_Throw')
				RemoveRole(object,'being_thrown')
				RemoveRole(char,'throwing')
				return
	if dist <= 50:
		ThrowPickup(char,object)
		AIRemoveGoalByID(goal_id,char, 'Goal_Throw')
	else:
		if dist <= m25distance.DISTANCE_MEDIUM:
			if AIIsFlying(char):
				AIDoMove(char,mode='land',priority=priority,time=0.5,description='land for throwing')
				return
		AIDoMove(char,mode,object,priority,time,0,'move to throwable')
		

#####################################################################################
#
# Groups
# 
# Description: collections of characters that can all be given orders at the same time
#
#####################################################################################		
			

# getting group values directly for a character

def CharGetGroupSize(char):
	group = CharGetGroup(char)
	if group == None: return None
	return len(group)
	
def CharGetGroupPosition(char):
	group = CharGetGroup(char)
	if group == None: return None
	num = 0
	for g in group:
		if g == char:
			return num
		num = num + 1
	return None
	
def CharGetGroup(char):
#	groupname = GetObjectVar(char,'_groupname',default=None)
	groupname = GetCharGroup(char)
	if groupname == None: return None
	return GetGroup(groupname)	
	
def CharGetGroupName(char):
	return GetCharGroup(char)
	#return GetObjectVar(char,'_groupname',default=None)
	
def CharGetGroupLeader(char):
	groupname = CharGetGroupName(char)
	if groupname != None:
		return Group_GetLeader(groupname)
	return None
	
	
# group management	
	
_ai_groups = {}

def AI_LoadGroupData():
	global _ai_groups
	if _ai_groups == {}:
		_ai_groups = m25getset.GetMissionVar('_ai_groups',default={})

def AI_SaveGroupData():
	global _ai_groups
	#clean out outdated entries to speed up the save
	for char in _ai_groups.keys():
		if not js.Object_Exists(char):
			del _ai_groups[char]
	m25getset.SetMissionVar('_ai_groups', _ai_groups)


def GroupExists(groupname):
	return GetGroup(groupname) != []
	
def GetGroup(groupname):
	group = GetMissionVar(groupname+'_group',default = None,nosave=1)
	if group == None:
		group = []
		#go through objects and recreate the group index
		objects = js.Mission_GetDynamicObjects()
		for obj in objects:
			chargroupname = GetCharGroup(obj)
			if chargroupname == groupname:
				group.append(obj)
		SetGroup(groupname, group)
	return group
	
def SetGroup(groupname, charlist):
	m25report.Report(GROUP_CHANNEL,`groupname` + ' set to ' + `charlist`)
	SetMissionVar(groupname+'_group',charlist,nosave=1)

def GetCharGroup(char):
	global _ai_groups
	AI_LoadGroupData()
	if _ai_groups.has_key(char): return _ai_groups[char]
	return None
	
def SetCharGroup(char,groupname,save=1):
	global _ai_groups
	_ai_groups[char] = groupname
	if save:
		AI_SaveGroupData()
	#SetObjectVar(char,'_groupname',groupname)
	
def AddToGroup(groupname, char, reg=1, save=1):
	if char == None:
		m25report.Report(GROUP_CHANNEL,`char` + ' cannot be added to ' + `groupname`)
		return
	group = GetGroup(groupname)
	# set up the group timer if we are starting from an empty group
# 	if (group == []) and (GetGroupVar(groupname, '_timerset',default=0,integer=1) == 0):
# 		m25event.SetTimer(5, 'GroupTimer', groupname)
# 		SetGroupVar(groupname, '_timerset', 1)
	if type(char) == types.ListType:
		for c in char:
			if not c in group:
				group.append(c)
				if GROUP_CHANNEL: m25report.Report(GROUP_CHANNEL,`c` + ' added to group ' + `groupname`)
				SetCharGroup(c, groupname, save)
				cshelper.regDeath(c,'GroupMemberKO')
				mlogreader.regObjectDestroy(c, 'm25ai.GroupMemberDestroyed',string=groupname)
	else:
		if not char in group:
			group.append(char)
			SetCharGroup(char,groupname, save)
			m25report.Report(GROUP_CHANNEL,`char` + ' added to group ' + `groupname`)
			cshelper.regDeath(char,'GroupMemberKO')
			mlogreader.regObjectDestroy(char, 'm25ai.GroupMemberDestroyed',string=groupname)
	SetGroup(groupname,group)
	


# GroupMemberDestroyed is called when a character in a group is destroyed.
# It is meant to preserve the integrity of groups if an object is destroyed before it is KO'd.  
# This function does NOT trigger any KO events as object destruction before KO only occurs in scripts
# It is up to the developer to handle other effects of destruction.
def GroupMemberDestroyed(event):
	char = event.object
	groupname = event.string
	group = GetGroup(groupname)
	if char in group:
		RemoveFromGroup(groupname,char)
		if Group_GetLeader(groupname) == char: Group_PromoteLeader(groupname)	


def RemoveFromGroup(groupname, char):
	group = GetGroup(groupname)
	if char in group:
		group.remove(char)
		SetCharGroup(char,None)
		m25report.Report(GROUP_CHANNEL,`char` + ' removed from group ' + `groupname`)
	SetGroup(groupname,group)
	
def SwitchGroup(groupname, char):
	old_group = CharGetGroupName(char)
	if old_group != None:
		RemoveFromGroup(old_group, char)
		AddToGroup(groupname, char, reg=0)
	else:
		AddToGroup(groupname, char, reg=1)
	
	
# group variables/state

def GetGroupVar(groupname, var, default=None, integer=0, number=0):
	return GetMissionVar('group_' + groupname + '_' + var, default=default, integer=integer, number=number)
	
def SetGroupVar(groupname, var, value):
	SetMissionVar('group_' + groupname + '_' + var, value)
	

# group event registration

def GroupAddReg(groupname, function_name, var):
	glist = GetGroupVar(groupname, var, default=[])
	glist.append(function_name)
	SetGroupVar(groupname, var, glist)		
	
def RegGroupLeaderKO(groupname,function_name):
	GroupAddReg(groupname,function_name,'koleader')
	
def RegGroupMemberKO(groupname,function_name):
	GroupAddReg(groupname,function_name,'komember')
	
def RegGroupPromoteLeader(groupname, function_name):
	GroupAddReg(groupname,function_name,'promote')
	
def RegGroupLeaderSet(groupname, function_name):
	GroupAddReg(groupname,function_name,'setleader')	
	
def RegGroupTimer(groupname, function_name):
	GroupAddReg(groupname,function_name,'timer')
	
def GetGroupLeaderKO(groupname):
	return GetGroupVar(groupname, 'koleader', default=[])
	
def GetGroupMemberKO(groupname):
	return GetGroupVar(groupname, 'komember', default=[])
	
def GetGroupPromoteLeader(groupname):
	return GetGroupVar(groupname, 'promote', default=[])
	
def GetGroupLeaderSet(groupname):
	return GetGroupVar(groupname, 'setleader', default=[])	
	
def GetGroupTimer(groupname):
	return GetGroupVar(groupname, 'timer', default=[])
	
	
#group events

def GroupEventCallback(groupname, func_name, values=[]):
	if groupname != None:
		glist = []
		exec 'glist = ff.Get' + func_name + '(' + `groupname` +  ')'
		for g in glist:
			try:
				exec 'ff.' + g + '(' + `groupname` + ',' + `values` + ')'
			except:
				m25report.Report(ERROR_CHANNEL, func_name + ' problem with registered function ' + str(g))
				print sys.exc_type, sys.exc_value
				raise

def GroupLeaderKO(event):
	groupname = GetCharGroup(event.object)
	if groupname != None:
		m25report.Report(GROUP_CHANNEL, `groupname` + ' leader koed ')
		GroupEventCallback(groupname,'GroupLeaderKO',values=event.object)
		Group_PromoteLeader(groupname)

# Note:  this will trigger mutliple times if you add a character to multiple groups
def GroupMemberKO(event):
	groupname = GetCharGroup(event.object)
	if groupname != None:
		m25report.Report(GROUP_CHANNEL, `groupname` + ' member koed ')
		GroupEventCallback(groupname,'GroupMemberKO',values=event.object)

def GroupTimer(event):
	groupname = event.object
	if groupname != None and (not m25missionover.GetMissionOver()):
		m25report.Report(GROUP_CHANNEL, `groupname` + ' timer ')
		all = GetGroup(groupname)
		alive = 0
		for a in all:
			if js.Object_Exists(a) and js.Object_IsAlive(a):
				alive = 1
		if alive:	
			GroupEventCallback(groupname,'GroupTimer')
			m25event.SetTimer(GetGroupTimeInterval(groupname), 'GroupTimer', groupname)
	else:
		# turn off the timer - this group is finished
		SetGroupVar(groupname, '_timerset', 0)
		
# settings

def GroupAutoGuard(groupname):
	return GetGroupVar(groupname,'_autoguard',default=1,integer=1)
	
def SetGroupAutoGuard(groupname,on=1):
	SetGroupVar(groupname,'_autoguard',on)
	
def GroupStayWithTeam(groupname):
	return GetGroupVar(groupname,'_staywithteam',default=1,integer=1)	
	
def SetGroupStayWithTeam(groupname,on=1):
	SetGroupVar(groupname,'_staywithteam',on)
	
def SetGroupTimeInterval(groupname,interval):
	SetGroupVar(groupname,'interval',interval)
	
def GetGroupTimeInterval(groupname):
	return GetGroupVar(groupname,'interval',default=5,number=1)
				
# leadership	
	
def Group_SetLeader(groupname, char):
	if char == None:
		m25report.Report(GROUP_CHANNEL,`char` + ' cannot be made leader of ' + `groupname`)
	group = GetGroup(groupname)
	if not char in group:
		AddToGroup(groupname,char)
	m25report.Report(GROUP_CHANNEL,`char` + ' made leader of group ' + `groupname`)
	if GroupAutoGuard(groupname):
		GroupCommand_Guard(groupname, char)
	if GroupStayWithTeam(groupname):
		AddGoal_StayWithTeam(char)
	cshelper.regDeath(char,'GroupLeaderKO')
	SetMissionVar(groupname+'_leader', char)
	GroupEventCallback(groupname,'GroupLeaderSet', char)
	
def Group_GetLeader(groupname):
	leader = GetMissionVar(groupname+'_leader', default=None)
	if leader == None:
		leader = Group_PromoteLeader(groupname)
	elif not (Object_Exists(leader) and Object_IsAlive(leader)):
		leader = Group_PromoteLeader(groupname)
	else:
		pass
	return leader
	
def Group_PromoteLeader(groupname):
	group = GetGroup(groupname)
	if group == []:
		m25report.Report(GROUP_CHANNEL, 'empty ' + `groupname` + ' cannot have a leader promoted ')
		return None
	#SetMissionVar(groupname+'_leader', None)
	GroupEventCallback(groupname,'GroupPromoteLeader')
	#leader may have been set by callbacks, so don't overwrite it if it was
	leader = GetMissionVar(groupname+'_leader', default=None)
	if leader and js.Object_Exists(leader) and js.Object_IsAlive(leader): return leader
	leader = None
	if leader == None:
		for g in group:
			if (Object_Exists(g) and Object_IsAlive(g)):
				if leader == None:
					leader = g
				# controllable character automatically become the leader
				if ffx.FFX_HasPortrait(g):
					leader = g
		if leader != None:
			m25report.Report(GROUP_CHANNEL, `leader` + ' automatically promoted to leader of group ' + `groupname`)
			Group_SetLeader(groupname, leader)
	return leader
	
	
	
# group commands	
	
def GroupCommand_AllowCallsForHelp(groupname):
	#disabled for now
	return
	group = GetGroup(groupname)
	if group != None:
		for g in group:
			if not AIHasGoal(g,'Goal_CallForHelp'):
				AddGoal_CallForHelp(g)

def GroupCommand_Attack(groupname, target):
	leader = Group_GetLeader(groupname)
	if leader:
		AddGoal_Attack(leader, target)				
				
def GroupCommand_FocusFire(groupname, target):
	group = GetGroup(groupname)
	if group != None:
		for g in group:
			AddGoal_Attack(g, target)
				
def GroupCommand_Guard(groupname, char):
	group = GetGroup(groupname)
	for g in group:
		if (Object_Exists(g) and Object_IsAlive(g)):
			goal_id = AIHasGoal(g,'Goal_Guard(')
			if goal_id:
				AIRemoveGoalByID(goal_id,g,'Goal_Guard(')
			if (g != char):	AddGoal_Guard(g,char)
	m25report.Report(GROUP_CHANNEL, `groupname` + ' set to guard ' + `char`)
	
def GroupCommand_GuardLeader(groupname, char):
	leader = Group_GetLeader(groupname)
	if leader == None:
		return 
	GroupCommand_Guard(groupname, leader)		
			
def GroupCommand_HuntEnemies(groupname):
	leader = Group_GetLeader(groupname)
	if leader == None:
		return 
	if not AIHasGoal(leader,'Goal_HuntEnemy('):
		AddGoal_HuntEnemy(leader)	
	#GroupCommand_Guard(groupname, leader)
	m25report.Report(GROUP_CHANNEL, `groupname` + ' set to hunt enemies ')
	
def GroupCommand_HuntHeroes(groupname):
	leader = Group_GetLeader(groupname)
	if leader == None: return
	if not AIHasGoal(leader, 'Goal_MoveToHero('):
		AddGoal_MoveToHero(leader)
		m25report.Report(GROUP_CHANNEL, `groupname` + ' set to hunt heroes ')

def GroupCommand_MoveToTarget(groupname, target, mindist=50, pct=100,priority=TACTIC_PRIORITY_MEDIUM-1, function=''):
	leader = Group_GetLeader(groupname)
	if leader == None: return 
	#AIRemoveGoal(leader,'Goal_MoveToTarget('):
	AddGoal_MoveToTarget(leader,target=target,mindist=mindist,priority=priority,pct=pct,fn=function)
	m25report.Report(GROUP_CHANNEL, `groupname` + ' set to move to target ' + `target`)		
		
def GroupCommand_MoveToMarker(groupname, marker, pct=100,priority=TACTIC_PRIORITY_MEDIUM-1, patrol=0, function=''):
	leader = Group_GetLeader(groupname)
	if leader == None: return 
	if not AIHasGoal(leader,'Goal_MoveToMarker('):
		AddGoal_MoveToMarker(leader,marker=marker,priority=priority,pct=pct,patrol=patrol,fn=function)
	m25report.Report(GROUP_CHANNEL, `groupname` + ' set to move to marker ' + `marker`)
	
def GroupCommand_RetrieveItem(groupname, item, priority=TACTIC_PRIORITY_MEDIUM, pct=80, function=''):
	leader = Group_GetLeader(groupname)
	if leader == None: return 
	if not AIHasGoal(leader,'Goal_RetrieveItem('):
		AddGoal_RetrieveItem(leader,item=item,priority=priority,pct=pct,fn=function)
	m25report.Report(GROUP_CHANNEL, `groupname` + ' set to retrieve item ' + item)	
	
def GroupCommand_SurroundEnemies(groupname):
	group = GetGroup(groupname)
	if group != None:
		for g in group:
			AddGoal_SurroundTarget(g)	
			
			
	
#####################################################################################
#
# Roles
# 
# Description: dynamic properties that can be applied to an object (used in subtype checks)
#
#####################################################################################
def GetRoles(char):
	return m25getset.GetObjectVar(char, 'ai_roles', default = [])
	
def SetRoles(char,roles):
	SetObjectVar(char, 'ai_roles', roles)	
	
def AddRole(char,role):	
	roles = GetRoles(char)
	roles.append(role)
	SetRoles(char,roles)
	
def ClearRoles(char):
	SetRoles(char,[])	
	
def RemoveRole(char,role):
	roles = GetRoles(char)
	num = 0
	for r in roles:
		if r == role:
			del roles[num]	
			SetRoles(char,roles)
			return
		num = num + 1
	
def HasRole(char, role):
	if IsList(role):
		for r in role:
			if not HasOneRole(char, r):
				return 0
		return 1
	else:
		return HasOneRole(char, role)

def HasOneRole(char, role):
	if role in ['none',None]: return 1
	charroles = GetRoles(char)
	return role in charroles	
	
#####################################################################################
#
# Subtypes
# 
# Description: find characters and objects with certain properties
#
#####################################################################################

#-------------------------------------------------------------------------------------
#
# Find Nearest
#
#-------------------------------------------------------------------------------------

def FindNearestOfType(object, reaction=TEAM_ENEMIES, maxdist=m25distance.DISTANCE_VISIBLE, subtype=[], default=None, targetlist=[]):
	current = Timestamp_GetObjectCachedAttr(object, `subtype`, delta=TARGET_CACHED_TIMER, default=None)
	if current != None:
		return current
	maxd = DistanceConvert(maxdist,object)
	tlist = targetlist
	if targetlist == []:
		tlist = Mission_GetDynamicObjects()
	current = default
	curdist = m25distance.DISTANCE_OUTOFRANGE
	for t in tlist:
		if FoundObjectValid(object,t, reaction) and IsSubtype(t,subtype,object):
			dist = GetDistance(object,t)
			if dist <= maxd:
				if dist < curdist:
					current = t
					curdist = dist
	Timestamp_SetObjectCachedAttr(object,`subtype`,current)
	return current
		
def FindAllOfType(object, reaction=TEAM_ENEMIES, mindist=0, maxdist=m25distance.DISTANCE_VISIBLE, subtype=[], default=None, targetlist=[], speedcheck=0):
	current = Timestamp_GetObjectCachedAttr(object, `subtype`+'alltype', delta=TARGET_CACHED_TIMER, default=None)
	if current != None:
		return current
	mind = DistanceConvert(mindist,object)
	maxd = DistanceConvert(maxdist,object)
	tlist = targetlist
	if targetlist == []:
		tlist = Mission_GetDynamicObjects()
	current = []
	for t in tlist:
		if speedcheck and (randint(1,5) == 1):
			continue
		if FoundObjectValid(object,t,reaction) and IsSubtype(t,subtype,object):
			dist = GetDistance(object,t) 
			if (dist >= mind) and (dist <= maxd):
				current.append(t)
	if current == []:
		return default
	Timestamp_SetObjectCachedAttr(object,`subtype`+'alltype',current)
	return current
	
def FindOneOfType(object, reaction=TEAM_ENEMIES, maxdist=m25distance.DISTANCE_VISIBLE, subtype=[], default=None, targetlist=[]):
	global valid_characters
	current = Timestamp_GetObjectCachedAttr(object, `subtype`, delta=TARGET_CACHED_TIMER, default=None)
	if current != None:
		return current
	maxd = DistanceConvert(maxdist,object)
	tlist = targetlist
	if targetlist == []:
		tlist = skXMapInfo.MapInfo_GetObjects(object,maxdist)
	for t in tlist:
		if FoundObjectValid(object,t, reaction) and IsSubtype(t,subtype,object):
			dist = m25distance.GetDistance(object,t)
			if dist <= maxd:
				Timestamp_SetObjectCachedAttr(object,`subtype`,t)
				return t
	return default
	

def FoundObjectValid(object,t,reaction):
	return (t!=object) and Object_Exists(t) \
		and Object_IsAlive(t) and TeamsReactionOfType(object, t, reaction) and \
		not (ST_decoy(t) and TeamsReactionOfType(object, t, TEAM_ALLIES)) 
	
#-------------------------------------------------------------------------------------
#
# Find Specific Objects (various functions)
#
#-------------------------------------------------------------------------------------

def FindNearestObjectOfType(char,reaction=TEAM_NEUTRAL,maxdist=m25distance.DISTANCE_WHOLEMAP,subtype=[],default=[]):
	current = Timestamp_GetObjectCachedAttr(char, `subtype`+'obj', delta=TARGET_CACHED_TIMER, default=None)
	if current != None:
		return current
	if 'throwable' in subtype:
		objects = M25GetThrowableObjects()
	else:
		objects = M25GetObjects()
	obj = FindNearestOfType(char,reaction,maxdist=maxdist,subtype=subtype,targetlist=objects,default=None)
	if obj == None:
		return default
	Timestamp_SetObjectCachedAttr(char, `subtype`+'obj', obj)
	return obj

def FindClosestThrowable(char,maxdist=m25distance.DISTANCE_VISIBLE,default=None):
	objects = skXMapInfo.MapInfo_GetObjects(js.Get_ObjectPos(char), 'nearest', tfFilter=ff.ThrowFilter)
	print objects
	return objects
	

	
def FindRandObjectInRange(char,reaction=TEAM_ANY,mindist=0, maxdist=m25distance.DISTANCE_WHOLEMAP, subtype='none',default=[]):
	current = Timestamp_GetObjectCachedAttr(char, `subtype`+'randobj', delta=TARGET_CACHED_TIMER, default=None)
	if current != None:
		return current
	objects = js.Mission_GetObjects()
	tries = 20
	done = 0
	num_objects = len(objects)-1
	mind,maxd = m25distance.DistanceConvert(mindist,char),m25distance.DistanceConvert(maxdist,char)
	while (not done) and (tries > 0):
		obj = objects[randint(0,num_objects)]
		if FoundObjectValid(char,obj,reaction) and IsSubtype(obj,subtype,char):
			dist = GetDistance(obj,char)
			if (dist >= mind) and (dist <= maxd):
				Timestamp_SetObjectCachedAttr(char, `subtype`+'randobj', obj)
				return obj
		tries = tries -1
	return None
	
def FindAllObjectsInRange(char,reaction=TEAM_NEUTRAL,mindist=0, maxdist=m25distance.DISTANCE_WHOLEMAP, subtype='none',default=[]):
	current = Timestamp_GetObjectCachedAttr(char, `subtype`+'allobj', delta=TARGET_CACHED_TIMER, default=None)
	if current != None:
		return current
	objects = M25GetObjects()
	objlist = FindAllOfType(char,reaction,mindist,maxdist,subtype,targetlist=objects,default=None)
	if objlist == None:
		return default
	Timestamp_SetObjectCachedAttr(char, `subtype`+'allobj', objlist)
	return objlist
		
def FindFleeObject(char,target,reaction=TEAM_NEUTRAL,mindist=0,maxdist=m25distance.DISTANCE_WHOLEMAP,subtype='none',default=[]):
	current = Timestamp_GetObjectCachedAttr(char, `subtype`+'fleeobj', delta=TARGET_CACHED_TIMER, default=None)
	if current != None:
		return current
	newsubtype=AddToList(subtype,'not building')
	objects = FindAllObjectsInRange(char,reaction,mindist,maxdist,newsubtype,default=None)
	if objects == None:
		return default
	mind = DistanceConvert(mindist,char)
	maxd = DistanceConvert(maxdist,char)
	objlist = []
	for obj in objects:
		if GetDistance(char,obj) < GetDistance(target,obj):
			objlist.append(obj)
	if objlist == []:
		return default
	else:
		obj = objlist[randint(0,len(objlist)-1)]
		Timestamp_SetObjectCachedAttr(char, `subtype`+'fleeobj', obj)
		return obj
			
		
#warning: maxdist not currently used		
def FindFleePoint(char,target,mindist=0,maxdist=m25distance.DISTANCE_LONG,default=None):
	pos = js.Get_ObjectPos(char)
	tpos = js.Get_ObjectPos(target)
	maxdist = m25distance.DistanceConvert(maxdist)
	found = 0
	bounds = FindMapBounds()
	minx, miny, maxx, maxy = bounds[0], bounds[1], bounds[2], bounds[3]
	maxdist = mindist + 50
	num = 0
	tries = 0
	while not found and tries < 50:
		tries = tries + 1
		newpos = (pos[0]+randint(-maxdist,maxdist),pos[1]+randint(-maxdist,maxdist),pos[2])
		v = js.Vector_Sub(pos, newpos)
		dist = math.sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2])
		vt = js.Vector_Sub(tpos,newpos)
		target_dist = math.sqrt(vt[0]*vt[0] + vt[1]*vt[1] + vt[2]*vt[2])
		if (newpos[0] >= minx) and (newpos[0] <= maxx) and (newpos[1] >= miny) and (newpos[1] <= maxy) and \
			dist >= mindist and \
			target_dist > dist:
			found = 1
		num = num + 1
		if num >= 50:
			newpos = pos
			found = 1
	return newpos
		
def FindRandomPoint(char,mindist=0,maxdist=m25distance.DISTANCE_LONG,default=None):
	pos = js.Get_ObjectPos(char)
	maxdist = m25distance.DistanceConvert(maxdist)
	found = 0
	bounds = FindMapBounds()
	minx, miny, maxx, maxy = bounds[0], bounds[1], bounds[2], bounds[3]
	tries = 0
	while not found and (tries < 50):
		tries = tries + 1
		newpos = (pos[0]+randint(-maxdist,maxdist),pos[1]+randint(-maxdist,maxdist),pos[2])
		v = js.Vector_Sub(pos, newpos)
		dist = math.sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2])
		if (newpos[0] >= minx) and (newpos[0] <= maxx) and (newpos[1] >= miny) and (newpos[1] <= maxy) and \
			dist >= mindist:
			found = 1
	return newpos	
	
def FindMapBounds():
	if _aicache.has_key('_mapbounds'):
		return _aicache['_mapbounds']
	objects = js.Mission_GetObjects()
	minx, miny, maxx, maxy = 10000000, 10000000, -10000000, -10000000
	for obj in objects:
		pos = js.Get_ObjectPos(obj)
		if pos[0] < minx: minx = pos[0]
		if pos[0] > maxx: maxx = pos[0]
		if pos[1] < miny: miny = pos[1]
		if pos[1] > maxy: maxy = pos[1]
	_aicache['_mapbounds'] = [minx, miny, maxx, maxy]
	return [minx, miny, maxx, maxy]
		
#-------------------------------------------------------------------------------------
#
# Condition Evaluators
# subtype = a condition that applies to the current target
# situation = a condition that applies to the user of the tactic
#
#-------------------------------------------------------------------------------------

subtypecompdict = {}

def IsSubtype(char,subtype,target=None,debug=0):
	if subtype in [[],'none',None]: return 1
	co = None
	if type(subtype) == types.IntType:
		co = m25aiopt.Opt_GetSubtypeCode(subtype)
	if co == None:	
		global subtypecompdict
		co = ''
		if subtypecompdict.has_key(`subtype`):
			co = subtypecompdict[`subtype`]
		else:
			cd = GenerateSubtypeCode(char, subtype, target)
			#print cd
			co = compile(cd, '<string>', 'exec')
			subtypecompdict[`subtype`] = co
		
	param,param2 = None,None		
	exec co
	return codeval

# 	if debug:
# 		print subtype
# 	if type(subtype) == types.ListType:
# 		for t in subtype:
# 			if not TrySubtype(char,t,target,debug):
# 				Timestamp_SetObjectCachedAttr(char, `subtype` + '_cache', 0)
# 				return 0
# 		Timestamp_SetObjectCachedAttr(char, `subtype` + '_cache', 1)
# 		return 1
# 	val = TrySubtype(char,subtype,target,debug)
# 	#Timestamp_SetObjectCachedAttr(char, `subtype` + '_cache', val)
# 	return val	
	
def InSituation(char,situation,target=None,debug=0):
	return IsSubtype(char,situation,target,debug)	
    
def GetNumSubtypes(char,subtype,target=None):
	num = 0
	if type(subtype) == types.ListType:
		for s in subtype:
			if IsSubtype(char,s,target):
				num = num + 1
	else:
		num = IsSubtype(char,subtype,target)
	return num


def TabDepth(depth=0):
	if depth <= 0: return ''
	st = ''
	for i in range(0,depth):
		st = st + '\t'
	return st	
	
def GenerateSubtypeCode(char, subtype, target=None, depth=0):
	global stcodedict
	code = ''
	n = 1
	if type(subtype) == types.ListType:	
		for t in subtype:
			code = code + GenOneSubtypeCode(char, t, target, depth)
			if n != len(subtype):
				code = code + TabDepth(depth) + 'if codeval == 1: \n'
			n = n + 1
			depth = depth + 1
	else:
		code = GenOneSubtypeCode(char, subtype, target, depth)
	return code
		
def GenOneSubtypeCode(char, subtype, target, depth=0):
	global stcodedict
	code = ''
	trylist = SplitAroundOr(subtype)
	if trylist[0] == 1:
		code = code + GenSingleSubtype(char, trylist[1], target, depth)
		code = code + TabDepth(depth) + 'if codeval == 0:\n'
		code = code + GenOneSubtypeCode(char, trylist[2], target, depth+1)
		return code
	else:
		return GenSingleSubtype(char, subtype, target, depth)
		
def GenSingleSubtype(char, subtype, target, depth=0):
	global stcodedict
	code = ''
	notval = 0
	st = subtype
	if GetLeftString(subtype,' ') == 'not': 
		notval = 1
		st = GetRightString(subtype, ' ')
	stype = GetLeftString(st,' ')
	p = GetRightString(st, ' ')
	param = GetLeftString(p, ' ')
	try:
		if int(param) != 0: param = int(param)
	except: 
		pass
	param2 = GetRightString(p, ' ')
	try: 
		if int(param2) != 0: param2 = int(param)
	except: 
		pass
	if stcodedict.has_key('ST_' + stype):
		if param != '':
			code = code + TabDepth(depth) + 'param = ' + `param` + '\n'
		if param2 != '':
			code = code + TabDepth(depth) + 'param2 = ' + `param2` + '\n'
		for line in stcodedict['ST_' + stype]:
			code = code + TabDepth(depth) + line
	else:
		code = TabDepth(depth) + 'codeval = TryOneSubtype(char, ' + `st` + ', target)\n'
	if notval:
		code = code + TabDepth(depth) + 'codeval = not codeval\n'
	return code
	
	
		
def TrySubtype(char,subtype,target,debug=0):
    if subtype == ' ':
        return 0
    trylist = SplitAroundOr(subtype)
    if trylist[0] == 1:
        if TryOneSubtype(char,trylist[1],target,debug):
            return 1
        else:
            return TrySubtype(char,trylist[2],target,debug)
    else:
        return TryOneSubtype(char,subtype,target,debug)

def AddToAICache(char,subtype,val):
	_aicache[char+'_'+subtype] = val        
        
def TryOneSubtype(char,subtype,target,debug=0):
	if debug:
		print subtype
	if subtype == 'flying': return AIIsFlying(char)
	val = Timestamp_GetObjectCachedAttr(char,`subtype` + '_cache',delta=AI_CACHE,integer=1)
	if val != None:
		return val
	else:
		val = TryOneSubtype2(char,subtype,target,debug)
		Timestamp_SetObjectCachedAttr(char, `subtype` + '_cache',val)
		return val
	
stcompdict = {}		
			
def TryOneSubtype2(char,subtype,target,debug=0):
	if GetLeftString(subtype,' ') == 'not':
		return not TrySubtype(char,GetRightString(subtype,' '),target)
	if _aicache.has_key(char+'_'+subtype):
		return _aicache[char+'_'+subtype]
	if IsFunction(subtype):
		exec 'val = ff.' + GetLeftString(subtype,'(') + '(' + `char` + ',' + `target` + ',' + GetRightString(subtype,'(')
		if debug:
			m25report.Report(AI_DEBUG_CHANNEL,char + ' ' + `subtype` + ' ' + `val`)
		return val
	if Object_GetTemplate(char) == subtype:
		if debug:
			m25report.Report(AI_DEBUG_CHANNEL,char + ' ' + `subtype` + ' 1')
		return 1
	val = 0
	val = TryUserSubtype(char, subtype)
	if val:
		if debug:
			m25report.Report(AI_DEBUG_CHANNEL,char + ' ' + `subtype` + ' 1')
		return 1
	val = HasRole(char, subtype)
	if val:
		if debug:
			m25report.Report(AI_DEBUG_CHANNEL,char + ' ' + `subtype` + ' 1')
		return 1
	param = GetRightString(subtype,' ',default=None)
	param2 = GetRightString(param,' ',default=None)
	name = GetLeftString(subtype, ' ')
	try:
		subtype_exec = 'ival = ff.ST_' + GetLeftString(subtype,' ') + '(' + `char` + ',target=' + `target` + ', param=' + `param` + ', param2=' + `param2` + ')'
		#global stcompdict
		#co = None
		#if subtype_exec in stcompdict.keys():
		#	co = stcompdict[subtype_exec]
		#else:
		#	co = compile(subtype_exec+'\n','err.log','exec')
		#	stcompdict[subtype_exec] = co
		#exec co
		exec subtype_exec
		#print 'VAL ' + `ival` + ' : ' + subtype_exec
		#if ival == 1:
		#	print 'VAL did equal 1'
		if debug:
			m25report.Report(AI_DEBUG_CHANNEL,char + ' ' + `subtype` + ' ' + `val`)
		return ival
	except:
		if sys.exc_type != AttributeError:
			#investigate how to report sys. properly!
			Report('Error in TryOneSubtype(): ' + char + ' is ' + subtype + ' : ' )
			print sys.exc_type, sys.exc_value
			if ALLOW_ERRORS:
				raise
			return 0
	if debug:
		m25report.Report(AI_DEBUG_CHANNEL,char + ' ' + `subtype` + ' 0')
	return 0


def AI_TemplateHasUserSubtype(char, subtype):
	subtypelist = GetAISubtypes(char, default=None, istemplate=1)
	return (subtypelist and (subtype in subtypelist))
	
def TryUserSubtype(char, subtype):
    types = GetAISubtypes(char,default=None)
    if types == None:
        return 0
    if subtype in types:
        return 1
    return 0        
               

#-------------------------------------------------------------------------------------
#
# Actual Conditions
#
#-------------------------------------------------------------------------------------    

# special default subtype always returns true
def ST_none(char,target=None,param=None,param2=None):
	return 1

# for use with FFX only
def ST_ffxattr(char,target=None,param=None,param2=None):
	try:
		val = ff.hasAttribute(char,param)
		AddToAICache(char,param,val)
		return ff.hasAttribute(char,param)
	except:
		return 0

def ST_in_database(char,target=None,param=None,param2=None):
	return ffx.FFX_ObjectGetAttr(char,'addedToDB')

def ST_air(char,target=None,param=None,param2=None):
	return (js.Object_GetAttr(char,'material') == 6) or (TryUserSubtype(char,'air'))

def ST_acid_burnt(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_ACID_BURNT  

def ST_acrobatic(char,target=None,param=None,param2=None):
	val = ffx.hasAttribute(char,'acrobatic')
	AddToAICache(char,'acrobatic',val)
	return val
	
def ST_alive(char,target=None,param=None,param2=None):
	if Object_Exists(char):
		return Object_IsAlive(char)
	else:
		return 0
	
def ST_alone(char,target=None,param=None,param2=None):
	return not ST_nearally(char, target, param, param2)

def ST_area_energized(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_AREA_ENERGIZED

def ST_area_acid(char,target=None,param=None,param2=None):
	val = HasPowerType(char, 'acid', 'area')
	#AddToAICache(char,'area_acid',val)
	return val
	
def ST_area_cold(char,target=None,param=None,param2=None):
	val = HasPowerType(char, 'cold', 'area')	
	#AddToAICache(char,'area_cold',val)
	return val

def ST_area_crushing(char,target=None,param=None,param2=None):
	val = HasPowerType(char, 'crushing', 'area')	
	#AddToAICache(char,'area_crushing',val)
	return val	
	
def ST_area_energy(char,target=None,param=None,param2=None):
	val = HasPowerType(char, 'energy', 'area')	
	#AddToAICache(char,'area_energy',val)
	return val	
	
def ST_area_electrical(char,target=None,param=None,param2=None):
	val = HasPowerType(char, 'electrical', 'area')	
	#AddToAICache(char,'area_electrical',val)
	return val

def ST_area_fire(char,target=None,param=None,param2=None):
	val = HasPowerType(char, 'fire', 'area')	
	#AddToAICache(char,'area_fire',val)
	return val	
	
def ST_area_mental(char,target=None,param=None,param2=None):
	val= HasPowerType(char, 'mental', 'area')	
	#AddToAICache(char,'area_mental',val)
	return val
		
def ST_area_mystical(char,target=None,param=None,param2=None):
	val= HasPowerType(char, 'mystical', 'area')	
	#AddToAICache(char,'area_mystical',val)
	return val
	
def ST_area_piercing(char,target=None,param=None,param2=None):
	val= HasPowerType(char, 'piercing', 'area')	
	#AddToAICache(char,'area_piercing',val)
	return val
	
def ST_area_radiation(char,target=None,param=None,param2=None):
	val = HasPowerType(char, 'radiation', 'area')	
	#AddToAICache(char,'area_radiation',val)
	return val
		
def ST_asleep(char,target=None,param=None,param2=None):
	return ffx.FFX_ObjectGetAttr(char,'ffx_sleep')

def ST_blanked(char,target=None,param=None,param2=None):
	return js.Object_GetPrimaryState(char) == js.PCSTATE_BLANK

def ST_blind(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_BLIND
	
def ST_buoyant(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_BUOYANT

def ST_building(char,target=None,param=None,param2=None):
	try:
		val= js.Object_GetClass(char) & js.OC_BUILDING	
	except:
		val = 0
	#AddToAICache(char,'building',val)
	return val

def ST_caged(char,target=None,param=None,param2=None):
	return 0	
	
def ST_char(char,target=None,param=None, param2=None):
	if not js.Object_GetAttr(char, 'physical'): 
		return 0
	else:
		c = js.Object_GetClass(char)
		val= (c & ( js.OC_CHARACTER |  js.OC_CONTROLLABLE |  js.OC_POLICE |  js.OC_CIVILIAN |  js.OC_MINION |  js.OC_VILLIAN))	
		return val
		
def ST_chargedup(char,target=None,param=None,param2=None):
	return HasRole(char,'charged_up')

def ST_civilian(char,target=None,param=None,param2=None):
	val= js.Object_GetClass(char) & js.OC_CIVILIAN
	#AddToAICache(char,'civilian',val)
	return val

def ST_cloaked(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_INVISIBLE	
	
def ST_clone(char,target=None,param=None,param2=None):
	val= AIIsClone(char)
	#AddToAICache(char,'clone',val)
	return val
	
def ST_cloth(char,target=None,param=None,param2=None):
	return (js.Object_GetAttr(char,'material') == 10) or (TryUserSubtype(char,'cloth'))


def ST_clustered(char,target=None,param=None,param2=None):
	return ST_swarmed(char,TEAM_ALLIES)

def ST_concrete(char,target=None,param=None,param2=None):
	return (js.Object_GetAttr(char,'material') == 8) or (TryUserSubtype(char,'concrete'))

def ST_dead(char,target=None,param=None,param2=None):
	if Object_Exists(char):
		return not Object_IsAlive(char)
	return 0

def ST_decoy(char,target=None,param=None,param2=None):
	val= (AIIsDecoy(char) or (js.Object_GetTemplate(char) == 'standard_image_projection'))
	return val

def ST_deep_sensored(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_DEEP_SENSORED
	
def ST_density_maxed(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_DENSITY  

def ST_density_increased(char,target=None,param=None,param2=None):
	return ffx.FFX_ObjectGetAttr(char,'densityFXOn')

# def ST_disabled(char,target=None,param=None,param2=None):
# 	if js.Object_GetPrimaryState(char) in [js.PCSTATE_BLANK,js.PCSTATE_BLIND,js.PCSTATE_DIMENSION_SHIFT,js.PCSTATE_PANIC,js.PCSTATE_STUNNED,js.PCSTATE_STATIS]:
# 		return 0
	
def ST_direct_acid(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'acid', 'direct')
	
def ST_direct_cold(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'cold', 'direct')	

def ST_direct_crushing(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'crushing', 'direct')	
	
def ST_direct_energy(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'energy', 'direct')	
	
def ST_direct_electrical(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'electrical', 'direct')

def ST_direct_fire(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'fire', 'direct')	
	
def ST_direct_mental(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'mental', 'direct')	
		
def ST_direct_mystical(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'mystical', 'direct')	
	
def ST_direct_piercing(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'piercing', 'direct')	
	
def ST_direct_radiation(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'radiation', 'direct')	
	
def ST_displaced(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_DISPLACE_IMAGE

def ST_doppelganger(char,target=None,param=None,param2=None):
	return js.Object_GetPrimaryState(char) == js.PCSTATE_DOPPELGANGER

def ST_electrical(char,target=None,param=None,param2=None):
	return ffx.isElectricalGrid(char)
	
def ST_electronic(char,target=None,param=None,param2=None):
	return ffx.isPhasingTemplate(char)

def ST_emitting_damage(char,target=None,param=None,param2=None):
	if (js.Object_GetSecondaryStates(char) & (js.SCSTATE_IRRADIATED | js.SCSTATE_AREA_ENERGIZED)) or HasRole(char,ROLE_AREA_ATTACK) or (ffx.FFX_ObjectGetAttr(char,'eTangled',0)==1):
		return 1
	else:
		return 0
	
def ST_empathy(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_EMPATHY	
	
def ST_energized(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_ENERGIZED	
	
def ST_energy(char,target=None,param=None,param2=None):
	return (js.Object_GetAttr(char,'material') == 5) or (TryUserSubtype(char,'energy'))

def ST_exiled(char,target=None,param=None,param2=None):
	return js.Object_GetPrimaryState(char)== js.PCSTATE_DIMENSION_SHIFT

def ST_enraged(char,target=None,param=None,param2=None):
	return js.Object_GetPrimaryState(char) == js.PCSTATE_ENRAGED

def ST_fasterthan(char,target=None,param=None,param2=None):
	return js.Object_GetAttr(char,'speed') > js.Object_GetAttr(target,'speed')	
	
def ST_ffx_weakened(char,target=None,param=None,param2=None):
	return ffx.FFX_ObjectGetAttr(char,'weak')	
	
def ST_fire(char,target=None,param=None,param2=None):
	return (js.Object_GetAttr(char,'material') == 4) or (TryUserSubtype(char,'fire'))

def ST_flammable(char,target=None,param=None,param2=None):
	if js.Object_GetTemplate(char) in ['cub_military_crate', 'ger_crate_ammo', 'truck_ger_ammo', 'cub_fuel_barrels', 'ger_barrel']:
		return 1
	else: return 0

def ST_flesh(char,target=None,param=None,param2=None):
	return (js.Object_GetAttr(char,'material') == 0) or (TryUserSubtype(char,'flesh'))

def ST_flying(char,target=None,param=None,param2=None):
	return AIIsFlying(char)

def ST_frozen(char,target=None,param=None,param2=None):
	return js.Object_GetPrimaryState(char) == js.PCSTATE_FROZEN

def ST_geneticdamage(char,target=None,param=None,param2=None):
	return 0

def ST_grappled(char,target=None,param=None,param2=None):
	return ffx.FFX_ObjectGetAttr(char,'rangedgrappled') or ffx.FFX_ObjectGetAttr(char,'grappled') or ffx.FFX_ObjectGetAttr(char,'rangedgrappledCD')

def ST_grappler(char,target=None,param=None,param2=None):
	return ffx.FFX_ObjectGetAttr(char,'rangedgrappler') or ffx.FFX_ObjectGetAttr(char,'grappler') or ffx.FFX_ObjectGetAttr(char,'rangedgrapplerCD')

def ST_hero(char,target=None,param=None,param2=None):
	return ffx.FFX_HasPortrait(char)

def ST_hexed(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_HEXED

def ST_higherenergy(char,target=None,param=None,param2=None):
	return CompareEnergy(Object_GetAttr(char,'energyPoints'),param,'>')
	
def ST_higherenergythan(char,target=None,param=None,param2=None):
	return Object_GetAttr(char,'energyPoints') > Object_GetAttr(target,'energyPoints')

def ST_higherhealth(char,target=None,param=None,param2=None):
	return Object_GetAttr(char,'health') > int(param)
	
def ST_higherhealththan(char,target=None,param=None,param2=None):
	return Object_GetAttr(char,'health') > Object_GetAttr(target,'health')	
	
def ST_higherhealthpct(char,target=None,param=None,param2=None):
	curpct = 100.0 * js.Object_GetAttr(char,'health') / js.Object_GetAttr(char,'maxHealth')
	if curpct >= int(param):
		return 1
	else: return 0		

def ST_highermass(char,target=None,param=None,param2=None):
	return Object_GetAttr(char,'mass') > int(param)

def ST_higherspeed(char,target=None,param=None,param2=None):
	return Object_GetAttr(char,'speed') > int(param)

def ST_higherstrength(char,target=None,param=None,param2=None):
	if HasStrengthAttr(char):
		return js.Object_GetAttr(char,'strength') > int(param)
	else: return 0

def ST_higherstrengththan(char,target=None,param=None,param2=None):
	if HasStrengthAttr(char):
		return js.Object_GetAttr(char,'strength') > js.Object_GetAttr(target,'strength')
	else: return 0
	 
def ST_hurt(char,target=None,param=None,param2=None):
	return js.Object_GetAttr(char,'health') < Object_GetAttr(char,'maxHealth')

def ST_hypnotized(char,target=None,param=None,param2=None):
	return js.Object_GetPrimaryState(char) == js.PCSTATE_HYPNOTISED

def ST_ice(char,target=None,param=None,param2=None):
	return (js.Object_GetAttr(char,'material') == 9) or (TryUserSubtype(char,'ice'))

def ST_ice_control(char,target=None,param=None,param2=None):
	return ffx.hasAttribute(char,'icecontrol')	
	
def ST_illusion(char,target=None,param=None,param2=None):
	return (ffx.FFX_ObjectGetAttr(char,'minionType') == 4)
	
def ST_invisible(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_INVISIBLE

def ST_irradiated(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_IRRADIATED

def ST_interogatable(char,target=None,param=None,param2=None):
	return js.Object_GetPrimaryState(char) == js.PCSTATE_INTEROGATE	

def ST_koed(char,target=None,param=None,param2=None):
	if Object_Exists(char):
		return not Object_IsAlive(char)
	return 0

def ST_lowerenergy(char,target=None,param=None,param2=None):
	return CompareEnergy(js.Object_GetAttr(char,'energyPoints'),param,'<')

def ST_lowerenergythan(char,target=None,param=None,param2=None):
	return Object_GetAttr(char,'energyPoints') < Object_GetAttr(target,'energyPoints')	
	
def ST_lowerhealth(char,target=None,param=None,param2=None):
	#print 'checking lowerhealth ' + char + ' ' + `int(param)`
	return js.Object_GetAttr(char,'health') < int(param)

def ST_lowerhealthpct(char,target=None,param=None,param2=None):
	curpct = 100.0* js.Object_GetAttr(char,'health') / js.Object_GetAttr(char,'maxHealth')
	#print 'lowerhealthpct ' + char + ' ' + `curpct` + ' < ' + `int(param)`
	if curpct <= int(param):
		return 1
	else: return 0	
	
def ST_lowerhealththan(char,target=None,param=None,param2=None):
	return Object_GetAttr(char,'health') < Object_GetAttr(target,'health')
	
def ST_lowermass(char,param=None):
	return js.Object_GetAttr(char,'mass') < int(param)

def ST_lowerspeed(char,target=None,param=None,param2=None):
	return js.Object_GetAttr(char,'speed') < int(param)
   
def ST_lowerstrength(char,target=None,param=None,param2=None):
	if HasStrengthAttr(char):
		return js.Object_GetAttr(char,'strength') < int(param)
	else: return 0

def ST_maxdist(char,target=None,param=None,param2=None):
	if (param==None) or target==None:
		return 0
	else: return (m25distance.GetDistance(char,target) <= m25distance.DistanceConvert(param))	
	
def ST_melee_acid(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'acid', 'melee')
	
def ST_melee_cold(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'cold', 'melee')	

def ST_melee_crushing(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'crushing', 'melee')	
	
def ST_melee_energy(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'energy', 'melee')	
	
def ST_melee_electrical(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'electrical', 'melee')

def ST_melee_fire(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'fire', 'melee')	
	
def ST_melee_mental(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'mental', 'melee')	
		
def ST_melee_mystical(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'mystical', 'melee')	
	
def ST_melee_piercing(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'piercing', 'melee')	
	
def ST_melee_radiation(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'radiation', 'melee')	
	
def ST_metal(char,target=None,param=None,param2=None):
	return (js.Object_GetAttr(char,'material') == 1) or (TryUserSubtype(char,'metal'))

def ST_mindist(char,target=None,param=None,param2=None):
	if (param==None) or target==None:
		return 0
	else: return (m25distance.GetDistance(char,target) >= m25distance.DistanceConvert(param))
	
def ST_minion(char,target=None,param=None,param2=None):
	return js.Object_GetClass(char) & js.OC_MINION

def ST_near(char, target=None, param=None, param2=None):
	dist = param2
	if param2 == None:
		dist = m25distance.DISTANCE_AREA_MEDIUM
	reaction = param
	if param == None:
		reaction = TARGET_ENEMY
	target = FindOneOfType(char,reaction,maxdist=dist)
	if target == None:
		return 0
	else: return 1

def ST_nearattackerally(char,target=None,param=None,param2=None):
	#print 'attackerally ' + `char` + `target` + `param`
	p = param
	if param == None:
		p = m25distance.DISTANCE_AREA_MEDIUM
	ally_list = filter(lambda x,c=char: m25team.AreAllies(x,c), AI_GetNearEnemies(target))
	if len(ally_list):
		return 1
	else:
		return 0

def ST_neardecoy(char,target=None,param=None,param2=None):
	p = m25distance.DistanceConvert(param)
	if param == None:
		p = m25distance.DISTANCE_LONG
	objects = js.Mission_GetDynamicObjects()
	for obj in objects:
		if ST_decoy(obj) and (m25distance.GetDistance(char,obj)<=p):
			return 1
	else: return 0
	
def ST_nearenemy(char,target=None,param=None,param2=None):
	p = param
	if param == None: p = m25distance.DISTANCE_AREA_MEDIUM
	enemy, distance = AI_GetNearestEnemy(char)
	return (enemy and (distance <= p))


def ST_nearfriend(char,target=None,param=None,param2=None):
	return ST_nearally(char,target,param,param2)
	
def ST_nearally(char,target=None,param=None,param2=None):
	p = param
	if param == None:
		p = m25distance.DISTANCE_AREA_MEDIUM
	ally, distance = AI_GetNearestAlly(char)
	return (ally and (distance <= p))
			
def ST_normal(char,target=None,param=None,param2=None):
	return js.Object_GetPrimaryState(char) == js.PCSTATE_NORMAL
	    
def ST_nulled(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_POWER_NULLIFICATION

def ST_object(char,target=None,param=None, param2=None):
	c = js.Object_GetClass(char)
	return not (c & ( ffx.FFX_CHARACTER |  js.OC_CONTROLLABLE |  js.OC_POLICE |  js.OC_CIVILIAN |  js.OC_MINION |  js.OC_VILLIAN))	

def ST_objectified(char,target=None,param=None,param2=None):
	return js.Object_GetPrimaryState(char) == js.PCSTATE_OBJECT
		
def ST_panicked(char,target=None,param=None,param2=None):
	return js.Object_GetPrimaryState(char) == js.PCSTATE_PANICKED

def ST_phased(char,target=None,param=None,param2=None):
	return ffx.FFX_ObjectGetAttr(char,'phased')

def ST_physical(char,target=None,param=None,param2=None):
	return js.Object_GetAttr(char,'physical')	
	
def ST_police(char,target=None,param=None,param2=None):
	return js.Object_GetClass(char) & js.OC_POLICE
	
def ST_polymorphed(char,target=None,param=None,param2=None):
	return js.Object_GetPrimaryState(char) == js.PCSTATE_POLYMORPHED

def ST_possessed(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_POSSESSED
	
def ST_powerup(char,target=None,param=None,param2=None):
	return js.Object_GetClass(char) & js.OC_POWERUP

def ST_radioactive(char,target=None,param=None,param2=None):
	if js.Object_GetTemplate(char) in ['abomb']:
		return 1
	else: return 0

def ST_ranged_acid(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'acid', 'ranged')
	
def ST_ranged_cold(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'cold', 'ranged')	

def ST_ranged_crushing(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'crushing', 'ranged')	
	
def ST_ranged_energy(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'energy', 'ranged')	
	
def ST_ranged_electrical(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'electrical', 'ranged')

def ST_ranged_fire(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'fire', 'ranged')	
	
def ST_ranged_mental(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'mental', 'ranged')	
		
def ST_ranged_mystical(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'mystical', 'ranged')	
	
def ST_ranged_piercing(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'piercing', 'ranged')	
	
def ST_ranged_radiation(char,target=None,param=None,param2=None):
	return HasPowerType(char, 'radiation', 'ranged')
	
	
def ST_spirit_released(char,target=None,param=None,param2=None):
	return ffx.FFX_ObjectGetAttr(char,'release_spirit')


def ST_resistant_acid(char,target=None,param=None,param2=None):
	return (IsSubtype(char,'fire or concrete or air',target)or TryUserSubtype(char,'resistant_acid')) and not TryUserSubtype(char,'weakness_acid')

def ST_resistant_area_acid(char,target=None,param=None,param2=None):
	return HasResistance(char, 'acid', 'area')

def ST_resistant_direct_acid(char,target=None,param=None,param2=None):
	return HasResistance(char, 'acid', 'direct')
	
def ST_resistant_melee_acid(char,target=None,param=None,param2=None):
	return HasResistance(char, 'acid', 'melee')

def ST_resistant_ranged_acid(char,target=None,param=None,param2=None):
	return HasResistance(char, 'acid', 'ranged')


def ST_resistant_cold(char,target=None,param=None,param2=None):
	return (IsSubtype(char,'frozen or stone or wood or concrete or air',target)or TryUserSubtype(char,'resistant_cold')) and not TryUserSubtype(char,'weakness_cold')

def ST_resistant_area_cold(char,target=None,param=None,param2=None):
	return HasResistance(char, 'cold', 'area')
	
def ST_resistant_direct_cold(char,target=None,param=None,param2=None):
	return HasResistance(char, 'cold', 'direct')
	
def ST_resistant_melee_cold(char,target=None,param=None,param2=None):
	return HasResistance(char, 'cold', 'melee')
	
def ST_resistant_ranged_cold(char,target=None,param=None,param2=None):
	return HasResistance(char, 'cold', 'ranged')
	
	
def ST_resistant_crushing(char,target=None,param=None,param2=None):
	return (IsSubtype(char,'metal or rubber or stone or wood or concrete or air or cloth',target)or TryUserSubtype(char,'resistant_crushing')) and not TryUserSubtype(char,'weakness_crushing')		
	
def ST_resistant_area_crushing(char,target=None,param=None,param2=None):
	return HasResistance(char, 'crushing', 'area')
	
def ST_resistant_direct_crushing(char,target=None,param=None,param2=None):
	return HasResistance(char, 'crushing', 'direct')
	
def ST_resistant_melee_crushing(char,target=None,param=None,param2=None):
	return HasResistance(char, 'crushing', 'melee')
	
def ST_resistant_ranged_crushing(char,target=None,param=None,param2=None):
	return HasResistance(char, 'crushing', 'ranged')
		
	
def ST_resistant_electrical(char,target=None,param=None,param2=None):
	return (IsSubtype(char,'fire or frozen or stone or concrete',target)or TryUserSubtype(char,'resistant_electrical')) and not TryUserSubtype(char,'weakness_electrical')

def ST_resistant_area_electrical(char,target=None,param=None,param2=None):
	return HasResistance(char, 'electrical', 'area')
	
def ST_resistant_direct_electrical(char,target=None,param=None,param2=None):
	return HasResistance(char, 'electrical', 'direct')
	
def ST_resistant_melee_electrical(char,target=None,param=None,param2=None):
	return HasResistance(char, 'electrical', 'melee')
	
def ST_resistant_ranged_electrical(char,target=None,param=None,param2=None):
	return HasResistance(char, 'electrical', 'ranged')
		
	
def ST_resistant_energy(char,target=None,param=None,param2=None):
	return (IsSubtype(char,'air',target)or TryUserSubtype(char,'resistant_energy')) and not TryUserSubtype(char,'weakness_energy')
	
def ST_resistant_area_energy(char,target=None,param=None,param2=None):
	return HasResistance(char, 'energy', 'area')
	
def ST_resistant_direct_energy(char,target=None,param=None,param2=None):
	return HasResistance(char, 'energy', 'direct')
	
def ST_resistant_melee_energy(char,target=None,param=None,param2=None):
	return HasResistance(char, 'energy', 'melee')
	
def ST_resistant_ranged_energy(char,target=None,param=None,param2=None):
	return HasResistance(char, 'energy', 'ranged')
	
	
def ST_resistant_fire(char,target=None,param=None,param2=None):
	return (IsSubtype(char,'fire or stone or concrete',target)or TryUserSubtype(char,'resistant_fire')) and not TryUserSubtype(char,'weakness_fire')	
	
def ST_resistant_area_fire(char,target=None,param=None,param2=None):
	return HasResistance(char, 'fire', 'area')
	
def ST_resistant_direct_fire(char,target=None,param=None,param2=None):
	return HasResistance(char, 'fire', 'direct')
	
def ST_resistant_melee_fire(char,target=None,param=None,param2=None):
	return HasResistance(char, 'fire', 'melee')
	
def ST_resistant_ranged_fire(char,target=None,param=None,param2=None):
	return HasResistance(char, 'fire', 'ranged')
			
	
def ST_resistant_area_mental(char,target=None,param=None,param2=None):
	return HasResistance(char, 'mental', 'area')

def ST_resistant_direct_mental(char,target=None,param=None,param2=None):
	return HasResistance(char, 'mental', 'direct')
	
def ST_resistant_melee_mental(char,target=None,param=None,param2=None):
	return HasResistance(char, 'mental', 'melee')
	
def ST_resistant_ranged_mental(char,target=None,param=None,param2=None):
	return HasResistance(char, 'mental', 'ranged')
	
	
	
def ST_resistant_area_mystical(char,target=None,param=None,param2=None):
	return HasResistance(char, 'mystical', 'area')

def ST_resistant_direct_mystical(char,target=None,param=None,param2=None):
	return HasResistance(char, 'mystical', 'direct')
	
def ST_resistant_melee_mystical(char,target=None,param=None,param2=None):
	return HasResistance(char, 'mystical', 'melee')
	
def ST_resistant_ranged_mystical(char,target=None,param=None,param2=None):
	return HasResistance(char, 'mystical', 'ranged')
	

def ST_resistant_piercing(char,target=None,param=None,param2=None):
	return (IsSubtype(char,'metal or rubber or stone or wood or concrete or air',target)or TryUserSubtype(char,'resistant_piercing')) and not TryUserSubtype(char,'weakness_piercing')
	
def ST_resistant_area_piercing(char,target=None,param=None,param2=None):
	return HasResistance(char, 'piercing', 'area')
	
def ST_resistant_direct_piercing(char,target=None,param=None,param2=None):
	return HasResistance(char, 'piercing', 'direct')
	
def ST_resistant_melee_piercing(char,target=None,param=None,param2=None):
	return HasResistance(char, 'piercing', 'melee')
	
def ST_resistant_ranged_piercing(char,target=None,param=None,param2=None):
	return HasResistance(char, 'piercing', 'ranged')


def ST_resistant_radiation(char,target=None,param=None,param2=None):
	return (IsSubtype(char,'metal or stone or concrete or air or cloth',target)or TryUserSubtype(char,'resistant_radiation')) and not TryUserSubtype(char,'weakness_radiation')	
	
def ST_resistant_area_radiation(char,target=None,param=None,param2=None):
	return HasResistance(char, 'radiation', 'area')
	
def ST_resistant_direct_radiation(char,target=None,param=None,param2=None):
	return HasResistance(char, 'radiation', 'direct')
	
def ST_resistant_melee_radiation(char,target=None,param=None,param2=None):
	return HasResistance(char, 'radiation', 'melee')
	
def ST_resistant_ranged_radiation(char,target=None,param=None,param2=None):
	return HasResistance(char, 'radiation', 'ranged')
	
		    
def ST_rubber(char,target=None,param=None,param2=None):
	return (js.Object_GetAttr(char,'material') == 3) or (TryUserSubtype(char,'rubber'))

def ST_samematerial(char, target=None, param=None, param2=None):
	if target == None: return 0
	return js.Object_GetAttr(char,'material') == js.Object_GetAttr(target,'material')	
	
def ST_scary(char,target=None,param=None,param2=None):
	return js.Object_GetAttr(char,'scary')	

def ST_shrunk(char,target=None,param=None,param2=None):
	if (js.Object_GetAttr(char,'speed') == 0) and (js.Object_GetAttr(char,'baseSpeed')==0) and (js.Object_GetSecondaryStates(char) & js.SCSTATE_INVISIBLE):
		return 1
	else: return 0
		
def ST_slowerthan(char,target=None,param=None,param2=None):
	return js.Object_GetAttr(char,'speed') < js.Object_GetAttr(target,'speed')	

def ST_spawn(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_SPAWN

def ST_speed_increased(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_SPEED_INCREASED

def ST_speed_decreased(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_SPEED_DECREASED

def ST_staticked(char,target=None,param=None,param2=None):
	return js.Object_GetPrimaryState(char) == js.PCSTATE_STATIC or ffx.FFX_ObjectGetAttr(char,'eTangled') or ffx.FFX_ObjectGetAttr(char,'tangled') or ffx.FFX_ObjectGetAttr(char,'rangedgrappled') or ffx.FFX_ObjectGetAttr(char,'grappled') or ffx.FFX_ObjectGetAttr(char,'rangedgrappledCD')

def ST_stasis(char,target=None,param=None,param2=None):
	return js.Object_GetPrimaryState(char) == js.PCSTATE_STATIC or ffx.FFX_ObjectGetAttr(char,'eTangled') or ffx.FFX_ObjectGetAttr(char,'tangled') or ffx.FFX_ObjectGetAttr(char,'rangedgrappled') or ffx.FFX_ObjectGetAttr(char,'grappled') or ffx.FFX_ObjectGetAttr(char,'rangedgrappledCD')

def ST_stone(char,target=None,param=None,param2=None):
	return (js.Object_GetAttr(char,'material') == 2) or (TryUserSubtype(char,'stone'))

def ST_strongerthan(char,target=None,param=None,param2=None):
	if HasStrengthAttr(char) and HasStrengthAttr(target):
		return js.Object_GetAttr(char,'strength') < js.Object_GetAttr(target,'strength')	
	else: return 0
	
def ST_studying_target(char,target=None,param=None,param2=None):
	return ffx.FFX_ObjectGetAttr(char,'addingToDB')

def ST_stunned(char,target=None,param=None,param2=None):
	return js.Object_GetPrimaryState(char) == js.PCSTATE_STUNNED

def ST_swarmed(char,target=None,param=None,param2=None):
	reaction = param
	if param == None:
		reaction = TEAM_ENEMIES
	xpos,ypos,zpos = js.Get_ObjectPos(char)
	tlist = filter(lambda x,object=char,r = reaction: m25team.TeamsReactionOfType(x,object,r), AI_GetGrid(xpos, ypos, distance=2, char=char))
	if tlist == None:
		return 0
	else:
		if len(tlist) >= 2:
			return 1
		else: 
			return 0

def ST_tangled(char,target=None,param=None,param2=None):
	return ffx.FFX_ObjectGetAttr(char,'eTangled') or ffx.FFX_ObjectGetAttr(char,'tangled')

def ST_terrain(char,target=None,param=None,param2=None):
	return js.Object_GetClass(char) & js.OC_TERRAIN

def ST_toadified(char,target=None,param=None,param2=None):
	return js.Object_GetPrimaryState(char) == js.PCSTATE_TOAD
	
def ST_throwable(char,target=None,param=None,param2=None):
	return js.Object_GetAttr(char,'throwable')	
	
def ST_throwing(char,target=None,param=None,param2=None):
	return GetObjectVar(char,'throwstate',default=0,integer=1)>0

def ST_tkflier(char,target=None,param=None,param2=None):
	return ffx.FFX_ObjectGetAttr(char,'tkFlightTarget')

def ST_tpallies(char,target=None,param=None,param2=None):
	return ffx.FFX_ObjectGetAttr(char,'tpallies')

def ST_tpfoes(char,target=None,param=None,param2=None):
	return ffx.FFX_ObjectGetAttr(char,'tpfoes')

def ST_usesenergy(char,target=None,param=None,param2=None):
	cache = GetObjectVar(char,'ai_usesenergyx',default=None,integer=1,nosave=1)
	if cache != None:
		return cache
	cdata = GetCharData(char)
	if cdata == {}:
		return 0
	else:
		cache = 0
		for powdata in cdata['powers'].keys():
			power = cdata['powers'][powdata]
			if 'EPCost' in power.keys():
				if power['EPCost'] > 0:
					cache = 1
		SetObjectVar(char,'ai_usesenergyx',cache,nosave=1)
		return cache
		
	
def ST_vampiric_control(char,target=None,param=None,param2=None):
	return ffx.FFX_ObjectGetAttr(char,'vMindControlOn')

def ST_vehicle(char,target=None,param=None,param2=None):
	return js.Object_GetClass(char) & js.OC_VEHICLE
	
def ST_vertigo(char,target=None,param=None,param2=None):
	return ffx.FFX_ObjectGetAttr(char,'vertigo')	
	
def ST_villain(char,target=None,param=None,param2=None):
	return js.Object_GetClass(char) & js.OC_VILLIAN
    
def ST_water_source(char,target=None,param=None,param2=None):
	if js.Object_GetTemplate(char) in ['pat_fire_hydrant', 'pat_park_fountain', 'park_fountain_berlin', 'park_fountain_marble_berlin']:
		return 1
	else: return 0

def ST_weak(char,target=None,param=None,param2=None):
	return IsSubtype(char,'stunned or staticked or nulled or frozen or blanked or panicked or blind or hexed',target)

def ST_weakened(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_WEAKENED
	
def ST_weakerthan(char,target=None,param=None,param2=None):
	if HasStrengthAttr(char) and HasStrengthAttr(target):
		return js.Object_GetAttr(char,'strength') > js.Object_GetAttr(target,'strength')	
	else: return 0
	
def ST_weakness_acid(char,target=None,param=None,param2=None):
	return (IsSubtype(char,'flesh or rubber or wood or cloth',target)or TryUserSubtype(char,'weakness_acid')) and not TryUserSubtype(char,'resistant_acid')
	
def ST_weakness_area_acid(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'acid', 'area')
	
def ST_weakness_direct_acid(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'acid', 'direct')
	
def ST_weakness_melee_acid(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'acid', 'melee')
	
def ST_weakness_ranged_acid(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'acid', 'ranged')	
	
	
def ST_weakness_cold(char,target=None,param=None,param2=None):
	return (IsSubtype(char,'fire or rubber or cloth',target)or TryUserSubtype(char,'weakness_cold')) and not TryUserSubtype(char,'resistant_cold')
	
def ST_weakness_area_cold(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'cold', 'area')
	
def ST_weakness_direct_cold(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'cold', 'direct')
	
def ST_weakness_melee_cold(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'cold', 'melee')
	
def ST_weakness_ranged_cold(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'cold', 'ranged')		
	

def ST_weakness_area_crushing(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'crushing', 'area')
	
def ST_weakness_direct_crushing(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'crushing', 'direct')
	
def ST_weakness_melee_crushing(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'crushing', 'melee')
	
def ST_weakness_ranged_crushing(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'crushing', 'ranged')	
	
	
def ST_weakness_electrical(char,target=None,param=None,param2=None):
	return (IsSubtype(char,'metal',target)or TryUserSubtype(char,'weakness_electrical')) and not TryUserSubtype(char,'resistant_electrical')

def ST_weakness_area_electrical(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'electrical', 'area')
	
def ST_weakness_direct_electrical(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'electrical', 'direct')
	
def ST_weakness_melee_electrical(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'electrical', 'melee')
	
def ST_weakness_ranged_electrical(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'electrical', 'ranged')	
	
	
def ST_weakness_area_energy(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'energy', 'area')
	
def ST_weakness_direct_energy(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'energy', 'direct')
	
def ST_weakness_melee_energy(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'energy', 'melee')
	
def ST_weakness_ranged_energy(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'energy', 'ranged')	
	
	
def ST_weakness_fire(char,target=None,param=None,param2=None):
	return (IsSubtype(char,'frozen or rubber or wood or cloth',target)or TryUserSubtype(char,'weakness_fire')) and not TryUserSubtype(char,'resistant_fire')
	
def ST_weakness_area_fire(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'fire', 'area')
	
def ST_weakness_direct_fire(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'fire', 'direct')
	
def ST_weakness_melee_fire(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'fire', 'melee')
	
def ST_weakness_ranged_fire(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'fire', 'ranged')	
	
	
def ST_weakness_area_mental(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'mental', 'area')
	
def ST_weakness_direct_mental(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'mental', 'direct')
	
def ST_weakness_melee_mental(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'mental', 'melee')
	
def ST_weakness_ranged_mental(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'mental', 'ranged')	
	
	
def ST_weakness_area_mystical(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'mystical', 'area')
	
def ST_weakness_direct_mystical(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'mystical', 'direct')
	
def ST_weakness_melee_mystical(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'mystical', 'melee')
	
def ST_weakness_ranged_mystical(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'mystical', 'ranged')	
	
	
def ST_weakness_piercing(char,target=None,param=None,param2=None):
	return (IsSubtype(char,'cloth',target)or TryUserSubtype(char,'weakness_piercing')) and not TryUserSubtype(char,'resistant_piercing')

def ST_weakness_area_piercing(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'piercing', 'area')
	
def ST_weakness_direct_piercing(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'piercing', 'direct')
	
def ST_weakness_melee_piercing(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'piercing', 'melee')
	
def ST_weakness_ranged_piercing(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'piercing', 'ranged')	
	
	
def ST_weakness_radiation(char,target=None,param=None,param2=None):
	return (IsSubtype(char,'flesh',target) or TryUserSubtype(char,'weakness_radiation')) and not TryUserSubtype(char,'resistant_radiation')	
	
def ST_weakness_area_radiation(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'radiation', 'area')
	
def ST_weakness_direct_radiation(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'radiation', 'direct')
	
def ST_weakness_melee_radiation(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'radiation', 'melee')
	
def ST_weakness_ranged_radiation(char,target=None,param=None,param2=None):
	return HasWeakness(char, 'radiation', 'ranged')	
	

def ST_wieldable(char,target=None,param=None,param2=None):
	return js.Object_GetAttr(char,'wieldable')

def ST_wood(char,target=None,param=None,param2=None):
	return js.Object_GetAttr(char,'material') == 7

def ST_zombie(char,target=None,param=None,param2=None):
	return js.Object_GetSecondaryStates(char) & js.SCSTATE_ZOMBIE
	
def HasStrengthAttr(char):
	oc = js.Object_GetClass(char) 
	if (oc & js.OC_POLICE) or (oc & js.OC_CONTROLLABLE) or (oc & js.OC_MINION) or (oc & js.OC_VILLIAN):
		return 1
	return 0

def HasActiveDefenceActive(char, damage_type_blocked, attack_mode_blocked, ignoreBought=0):
	#cache
	val = m25getset.GetObjectVar(char, 'defense ' + `damage_type_blocked` + '_' + `attack_mode_blocked`, default=-1,nosave=1)
	if val == 0: return 0
	if val  ==1 and HasRole(char, 'active_defense'): return 1
	cdata = GetCharData(char)
	if cdata == {}:
		return 0
	for pname in cdata['powers'].keys():
		if ignoreBought or ( cdata['powerLevels'][pname] > 0 ):
			power = cdata['powers'][pname]
			if power['PowerType'] == datfiles.PT_ACTIVE_DEFENCE:
				if ( power['DamageTypesBlocked'] & damage_type_blocked ) and \
				( power['AttackModesBlocked'] & attack_mode_blocked ):
					m25getset.GetObjectVar(char, 'defense ' + `damage_type_blocked` + '_' + `attack_mode_blocked`, 1,nosave=1)
					if HasRole(char,GetPowerName(char,power['PowerName'])):
#					if HasRole(char,power['PowerName']):
						return 1
	m25getset.GetObjectVar(char, 'defense ' + `damage_type_blocked` + '_' + `attack_mode_blocked`, 0,nosave=1)
	return 0

# subtype helper functions    
    
ai_damage_mode = {'melee':datfiles.PT_MELEE_BLOCKED, 'ranged':datfiles.PT_RANGED_BLOCKED, 
	'area':datfiles.PT_AREA_BLOCKED, 'direct':datfiles.PT_DIRECT_BLOCKED}

ai_damage_type = {
	'crushing':datfiles.PT_DAMAGE_BLOCKED_CRUSH,
	'piercing':datfiles.PT_DAMAGE_BLOCKED_PIERCE,
	'energy':datfiles.PT_DAMAGE_BLOCKED_ENERGY,
	'fire':datfiles.PT_DAMAGE_BLOCKED_FIRE,
	'radiation':datfiles.PT_DAMAGE_BLOCKED_RADIATION,
	'electrical':datfiles.PT_DAMAGE_BLOCKED_ELECTRICAL,
	'acid':datfiles.PT_DAMAGE_BLOCKED_ACID,
	'cold':datfiles.PT_DAMAGE_BLOCKED_COLD,
	'water':datfiles.PT_DAMAGE_BLOCKED_WATER,
	'mental':datfiles.PT_DAMAGE_BLOCKED_MENTAL,
	'mystical':datfiles.PT_DAMAGE_BLOCKED_MYSTICAL}

ai_pdamage_mode = {'melee':datfiles.PT_MELEE, 'ranged':datfiles.PT_RANGED, 
	'area':datfiles.PT_AREA, 'direct':datfiles.PT_DIRECT}

ai_pdamage_type = {
	'crushing':[datfiles.PT_DAMAGE_CRUSH],
	'piercing':[datfiles.PT_DAMAGE_PIERCE],
	'energy':[datfiles.PT_DAMAGE_ENERGY],
	'fire':[datfiles.PT_DAMAGE_FIRE],
	'radiation':[datfiles.PT_DAMAGE_RADIATION, datfiles.PT_DAMAGE_IRRADIATE],
	'electrical':[datfiles.PT_DAMAGE_ELECTRICAL],
	'acid':[datfiles.PT_DAMAGE_ACID, datfiles.PT_DAMAGE_ACID_BURN],
	'cold':[datfiles.PT_DAMAGE_COLD, datfiles.PT_DAMAGE_FREEZE],
	'water':[datfiles.PT_DAMAGE_WATER],
	'mental':[datfiles.PT_DAMAGE_BLIND, datfiles.PT_DAMAGE_HYPNOSIS, datfiles.PT_DAMAGE_MENTAL_BLANK, datfiles.PT_DAMAGE_PANIC, datfiles.PT_DAMAGE_RAGE],
	'mystical':[datfiles.PT_DAMAGE_HEX, datfiles.PT_DAMAGE_OBJECTIFY, datfiles.PT_DAMAGE_POLYMORPH, datfiles.PT_DAMAGE_STASIS, datfiles.PT_DAMAGE_TOADIFY]}	
		
ai_material_resist = {
	'acid': ['fire','concrete','air'],
	'cold': ['frozen','wood','concrete','air'],
	'crushing':['energy', 'metal', 'rubber', 'stone','concrete','air','cloth'],
	'electrical':['rubber', 'wood','concrete'],
	'energy':['frozen','wood','air'],
	'fire':['fire','metal','stone','concrete'],
	'mental':[],
	'mystical':['energy','metal'],
	'piercing':['energy','frozen','metal','stone','wood','concrete','air'],
	'radiation':['fire','metal','stone','concrete','air','cloth'],
	'stun':[],
}

ai_attribute_resist = {
	'acid': ['neutralize'],
	'cold':['hirsute','temperature control'],
	'crushing':['solid skeleton'],
	'electrical':['charged'],
	'energy':['strange visitor'],
	'fire':['temperature control','volcanic'],
	'mental':['cybernetic brain','disciplined','level headed'],
	'mystical':['unbeliever','level headed'],
	'piercing':['armoured'],
	'radiation':['invertebrate', 'radioactive'],
	'stun':['grim resolve'],
}

ai_material_weak = {
	'acid': ['rubber','wood','cloth'],
	'cold': ['fire', 'rubber', 'coth'],
	'crushing':[],
	'electrical':['energy','metal'],
	'energy':['energy','metal'],
	'fire': ['frozen','rubber','wood','cloth'],
	'mental':[],
	'mystical':['stone','wood',],
	'piercing':['cloth'],
	'radiation':['energy','rubber','wood'],
	'stun':[],
}

ai_attribute_weak = {
	'acid': ['rapid metabolism'],
	'cold': ['cold blooded'],
	'crushing':['glass bones'],
	'electrical':['grounded'],
	'energy':['cosmic chump'],
	'fire':['heat sensitive'],
	'mental':['weak minded'],
	'mystical':['bedeviled'],
	'piercing':['thin skinned'],
	'radiation':['rapid metabolism'],
	'stun':[],
}	
	
matname = {0:'flesh',1:'metal', 2:'stone', 3:'rubber', 4:'fire', 5:'energy', 6:'air', 7:'wood', 8:'concrete', 9:'ice', 10:'cloth', 11:'??'}

def AddOred(string, addendum):
	if string == '':
		return addendum
	if addendum == '':
		return string
	return string + ' or ' + addendum

def HasResistance(char, dmg_type, dmg_mode):
	cachename = 'ai_resistant_'+dmg_mode+'_'+dmg_type
	resist = GetObjectVar(char, cachename, default = None, integer=1,nosave=1)
	if resist==1: return 1
	damage_type = ai_damage_type[dmg_type]
	damage_mode = ai_damage_mode[dmg_mode]
	materials = ai_material_resist[dmg_type]
	attributes = ai_attribute_resist[dmg_type]
	#print resist,dmg_type, dmg_mode, damage_type, damage_mode, matname[int(js.Object_GetAttr(char,'material'))], materials, attributes
	if HasActiveDefenceActive(char,damage_type, damage_mode): return 1
	if resist==0: return 0
	# from here on we have to produce a value to cache
	cachevalue = 0
	if js.Object_GetAttr(char,'complex') == 0:
		cachevalue = 0
	else:
		try:
			if (matname[int(js.Object_GetAttr(char,'material'))] in materials) or \
				chardata.Character_HasPassiveDefence(char, damage_type, damage_mode) or \
				HasOneAttribute(char, attributes):
				cachevalue = 1
		except:
			cachevalue = 0
	SetObjectVar(char, cachename, cachevalue,nosave=1)
	return cachevalue

def HasWeakness(char, dmg_type, dmg_mode):
	cachename = 'ai_weakness_'+dmg_mode+'_'+dmg_type
	weak = GetObjectVar(char, cachename, default = None,nosave=1)
	if weak: return 1
	damage_type = ai_damage_type[dmg_type]
	damage_mode = ai_damage_mode[dmg_mode]
	if HasActiveDefenceActive(char,damage_type, damage_mode): return 0
	if weak == 0: return 0
	# from here on we have to produce a value to cache
	cachevalue = 0
	materials = ai_material_weak[dmg_type]
	attributes = ai_attribute_weak[dmg_type]
	if js.Object_GetAttr(char,'complex') == 0:
		cachevalue = 0
	else:
		try:
			if (matname[int(js.Object_GetAttr(char,'material'))] in materials or HasOneAttribute(char, attributes) ) and \
				(chardata.Character_HasPassiveDefence(char, damage_type, damage_mode) == 0):
				cachevalue = 1
		except:
			cachevalue = 0
	SetObjectVar(char, cachename, cachevalue,nosave=1)
	return cachevalue		
	
def HasOneAttribute(char, attributes):
	if js.Object_GetAttr(char,'complex') == 0:
		return 0
	for a in attributes:
		try:
			if chardata.Character_HasAttribute(char,a):
				return 1
		except:
			return 0
	return 0	
    
def HasPowerType(char, dmg_type, dmg_mode):
	cachename = 'ai_haspowertype_'+dmg_mode+'_'+dmg_type
	check = GetObjectVar(char, cachename, default = None, integer=1,nosave=1)
	if check != None:
		return check
	damage_type = ai_pdamage_type[dmg_type]
	damage_mode = ai_pdamage_mode[dmg_mode]   
	cdata = GetCharData(char)
	if cdata == {}:
		return 0
	for powdata in cdata['powers'].keys():
		power = cdata['powers'][powdata]
		if 'DamageType' in power.keys():
			if (power['PowerType'] == damage_mode) and (power['DamageType'] in damage_type):
				Report(AI_CHANNEL,char + ' has damage type ' + dmg_mode + '_' + dmg_type)
				SetObjectVar(char, cachename, 1,nosave=1)
				return 1
	SetObjectVar(char, cachename, 0,nosave=1)
	return 0
	
#####################################################################################
#
# Tactics - Conditions
# 
# Description:  Checks to perform before a tactic executes
#
#####################################################################################

def AITacticCheck(tactic_id=0,char='',energy=0,pct=20,priority=TACTIC_PRIORITY_LOW,role='none',situation='none',target=None,debug=0):
	#return 0 #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	#if not AIEnabledCheck(char):
	#	if debug:
	#		Report(AI_DEBUG_CHANNEL, 'ai not enabled ' + char)
	#	return 0
	if (energy != 0) and not AIEnergyCheck(char,energy):
		if debug:
			Report(AI_DEBUG_CHANNEL, 'energy not passed ' + char)
		return 0		
	if not AIPriorityCheck(char,tactic_id,target,priority):
		if debug:
			Report(AI_DEBUG_CHANNEL, 'priority not passed ' + char)
		return 0
	if not AIPercentCheck(char,tactic_id,target,pct):
		if debug:
			Report(AI_DEBUG_CHANNEL, 'pct not passed ' + char)
		return 0
	# do character state check before all tactics
	#if not AICharacterStateCheck(char):
	#	if debug:
	#		Report(AI_DEBUG_CHANNEL, 'char state failed ' + char)
	#	return 0
	if (role != 'none') and not AIRoleCheck(char,role):
		if debug:
			Report(AI_DEBUG_CHANNEL, 'incorrect role ' + char)
		return 0
	if (situation != 'none') and not AISituationCheck(char,situation,target,debug):
		if debug:
			Report(AI_DEBUG_CHANNEL, 'situation inappropriate ' + char)
			Report(AI_DEBUG_CHANNEL, `situation`)
		return 0
	#if debug:
	#	Report(AI_DEBUG_CHANNEL, 'ALL CHECKS PASSED ' + char)
	return 1

def AIEnabledCheck(char):
	if IsAIDisabled(char) or (IsAICustomDisabled(char) and not IsAIGoalsOnly(char)) or IsAIAllDisabled():
		return 0
	return 1	
	
def AIEnergyCheck(char, energy):
	if Object_Exists(char):
		energy = m25energy.ConvertEnergy(energy)
		#always allow if no energy required
		if (energy == 0): return 1
		#cannot use powers that require energy if power-nulled
		if (js.Object_GetSecondaryStates(char) & SCSTATE_POWER_NULLIFICATION) and (energy > 0):  
			return 0
		reserve = m25energy.ConvertEnergy(GetAIEnergyReserve(char,default=0))
		ep = Object_GetAttr(char, 'energyPoints')
		#if we haven't reached our energy reserve, don't allow energy use
		if (ep < reserve): return 0
		#if we have enough energy, use the power
		if (ep >= energy): return 1
	return 0	
	
def AIPercentCheck(char,tactic_id, target, pct):
	newpct = pct
	if IsFunction(pct):
		numchars = m25getset.GetMissionVar('_ai_active_characters', default=0,nosave=1)
 		if (numchars < 9): 
			newpct = DetermineDynamicTacticValue(char, tactic_id, target, 'percent', pct)
		else:
			newpct = 50
	return randint(1,100) <= newpct

def AIPriorityCheck(char,tactic_id,target, priority):
	newpriority = priority
	if IsFunction(priority):
		newpriority = DetermineDynamicTacticValue(char, tactic_id, target, 'priority', priority)
	#print 'AIPriorityCheck ' + `newpriority`
	AISetCurrentTacticPriority(char,newpriority)
	return newpriority > GetAIPriority(char)
	
def ConvertPriority(priority):
	if priority == 'low':
		return TACTIC_PRIORITY_LOW
	if priority == 'medium':
		return TACTIC_PRIORITY_MEDIUM
	if priority == 'high':
		return TACTIC_PRIORITY_HIGH
	
def AICharacterStateCheck(char):
	p_state = js.Object_GetPrimaryState(char)
	# stumpy 2006-08-26 - added stunned as first state (very common state).
	if AIInKnockBack(char) or (p_state in [js.PCSTATE_STUNNED, js.PCSTATE_OBJECT, js.PCSTATE_TOAD, js.PCSTATE_FROZEN, js.PCSTATE_STATIC, js.PCSTATE_BLANK, js.PCSTATE_DIMENSION_SHIFT, js.PCSTATE_INTEROGATE, js.PCSTATE_POLYMORPHED]):
		return 0
	return 1
	
def AIRoleCheck(char, role):
	if role == 'none': return 1
	return HasRole(char, role)	
	
def AISituationCheck(char, situation, target, debug):
	if situation == 'none': return 1
	return InSituation(char, situation, target, debug)
	
def AIInKnockBack(char):
	return MLOG_IsInKnockBack(char)

	
#####################################################################################
#
# Tactics - Generic
# 
# Description: tactics for generic attacks, moves, fleeing, etc.
#	tactics use these to minimize duplicated code (and bugs!)
#
#####################################################################################

#-------------------------------------------------------------------------------------
#
# Helper Functions
#
#-------------------------------------------------------------------------------------

def AddToList(vlist, value):
	newlist = vlist
	if IsList(vlist):
		newlist = vlist[0:]
		del vlist
		if IsList(value):
			for v in value:
				newlist.append(v)
		else:
			newlist.append(value)
	elif vlist == 'none' or vlist == None:
		newlist = value
	else:
		if IsList(value):
			newlist = value[0:]
			del value
			newlist.append(vlist)
		else:
			newlist = [vlist,value]
	return newlist

def AIReadActualValues(power, energy,mindist,maxdist):
	if energy == None:
		energy = datfiles.PT_EP_COST_NAMES[datfiles.Campaign_ReadPowers()[power]['EPCost']]
	if mindist == None:
		mindist = datfiles.PT_RANGE_NAMES[datfiles.Campaign_ReadPowers()[power]['RangeMin']]
	if maxdist == None:
		maxdist = datfiles.PT_RANGE_NAMES[datfiles.Campaign_ReadPowers()[power]['RangeMax']]
	#if radius == None:
	#	radius = datfiles.PT_RADIUS_NAMES[datfiles.Campaign_ReadPowers()[power]['Radius']]
	return [energy, mindist, maxdist]		
	
def AngleTooHigh(char, target):
	pos1 = js.Get_ObjectPos(char)
	pos2 = js.Get_ObjectPos(target)
	if math.fabs(pos1[2]-pos2[2]) > 20:
		v = js.Vector_Sub(pos1,pos2)
		xydist = math.sqrt(v[0]*v[0] + v[1]*v[1])
		zdist = math.sqrt(v[2]*v[2])
		if (zdist <= 100) and (AIIsFlying(char)): return 0
		if (math.fabs(xydist) < 5) or (math.fabs(zdist/xydist) >= 1.5):
			return 1
	return 0

dmgconversion = {datfiles.PT_DAMAGE_ACID : 'acid',
		datfiles.PT_DAMAGE_ACID_BURN : 'acid',
		datfiles.PT_DAMAGE_BLIND : 'mental',
		datfiles.PT_DAMAGE_COLD : 'cold',
		datfiles.PT_DAMAGE_CRUSH : 'crushing',
		datfiles.PT_DAMAGE_ELECTRICAL : 'electrical',
		datfiles.PT_DAMAGE_ENERGY : 'energy',
		datfiles.PT_DAMAGE_FIRE : 'fire',
		datfiles.PT_DAMAGE_FREEZE : 'cold',
		datfiles.PT_DAMAGE_HEX : 'mystical',
		datfiles.PT_DAMAGE_HYPNOSIS : 'mental',
		datfiles.PT_DAMAGE_IRRADIATE : 'radiation',
		datfiles.PT_DAMAGE_MENTAL_BLANK : 'mental',
		datfiles.PT_DAMAGE_OBJECTIFY : 'mystical',
		datfiles.PT_DAMAGE_PANIC : 'mental',
		datfiles.PT_DAMAGE_PIERCE : 'piercing',
		datfiles.PT_DAMAGE_POLYMORPH : 'mystical',
		datfiles.PT_DAMAGE_RADIATION : 'radiation',
		datfiles.PT_DAMAGE_RAGE : 'mental',
		datfiles.PT_DAMAGE_SPECIAL : 'none',
		datfiles.PT_DAMAGE_SPEED_DECREASE : 'none',
		datfiles.PT_DAMAGE_SPEED_INCREASE : 'none',
		datfiles.PT_DAMAGE_STASIS : 'mystical',
		datfiles.PT_DAMAGE_TOADIFY : 'mystical',
		datfiles.PT_DAMAGE_WATER : 'water',
		datfiles.PT_DAMAGE_WEAKEN : 'none'}
		
dmgclassconversion = {datfiles.PT_MELEE : 'melee',
		datfiles.PT_RANGED : 'ranged',
		datfiles.PT_AREA : 'area',
		datfiles.PT_DIRECT : 'direct',
		datfiles.PT_ACTIVE_DEFENCE : 'none',
		datfiles.PT_PASSIVE_DEFENCE : 'none',
		datfiles.PT_SPECIAL : 'special'}		
	
		

def GetPowerPct(power):
	cached = GetMissionVar(power+'_cachepct',default=None,nosave=1)
	if cached != None:
		return cached		
	powdata = datfiles.Campaign_ReadPowers()[power]		
	damagetype = powdata['DamageType']
	dmgname = dmgconversion[powdata['DamageType']]
	if dmgname == 'none':
		return 30
	dmgclass = dmgclassconversion[powdata['PowerType']]
	resist = 'resistant_' + dmgclass + '_' + dmgname
	weak = 'weakness_' + dmgclass + '_' + dmgname
	val =  'DSubtype(subtypelow=' + `resist` + ',subtypehigh=' + `weak` + ')'
	SetMissionVar(power+'_cachepct',val,nosave=1)
	return val


	
targetsubtypespecial = {datfiles.PT_SPECIAL_AFTERBURNER : [],
	datfiles.PT_SPECIAL_ALTERATION : ['not blind','not panicked','not hypnotized','not staticked'],
	datfiles.PT_SPECIAL_ALTER_GENETIC_STRUCTURE : [],
	datfiles.PT_SPECIAL_ANIMATE_DEAD : ['dead or not alive or koed or lowerhealth 1 or lowerhealthpct 1'],
	datfiles.PT_SPECIAL_ANIMATE_DEAD_MINIONS :['dead or not alive or koed'],
	datfiles.PT_SPECIAL_BEZERKER :[],
	datfiles.PT_SPECIAL_BUOYANCY : ['not buoyant'],
	datfiles.PT_SPECIAL_BURROW : [],
	datfiles.PT_SPECIAL_CEREBRAL_NORMALISATION : ['blind or panicked or hypnotized or enraged'],
	datfiles.PT_SPECIAL_CLOAK : [],
	datfiles.PT_SPECIAL_CLONE_SELF : [],
	datfiles.PT_SPECIAL_CLONE_TARGET : [],
	datfiles.PT_SPECIAL_DEEP_SENSOR : [],
	datfiles.PT_SPECIAL_DENSITY_MAXIMIZE : [],
	datfiles.PT_SPECIAL_DIMENSION_SHIFT : ['not exiled'],
	datfiles.PT_SPECIAL_DISPLACE_IMAGE : [],
	datfiles.PT_SPECIAL_DISRUPT_OBJECT : [],
	datfiles.PT_SPECIAL_DOPPELGANGER : [],
	datfiles.PT_SPECIAL_ELECTRIFY_SELF : [],
	datfiles.PT_SPECIAL_EMPATHY : [],
	datfiles.PT_SPECIAL_ENERGIZE : ['lowerenergy 33'],
	datfiles.PT_SPECIAL_ENERGIZE_AREA : [],
	datfiles.PT_SPECIAL_EP_DRAW : ['higherenergy 33'],
	datfiles.PT_SPECIAL_EP_TRANSFER : ['lowerenergy 33'],
	datfiles.PT_SPECIAL_HP_BALANCE : ['higherhealththan'],
	datfiles.PT_SPECIAL_IGNITE : [],
	datfiles.PT_SPECIAL_IMAGE_PROJECTOR : [],
	datfiles.PT_SPECIAL_LAWORDER : [],
	datfiles.PT_SPECIAL_LIFE_GIFT : ['lowerhealthpct 50'],
	datfiles.PT_SPECIAL_LIVING_FIRES : [],
	datfiles.PT_SPECIAL_MOLECULAR_EXCITE : [],
	datfiles.PT_SPECIAL_MORPH : [],
	datfiles.PT_SPECIAL_NONE : [],
	datfiles.PT_SPECIAL_NORMALISE : ['blind or panicked or hypnotized or enraged'],
	datfiles.PT_SPECIAL_PATRIOT_CHARGE : [],
	datfiles.PT_SPECIAL_POSSESSION : ['not possessed'],
	datfiles.PT_SPECIAL_POWER_NULLIFICATION : ['usesenergy'],
	datfiles.PT_SPECIAL_RALLY : [],
	datfiles.PT_SPECIAL_REMOVE_DEFENCE : [],
	datfiles.PT_SPECIAL_SACRIFICE : [],
	datfiles.PT_SPECIAL_SONIC_BOOM : [],
	datfiles.PT_SPECIAL_SPEEDING_BULLET : [],
	datfiles.PT_SPECIAL_SPRINT : [],
	datfiles.PT_SPECIAL_TELEPORT_SELF : [],
	datfiles.PT_SPECIAL_TELEPORT_SWAP : [],
	datfiles.PT_SPECIAL_TUMBLE : [],
	datfiles.PT_SPECIAL_TUNNEL : [],
	datfiles.PT_SPECIAL_UNCLOAK : []}	
	
targetsubtypedamage = {datfiles.PT_DAMAGE_ACID : [],
	datfiles.PT_DAMAGE_ACID_BURN : ['not acid_burnt'],
	datfiles.PT_DAMAGE_BLIND : ['not blind'],
	datfiles.PT_DAMAGE_COLD : [],
	datfiles.PT_DAMAGE_CRUSH : [],
	datfiles.PT_DAMAGE_ELECTRICAL : [],
	datfiles.PT_DAMAGE_ENERGY : [],
	datfiles.PT_DAMAGE_FIRE : [],
	datfiles.PT_DAMAGE_FREEZE : ['not frozen'],
	datfiles.PT_DAMAGE_HEX : ['not hexed'],
	datfiles.PT_DAMAGE_HYPNOSIS : ['not hypnotized'],
	datfiles.PT_DAMAGE_IRRADIATE : ['not irradiated'],
	datfiles.PT_DAMAGE_MENTAL_BLANK : ['not blanked'],
	datfiles.PT_DAMAGE_OBJECTIFY : ['not objectified'],
	datfiles.PT_DAMAGE_PANIC : ['not panicked'],
	datfiles.PT_DAMAGE_PIERCE : [],
	datfiles.PT_DAMAGE_POLYMORPH : ['not alone long'],
	datfiles.PT_DAMAGE_RADIATION : [],
	datfiles.PT_DAMAGE_RAGE : ['not enraged','nearfriend'],
	datfiles.PT_DAMAGE_SPECIAL : [],
	datfiles.PT_DAMAGE_SPEED_DECREASE : ['not speed_decreased'],
	datfiles.PT_DAMAGE_SPEED_INCREASE : ['not speed_increased'],
	datfiles.PT_DAMAGE_STASIS : ['not staticked'],
	datfiles.PT_DAMAGE_TOADIFY : ['not toadified'],
	datfiles.PT_DAMAGE_WATER : [],
	datfiles.PT_DAMAGE_WEAKEN : ['not weakened']}	
		
def GetSpecialSubtype(char, powername):
	cached = GetMissionVar(powername+'_subtype',default=None,nosave=1)
	if  cached != None:
		return cached
	power = datfiles.Campaign_ReadPowers()[powername]
	target = []
	if (not (power['PowerType'] in [datfiles.PT_PASSIVE_DEFENCE,datfiles.PT_ACTIVE_DEFENCE])):
		sp = GetFFXSubtype(char,power['PowerName'])
		if sp == []:
			sp = targetsubtypespecial[power['SpecialType']]
		if sp == []:
			sp = targetsubtypedamage[power['DamageType']]
		if sp == []:
			if datfiles.PT_DAMAGE_MAGNITUDE_NAMES[power['Magnitude']] == 'none':
				if datfiles.PT_STUN_NAMES[power['Stun']] != 'none':
					sp = ['not stunned']
		if sp != []:
			for s in sp:
				if not (s in target):
					target.append(s)
	if target == []:
		target = 'none'
	SetMissionVar(powername+'_subtype', target,nosave=1)
	return target

targetsubtypeffx = {'absorb': ['not cloth','not flesh', 'not samematerial'],
'castle': [],
'cloak' : ['not cloaked'],
'crystallise': ['not caged'],
'firecage': ['not caged'],
'gravityincrease': ['higherstrength 0'],
'heal': ['lowerhealthpct 50'],
'icepatch': ['not ice_control', 'not acrobatic', 'not flying'],
'leadership': [],  #needs situation and subtype!
'lifedrain': [],
'magneticpull': ['metal'],
'magneticthrow': ['metal'],
'magnetise': [],
'mentaldamage': [],
'mimic': [],
'staticcage': ['not caged'],
'strengthdrain': ['higherstrength 0'],
'phase': ['not displaced'],  #the ai shouldn't be throwing this one around
'pull': [],
'selfheal': ['lowerhealthpct 50'],
'shrink': ['not shrunk'],
'speedboost': ['lowerspeed 5'],
'strobe': [],
'summon': [],
'taunt': [],
'teleport': [],
'throw': [],
'transmute': [],
'vertigo': ['not vertigo'],
'weaken': ['not ffx_weakened'],
'zerog': ['not flying'],
}	
	
def GetFFXSubtype(char, powername):
	carrier = FindFFXCarrier(char,powername)
	if carrier == None: return []
	if targetsubtypeffx.has_key(carrier):
		#print 'FindFFXCarrier %s %s %s'%(char,powername, `targetsubtypeffx[carrier]`)
		return targetsubtypeffx[carrier]
	else:
		return []

def FindFFXCarrier(char,powername):
	template = m25template.GetComplexTemplate(char)
	power = datfiles.Campaign_ReadPowers()[powername]
	truepowername = GetTruePowerName(char,powername,default=powername)
	if power.has_key('SpecialType'): specialtype = power['SpecialType']
	else:
		specialtype = datfiles.PT_SPECIAL_NONE
	if power.has_key('DamageType'): damagetype = power['DamageType']
	else: 
		damagetype = datfiles.PT_DAMAGE_WATER
	for e in ffxcustom.FFX_CARRIERS:
		if e[0] == template:
			if e[3] == truepowername:
				#print e[1], e[2]
				return e[2]
			else:
				if e[3] == 'all':
					#print specialtypestring[specialtype], damagetypestring[damagetype]
					if (e[1] == specialtypestring[specialtype]) or (e[1] == damagetypestring[damagetype]):
						#print e[1], e[2]
						return e[2]
	return None

		
def ClearTargetRadius(char, target, dist):
	if AIHasGoal(char,'Goal_MoveAway'): return
	all = js.Mission_GetDynamicObjects()
	for a in all:
		if AreAllies(char,a) and (not AIHasGoal(a,'Goal_MoveAway')) and not a==char:
			AddGoal_MoveAway(a,target,dist=dist)
	
	
#-------------------------------------------------------------------------------------
#
# Generic Tactics
#
#-------------------------------------------------------------------------------------	
	
def AISupport_CheckConditions(char, target, distance, pct, priority, mindist, maxdist, subtype, situation, role, energy=0):
	return target != None and (randint(1,100)<pct) and (GetAIPriority(char) < priority) and (distance > mindist) and (distance < maxdist) and IsSubtype(target,subtype,char) and IsSubtype(char,situation,char)
# 	if randint(1,100)>pct: return 0
# 	if GetAIPriority(char) >= priority: return 0
# 	if target == None: return 0
# 	if (distance < mindist): return 0
# 	if (distance > maxdist): return 0
# 	if (subtype != None) and not IsSubtype(target,subtype,char): return 0
# 	if (situation != None) and not IsSubtype(char,situation,target): return 0
# 	return 1

def AISupport_CheckFleeConditions(char, target, distance, pct, priority, mindist, maxdist, subtype, situation, role):
	if target == None: return 0
	if randint(1,100)>pct: return 0
	if GetAIPriority(char) >= priority: return 0
	if target == None: return 0
	if (distance > mindist): return 0
	if (subtype != None) and not IsSubtype(target,subtype,char): return 0
	if (situation != None) and not IsSubtype(char,situation,target): return 0
	return 1
	
def AISupport_DoFunction(char, function, target, priority, time, desc=''):
	AIDisableStandard(char)
	SetAIPriority(char,priority)
	SetPerformingTactic(char)
	regIdleStart(char, 'AITacticDone', user=priority)
	m25event.SetTimer(time/2, 'AICheckNotActing', char, `priority`, user=GetAITimeStamp(char))
	try:
		eval(function)
	except:
		eval('ff.%s'%function)
	if AI_CHANNEL:
		Report(AI_CHANNEL,char + ' : ' + `m25template.GetComplexTemplate(char)` + ' ' + function + ' ' + desc + ' ' + target)

def AISupport_DoFunctionTimed(char, function, target, priority, time, desc=''):
	AIDisableStandard(char)
	m25event.SetTimer(time, 'AITacticDone', char, user=priority)
	SetAIPriority(char,priority)
	SetPerformingTactic(char)
	try:
		eval(function)
	except:
		eval('ff.%s'%function)
	if AI_CHANNEL:
		Report(AI_CHANNEL,char + ' : ' + `m25template.GetComplexTemplate(char)` + ' ' + function + ' ' + desc + ' ' + `target`)
		
		
def AISupport_DoMove(char, target, movetype, speed, loc, char_pos, priority, time, desc=''):		
	AIDisableStandard(char)
	SetAIPriority(char,priority)
	SetPerformingTactic(char)
	Trigger_Move(char,movetype,loc,'',0,speed)
	multiplier = 20
	if movetype == 'fly': multiplier = 40
	arrival_time = 0
	try:
		arrival_time = m25distance.GetTupleDistance(loc,char_pos)/((js.Object_GetAttr(char,'speed')+1)*multiplier)
	except:
		arrival_time = 0
	if arrival_time < 0.2: arrival_time = 0.2
	else: 
		if arrival_time > time: arrival_time = time
	m25event.SetTimer(arrival_time/2,'AICheckNotMoving',char,target,user=1024*priority+GetAITimeStamp(char))
	m25event.SetTimer(arrival_time,'AITacticDone',char,user=priority)
	if AI_CHANNEL:
		Report(AI_CHANNEL,char + ':' + `m25template.GetComplexTemplate(char)` + ' ' + desc + ' ' + `target`)		
		
def AISupport_DoSelfPowerTimed(char, power, target, priority, time, desc=''):
	AIDisableStandard(char)
	SetAIPriority(char,priority)
	js.Trigger_Power(char,char,power,'')
	m25event.SetTimer(time, 'AITacticDone', char, user=priority)
	SetPerformingTactic(char)
	if AI_CHANNEL:
		org_power = power
		power = GetTruePowerName(char, power, default=None)
		if power == None:
			power = org_power
		Report(AI_CHANNEL,char + ':' + `m25template.GetComplexTemplate(char)` + ' ' + `power` + ' ' + desc + ' ' + `target`)

def AISupport_DoPower(char, power, target, priority, time, carrier=-1, level=0, desc=''):
	AIDisableStandard(char)
	SetAIPriority(char,priority)
	if carrier != -1:
		#voodoo for Dr. Mike's carrier attack system
		user = carrier
		js.Trigger_Power(char,target,power,'SetupCustomCarrier',level,js.Get_ObjectPos(char),user)
	else:
		js.Trigger_Power(char,target,power,'',level)
	regIdleStart(char, 'AITacticDone', user=priority)
	m25event.SetTimer(1.0, 'AICheckNotActing', char, `priority`, user=GetAITimeStamp(char))
	SetPerformingTactic(char)
	if AI_CHANNEL:
		org_power = power
		power = GetTruePowerName(char, power, default=None)
		if power == None:
			power = org_power
		Report(AI_CHANNEL,char + ':' + `m25template.GetComplexTemplate(char)` + ' ' + `power` + ' ' + desc + ' ' + `target`)

Global_RegisterAttr('ai_canfly',-1)
		
def AIDetermineMoveType(char,target,distance,flee=0):
	if not flee and distance < m25distance.DISTANCE_MEDIUM:
		return 'move'
	canfly = js.Object_GetAttr(char,'ai_canfly')
	#canfly = m25getset.GetObjectVar(char,'ai_canfly',default=None,nosave=1)
	if canfly == -1:
		if js.Object_GetAttr(char,'complex') != 0:
			if chardata.Character_HasAttribute(char,'flier') or chardata.Character_HasAttribute(char,'levitate'):
				canfly = 1
		if canfly == -1:
			canfly = 0
		js.Object_SetAttr(char, 'ai_canfly', canfly)
		#m25getset.SetObjectVar(char, 'ai_canfly',canfly,nosave=1)
	if canfly == 1:
		return 'fly'
	elif (js.Object_GetSecondaryStates(char) & js.SCSTATE_BUOYANT):
		return 'fly'
	else:
		return 'move'		
		
#--------------------------------------------------------------------

def AIGenericActiveDefense(tactic_id, char, object, flt, user, power='', defense_time=20, pct=80, time=3, energy=0, mindist=0, maxdist=m25distance.DISTANCE_LONG, type=TARGET_ENEMY, subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_HIGH, desc='active defense'):
	target = Target_GetCurrentTarget(char)
	if not AISupport_CheckConditions(char,target,Target_GetDistance(char),pct,priority,mindist,maxdist,subtype,situation,role,energy):
		return 0
	AISupport_DoSelfPowerTimed(char, power, target, priority, time, desc)
	AddRole(char,power)
	AddRole(char,'active_defense')
	m25event.SetTimer(defense_time,'AIActiveDefenseOff',char,power)
	return 1
	
def AIActiveDefenseOff(event):
	char = event.object
	power = event.string
	RemoveRole(char,power)	
	RemoveRole(char, 'active_defense')	

def AIGenericFleeOpt(tactic_id, char, object, flt, user, pct=50, time=0.5, energy=0, mindist=m25distance.DISTANCE_SHORT, maxdist=m25distance.DISTANCE_VISIBLE, type=TARGET_ENEMY, subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM, movetype=None, desc='flee from', speed=0):
	if type == 'any': return 0
	target, dist = AI_GetNearestEnemy(char)
	if (target == None) or (AI_GetIgnore(char,target)): return 0
	if not AISupport_CheckFleeConditions(char, target, dist, pct, priority, mindist, maxdist, subtype, situation, role):
		return 0
	try:
		tpos = js.Get_ObjectPos(target)
		cpos = js.Get_ObjectPos(char)
		angle = skXAngles.fSkXGetAngle (tpos, cpos)
		angle = angle + randint(-30,30)
		loc = skXAngles.fSkXGetPointFromVector(tpos,angle,maxdist-10)
		loc = skXMapInfo.MapInfo_EnsurePointIsOutsideBuilding(loc)
		if movetype == None:
			movetype = AIDetermineMoveType(char,target,dist,flee=1)
#		AISupport_DoMove(char, target, AIDetermineMoveType(char,target,dist,flee=1), speed, loc, cpos, priority, time, desc)
		AISupport_DoMove(char, target, movetype, speed, loc, cpos, priority, time, desc=desc)
		return 1
	except: 
		print 'exception in generic flee opt for %s'%char
		return 0
	

def AIGenericMoveOpt(tactic_id, char, object, flt, user, pct=90, time=2.0, energy=0, mindist=m25distance.DISTANCE_SHORT, maxdist=m25distance.DISTANCE_VISIBLE, type=TARGET_ENEMY, subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_LOW, movetype=None, desc='move to', speed=0):
	target = Target_GetCurrentTarget(char,type,subtype)
	dist = Target_GetDistance(char,type)
	if not AISupport_CheckConditions(char, target, dist, pct, priority, mindist, maxdist, subtype, situation, role):
		return 0
	tpos = js.Get_ObjectPos(target)
	cpos = js.Get_ObjectPos(char)
	if mindist > 60:
		# move to the minimum distance away from the target
		angle = skXAngles.fSkXGetAngle (tpos, cpos)
		loc = skXAngles.fSkXGetPointFromVector(tpos,angle,mindist-10)
		loc = skXMapInfo.MapInfo_EnsurePointIsOutsideBuilding(loc)
	else: loc = tpos
	if movetype == None:
		movetype = AIDetermineMoveType(char,target,dist,flee=0)
#	AISupport_DoMove(char, target, AIDetermineMoveType(char,target,dist), speed, loc, cpos, priority, time, desc=desc)
	AISupport_DoMove(char, target, movetype, speed, loc, cpos, priority, time, desc=desc)
	return 1

		
def AIGenericMoveTypeOpt(tactic_id, char, object, flt, user, pct=90, time=2.0, energy=0, mindist=m25distance.DISTANCE_SHORT, maxdist=m25distance.DISTANCE_VISIBLE, type=TARGET_ENEMY, subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_LOW, mode=None, desc='movement', speed=0):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role):
		return 0
	AIDisableStandard(char)
	SetAIPriority(char,priority)
	SetPerformingTactic(char)
	Trigger_Move(char,mode)
	m25event.SetTimer(time,'AITacticDone',char,user=priority)
	if AI_CHANNEL:
		Report(AI_CHANNEL,char + ':' + `m25template.GetComplexTemplate(char)` + ' ' + desc + ' ' + `target`)	

	
def AIGenericPowerOnTarget(tactic_id, char, object, flt, user, power='', carrier=0, target=None, pct=None, time=3, energy=0, mindist=None, maxdist=None, type=TARGET_ENEMY, subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM, level=0, fn='', desc='attack on',onself=0,end_function=None, end_time=5, overpower=1, use_timer=0, debug=0):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char,target,Target_GetDistance(char,type),pct,priority,mindist,maxdist,subtype,situation,role,energy):
		return 0
	#check for AI overpowering its attack
	level = 0
	if overpower:
		ep = js.Object_GetAttr(char,'energyPoints')
		## AI did not seem to want to overpower attacks if they had an energy reserve.
		## Used m25energy.ConvertEnergy to convert energy_reserve to a number and calculated the amount of ep left after overpowering to make sure overpowering would not deplete energy below the reserve.
		## -TMX 07/30/2008
		reserve = m25energy.ConvertEnergy(GetAIEnergyReserve(char,default=0))
		newEP=ep-(energy+33)
		if ep >= energy+33 and newEP > reserve:
			if AI_CHANNEL:
				m25report.Report(AI_CHANNEL, char + ' overpowers his/her/its attack!')
			level = 1
	AISupport_DoPower(char, power, target, priority, time, carrier=carrier, level=level, desc=desc)
	if end_function:
		m25event.SetTimer(end_time, end_function, char, user=priority)
	return 1



def AIGenericPowerOnSelf(tactic_id, char, object, flt, user, power='', carrier=-1, target=None, pct=None, time=3, energy=0, mindist=None, maxdist=None, type=TARGET_ENEMY, subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM, level=0, fn='', desc='self power', end_function=None, end_time=5, overpower=1, use_timer=0, debug=0):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role, energy):
		return 0
	AISupport_DoPower(char, power, char, priority, time, carrier=carrier, level=0, desc=desc)
	return 1
	
def AIGenericFunctionOnTargetTimed(tactic_id, char, object, flt, user, power='', target=None, pct=None, time=3, energy=0, mindist=None, maxdist=None, type=TARGET_ENEMY, subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM, level=0, fn='', desc='attack on',onself=0,end_function=None, end_time=5, overpower=0, use_timer=0, debug=0):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role, energy):
		return 0
	AISupport_DoFunctionTimed(char, power, target, priority, time, desc=desc)
	return 1	
	
def AIGenericFunctionOnTarget(tactic_id, char, object, flt, user, power='', target=None, pct=None, time=3, energy=0, mindist=None, maxdist=None, type=TARGET_ENEMY, subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM, level=0, fn='', desc='attack on',onself=0,end_function=None, end_time=5, overpower=0, use_timer=0, debug=0):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role, energy):
		return 0
	AISupport_DoFunction(char, power, target, priority, time, desc=desc)
	return 1		
	
def AIGenericTravelPower(tactic_id, char, object, flt, user, power='', pct=20, time=5, energy=0, mindist=m25distance.DISTANCE_SHORT, maxdist=m25distance.DISTANCE_LONG, type=TARGET_ENEMY, subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM, fn='', desc='travel away to', objectdist=m25distance.DISTANCE_SHORT, travel_function=''):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckFleeConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role):
		return 0
	tpos = js.Get_ObjectPos(target)
	cpos = js.Get_ObjectPos(char)
	angle = skXAngles.fSkXGetAngle (tpos, cpos)
	angle = angle + randint(-30,30)
	loc = skXAngles.fSkXGetPointFromVector(tpos,angle,maxdist-10)
	loc = skXMapInfo.MapInfo_EnsurePointIsOutsideBuilding(loc)
	if js.Marker_Exists('ai_travel_marker'):
			js.Marker_SetPos('ai_travel_marker', loc)
			js.Object_MoveToMarker('ai_travel_object','ai_travel_marker')
	else:
		js.Marker_Spawn('ai_travel_marker', loc)
		js.Object_Spawn('ai_travel_object', 'ffx_tiggot', 'ai_travel_marker')
		js.Object_SetSecondaryState('ai_travel_object', js.SCSTATE_INVISIBLE, 1, js.PERMANENT_STATE)
		js.Object_SetAttr('ai_travel_object', 'minHealth', 1)
	AISupport_DoPower(char, power, 'ai_travel_object', priority, time, carrier=-1, level=0, desc=desc)
	return 1

def AIGenericTravelFunction(tactic_id, char, object, flt, user, power='', pct=20, time=5, energy=0, mindist=0, maxdist=m25distance.DISTANCE_LONG, type=TARGET_ENEMY, subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM, fn='', desc='travel away to', objectdist=m25distance.DISTANCE_SHORT, travel_function=''):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role, energy):
		return 0
	AIDisableStandard(char)
	SetAIPriority(char,priority)
	object = FindRandObjectInRange(target,mindist=objectdist,maxdist=maxdist,subtype=['not building','object'],default=None)
	eval(power)
	regIdleStart(char, 'AITacticDone', user=priority)
	m25event.SetTimer(time/2, 'AICheckNotActing', char, `priority`, user=GetAITimeStamp(char))
	SetPerformingTactic(char)
	if AI_CHANNEL:
		Report(AI_CHANNEL,char + ':' + `m25template.GetComplexTemplate(char)` + ' ' + power + ' ' + desc + ' ' + `target`)
	return 1	

def AIGenericTravelFunctionTimed(tactic_id, char, object, flt, user, power='', pct=20, time=5, energy=0, mindist=0, maxdist=m25distance.DISTANCE_LONG, type=TARGET_ENEMY, subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM, fn='', desc='travel away to', objectdist=m25distance.DISTANCE_SHORT, travel_function=''):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role, energy):
		return 0
	AIDisableStandard(char)
	SetAIPriority(char,priority)
	object = FindRandObjectInRange(target,mindist=objectdist,maxdist=maxdist,subtype=['not building','object'],default=None)
	eval(power)
	m25event.SetTimer(time, 'AITacticDone', char, user=priority)
	SetPerformingTactic(char)
	if AI_CHANNEL:
		Report(AI_CHANNEL,char + ':' + `m25template.GetComplexTemplate(char)` + ' ' + power + ' ' + desc + ' ' + `target`)
	return 1	

#--------------------------------------------------------------------


def DetermineMoveType(char,target,flee=0):
	dist = m25distance.GetDistance(char,target)
	if not flee and dist < m25distance.DISTANCE_MEDIUM:
		return 'move'
	canfly = m25getset.GetObjectVar(char,'ai_canfly',default=None,nosave=1)
	if canfly == None:
		if js.Object_GetAttr(char,'complex') != 0:
			if chardata.Character_HasAttribute(char,'flier') or chardata.Character_HasAttribute(char,'levitate'):
				canfly = 1
		if canfly == None:
			canfly = 0
		m25getset.SetObjectVar(char, 'ai_canfly',canfly,nosave=1)
		return DetermineMoveType(char,target)
	elif canfly == 1:
		return 'fly'
	elif (js.Object_GetSecondaryStates(char) & js.SCSTATE_BUOYANT):
		return 'fly'
	else:
		return 'move'

	
def AIGenericMove(tactic_id, char, object, flt, user, pct=100, time=0.5, mindist=20, maxdist=m25distance.DISTANCE_VISIBLE, type=TARGET_ENEMY, subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_LOW, mode=None, speed=0,desc='move to', stopdist=-1):
	#start = ff.clock()
	target = FindTacticTarget(tactic_id, char, mindist=0, maxdist=maxdist, energy=0, pct=pct, priority=priority, role=role, situation=situation, reaction=type, target_trait=subtype)
	#mark = ff.clock()
	#print 'aigenericmove find target ' + `mark-start`
	#print target
	if target == None:
		return 0
	#if not AITacticCheck(tactic_id, char, 0, pct, priority, role, situation, target):
	#	return 0
	dist = 0
	#print 'go!'+`mindist`
	if m25distance.GetDistance(char,target) > m25distance.DistanceConvert(mindist):
		#mark2 = ff.clock()
		#print 'aigenericmove get dist ' + `mark2 - mark`
		if AngleTooHigh(char,target):
			AIGenericRandMove(tactic_id, char, object, flt, user, pct=100, mindist=0, maxdist=100, desc='random move on bad attack angle')
			return 1
		#mark3 = ff.clock()
		#print 'aigenericmove angle too high ' + `mark3 - mark2`
		if stopdist == -1:
			dist = m25distance.DistanceConvert(mindist)-10
		else:
			dist = m25distance.DistanceConvert(stopdist)
		#mark4 = ff.clock()
		#print 'aigenericmove stopdist convert ' + `mark4-mark3`
		location = 0
		location = js.Get_ObjectPos(target)
		if dist > 50 and CharGetGroup(char) != None and (CharGetGroupSize(char) > 1) and (not ST_building(target)):
			miAngle = skXAngles.fSkXGetAngleOfSegment (CharGetGroupSize(char),CharGetGroupPosition(char))
			location = skXAngles.fSkXGetPointFromVector (js.Get_ObjectPos(target), miAngle, m25distance.DistanceConvert(mindist)-10)
			#try to ensure that we are at target level
			#target = (location[0], location[1], js.Get_ObjectPos(target)[2])
		if location != 0:		
			target = skXMapInfo.MapInfo_EnsurePointIsOutsideBuilding(location)
			#target= location
			dist = 0
		#mark5 = ff.clock()
		#print 'aigenericmover skXAngles ' + `mark5-mark4`
		#total = GetNumTargettedBy(target,TARGET_ENEMY)
		#num = GetTargettedByPosition(char,TARGET_ENEMY,target)
		#miAngle = skXAngles.fSkXGetAngleOfSegment (total,num)
		#dist = 0
		AIDoMove(char,mode,target,priority,time,speed,desc,mindist=dist)
		#mark6 = ff.clock()
		#print 'aigenericmover do move ' + `mark6-mark5`
		return 1
	return 0
	
def AIGenericMoveType(tactic_id, char, object, flt, user, pct=20, time=0.5, mindist=20, maxdist=m25distance.DISTANCE_VISIBLE, type=TARGET_ENEMY, subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_LOW, mode='hover', desc='move to'):
	target = FindTacticTarget(tactic_id, char, mindist=mindist, maxdist=maxdist, energy=0, pct=pct, priority=priority, role=role, situation=situation, reaction=type, target_trait=subtype)
	if target != None:
		AIDoMove(char,mode=mode,priority=priority,time=time,description=desc)			
		return 1
	return 0


	
					
def AIGenericFlee(tactic_id, char, object, flt, user, pct=20, time=0.5, mindist=m25distance.DISTANCE_SHORT, maxdist=m25distance.DISTANCE_VISIBLE, type=TARGET_ENEMY, subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM, mode=None, desc='flee to'):
	#target = FindNearestOfType(char, reaction=type, maxdist=mindist, subtype=subtype, default=None)
	target = FindOneOfType(char, reaction=type, maxdist=mindist, subtype=subtype, default=None)
	if target == None:
		return 0
	if not AITacticCheck(tactic_id, char, 0, pct, priority, role, situation, target):
		return 0
	if (GetDistance(char,target) <= DistanceConvert(mindist)) and IsSubtype(target,subtype,char):
		#object = FindFleeObject(char,target,mindist=50,maxdist=maxdist,default=None)
		#object = FindFleePoint(char,target,mindist=50,maxdist=maxdist,default=None)
		target = Target_GetCurrentTarget(char, TARGET_ENEMY)
		if 1==0 and m25distance.GetDistance(char, target) > 100 and CharGetGroup(char) != None and (not ST_building(target)):
			miAngle = skXAngles.fSkXGetAngleOfSegment (CharGetGroupSize(char),CharGetGroupPosition(char))
			object = skXAngles.fSkXGetPointFromVector (js.Get_ObjectPos(target), miAngle, m25distance.DistanceConvert(maxdist)+10)
			#try to ensure that we are at target level
			#target = (location[0], location[1], js.Get_ObjectPos(target)[2])
		else:
			if target == None: return 0
			try:
				angle = skXAngles.fSkXGetAngle (js.Get_ObjectPos(target), Get_ObjectPos(char))
				angle = angle + randint(-30,30)
				object = skXAngles.fSkXGetPointFromVector(js.Get_ObjectPos(target),angle,m25distance.DistanceConvert(maxdist)-10)
			except:
				object = FindFleePoint(char,target,mindist=50,maxdist=maxdist,default=None)
		
		#total = GetNumTargettedBy(target,TARGET_ENEMY)
		#num = GetTargettedByPosition(char,TARGET_ENEMY,target)
		#miAngle = skXAngles.fSkXGetAngleOfSegment (total,num)
		#object = skXAngles.fSkXGetPointFromVector (js.Get_ObjectPos(target), miAngle, m25distance.DistanceConvert(mindist))
		#dist = 0
		if object != None:
			object = skXMapInfo.MapInfo_EnsurePointIsOutsideBuilding(object)
			AIDoMove(char,mode,object,priority,time,0,desc,flee=1)
			return 1
	return 0

def AIGenericRandMove(tactic_id, char, object, flt, user, pct=20, time=AI_STANDARD_MOVE_TIME, mindist=m25distance.DISTANCE_SHORT, maxdist=m25distance.DISTANCE_VISIBLE, type=TARGET_ENEMY, subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM, mode='move', desc='random move', movedist=m25distance.DISTANCE_VISIBLE):
	target = FindTacticTarget(tactic_id, char, mindist=mindist, maxdist=maxdist, energy=0, pct=pct, priority=priority, role=role, situation=situation, reaction=type, target_trait=subtype)
	if target != None:
			#object = FindRandObjectInRange(char,mindist=m25distance.DISTANCE_SHORT,maxdist=maxdist,default=None)
			object = FindRandomPoint(char,mindist=mindist,maxdist=movedist,default=None)
			if object != None:
				AIDoMove(char,mode,object,priority,time,0,desc)
				return 1
	return 0
						
			
#####################################################################################
#
# Tactics - Main Game Tactics
# 
# Description: functions that tell a character to attack, move, etc.
#
#####################################################################################		

# active defenses


def ConvertDefenceTime(power):
	dt = 0
	val = m25getset.GetMissionVar('ai_'+power+'_duration',default=None,nosave=1)
	if val != None:
		dt = int(val)
	else:
		dt = datfiles.Campaign_ReadPowers()[power]['Duration']
	#print 'Duration ' + `dt`
	vals = [10,20,30]
	if val == None:
		m25getset.SetMissionVar('ai_'+power+'_duration',val,nosave=1)
	return vals[dt]
		
def TActiveDefenseRemote(tactic_id, char, object, flt, user, power, pct=50, time=1, energy=0, mindist=0, maxdist=m25distance.DISTANCE_LONG, type='friend',subtype='none',situation='none', role='none', priority=TACTIC_PRIORITY_HIGH, defense_time=20, threat_situation='none', threat_subtype='none', threat_reaction=TARGET_ENEMY):
	#check if already active
	#I'm using roles here to expose it to the aidata definition file (users can check if the power is active in their tactics)
	target = FindTacticTarget(tactic_id, char, mindist=mindist, maxdist=maxdist, energy=energy, pct=pct, priority=priority, role=role, situation=situation, reaction=type, target_trait=subtype)
	if target != None:
		if HasRole(target,power):
			return
		tempsubtype = GetObjectVar(char,power+'defences',default=None,nosave=1)
		if tempsubtype == None:
			#datfiles.ShowPower(datfiles.Campaign_ReadPowers()[power])
			allpowers = datfiles.Campaign_ReadPowers()
			powdata = allpowers[power]
			tempsubtype = ConvertDefenceTypes(powdata)
			SetObjectVar(char,power+'defences',tempsubtype,nosave=1)
		threat_subtype = AddToList(threat_subtype,tempsubtype)
		#print power, threat_subtype
		threat = FindNearestOfType(target, reaction=threat_reaction, maxdist=maxdist, subtype=threat_subtype, default=None)
		#print char, power, target, threat, threat_subtype
		if threat == None:
			return
		AddRole(target, power)
		AddRole(target, 'active_defense')
		m25event.SetTimer(0.5,'ADRemoteSpur',target)
		defense_time = ConvertDefenceTime(power)
		m25event.SetTimer(defense_time,'AIActiveDefenseOff',target,power)
		AIDoPower(char,power,target,priority,0,time,'','remote active defense on')				
			
def ADRemoteSpur(event):
	char = event.object
	AI_Animate(char,'area')
		
def ConvertDefenceTypes(power):
	cached = GetMissionVar(power['PowerName'],default=None,nosave=1)
	if  cached != None:
		return cached
	modes = {datfiles.PT_MELEE_BLOCKED:'melee', datfiles.PT_RANGED_BLOCKED:'ranged', datfiles.PT_AREA_BLOCKED:'area', datfiles.PT_DIRECT_BLOCKED:'direct', datfiles.PT_SPECIAL_BLOCKED:'special'}
	subtype = ''
	for m in modes.keys():
		if power['AttackModesBlocked'] & m:
			subtype = AddOred(subtype, ConvertDefenceMode(power, modes[m]))
	if subtype == '':
		return 'none'
	return subtype

def ConvertDefenceMode(power, mode):
	damtypes = {datfiles.PT_DAMAGE_BLOCKED_CRUSH:'crushing', datfiles.PT_DAMAGE_BLOCKED_PIERCE:'piercing', datfiles.PT_DAMAGE_BLOCKED_ENERGY:'energy',
		datfiles.PT_DAMAGE_BLOCKED_FIRE:'fire', datfiles.PT_DAMAGE_BLOCKED_RADIATION:'radiation', datfiles.PT_DAMAGE_BLOCKED_ELECTRICAL:'electrical',
		datfiles.PT_DAMAGE_BLOCKED_ACID:'acid', datfiles.PT_DAMAGE_BLOCKED_COLD:'cold', datfiles.PT_DAMAGE_BLOCKED_WATER:'water',
		datfiles.PT_DAMAGE_BLOCKED_MENTAL:'mental', datfiles.PT_DAMAGE_BLOCKED_MYSTICAL:'mystical'}
	subtype = ''
	for dt in damtypes.keys():
		if power['DamageTypesBlocked'] & dt:
			#print mode + '_' + damtypes[dt]
			subtype = AddOred(subtype, mode + '_' + damtypes[dt])
			#print subtype
	return subtype


def AIChargeUp(tactic_id, char, object, flt, user, power, pct=50, time=3, energy=None, mindist=0, maxdist=m25distance.DISTANCE_MELEE, type=TARGET_ENEMY, subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM, level=0, fn=''):
	if HasRole(char, 'charged_up'):
		return 0
	if AIGenericPowerOnSelf(tactic_id, char, object, flt, user, power,time=time,energy=energy,mindist=mindist,maxdist=maxdist,type=type,subtype=subtype,situation=situation,role=role,priority=priority):
		AddRole(char,'charged_up')
		m25event.SetTimer(10,'AIChargedUpOffOpt',char)
		return 1
	else: return 0
	
def AIChargedUpOffOpt(event):
	char = event.object
	RemoveRole(char,'charged_up')	
	
	
def TCloneSelfOpt(tactic_id, char, object, flt, user, power, pct=80, time=3, energy=None, mindist=0, maxdist=m25distance.DISTANCE_LONG, type=TARGET_ENEMY, subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_HIGH, level=0, fn='', maxclones=4, clone_can_clone=0):
	if (not clone_can_clone) and AIIsClone(char):
		return 0
	if js.Object_Exists(char):
		if (js.Object_GetAttr(char,'cloneCount') >= maxclones):
			return 0
	return AIGenericPowerOnSelf(tactic_id,char,object,flt,user,power,pct=pct,time=time,energy=energy,mindist=mindist,maxdist=maxdist,type=type,subtype=subtype,situation=situation,role=role,priority=priority)
	
	
def AIDisruptObject(tactic_id, char, object, flt, user, power='', carrier=-1, target=None, pct=None, time=3, energy=None, mindist=None, maxdist=None, type=TARGET_ENEMY, subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM, level=0, fn='', desc='disrupt object',onself=0,end_function=None, end_time=5, overpower=1, use_timer=0, debug=0):
	target = Target_GetCurrentTarget(char)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char), pct, priority, mindist, maxdist, subtype, situation, role):
		return 0
	object, distance = AI_GetNearestObject(target)
	#object = skXMapInfo.MapInfo_GetObjects(js.Get_ObjectPos(target), 'nearest', tfFilter=ff.AIDisruptFilter) 
	AISupport_DoPower(char, power, object, priority, time, carrier=carrier, level=0, desc=desc)
	return 1

# def AIDisruptFilter(obj):
# 	if not (js.Object_GetClass(obj) & ( ffx.FFX_CHARACTER |  js.OC_CONTROLLABLE |  js.OC_POLICE |  js.OC_CIVILIAN |  js.OC_MINION |  js.OC_VILLIAN)):
# 		return 1
# 	return 0	
				
# movement

def TDodgeMissiles(tactic_id, char, object, flt, user, pct=100, time=AI_STANDARD_MOVE_TIME, mindist=60, maxdist=200, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_HIGH):
	if not AITacticCheck(tactic_id, char, 0, pct, priority, role, situation, None):
		return 0
	objects = js.Mission_GetDynamicObjects()
	for obj in objects:
		if Object_GetTemplate(obj)=='projectile_standin' and obj[:5]=='mobj_':
			dist = m25distance.GetDistance(char,obj)
			if (dist >= DistanceConvert(mindist)) and (dist <= DistanceConvert(maxdist)):
				fleepoint = FindFleePoint(char,obj,mindist=50,maxdist=maxdist,default=None)
				if fleepoint != None:
					AIDoMove(char,'move',fleepoint,priority,time,0,'dodge missiles')


def TRandMove(tactic_id, char, object, flt, user, pct=50, time=AI_STANDARD_MOVE_TIME, mindist=m25distance.DISTANCE_LONG, maxdist=m25distance.DISTANCE_WHOLEMAP, type=TARGET_ENEMY, subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_LOW, movetype=None):
	AIGenericRandMove(tactic_id,char,object,flt,user, pct,time,mindist,maxdist,type,subtype,situation,role,priority,movetype,'random move to')


def TUncloak(tactic_id, char, object, flt, user, power,pct=50, time=3, energy=None, mindist=0, maxdist=m25distance.DISTANCE_MELEE, type=TARGET_ENEMY, subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM, level=0, fn='', carrier=-1):
	return 0
	
	
# rally	
	
def TRally(tactic_id, char, object, flt, user, power,pct=50, time=3, energy=None, mindist=0, maxdist=m25distance.DISTANCE_MELEE, type=TARGET_ENEMY, subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM, level=0, fn='', carrier=-1):
	if HasRole(char, 'rally_used'):
		return
	if AIGenericPowerOnSelf(tactic_id, char, object, flt, user, power,time=time,energy=energy,mindist=mindist,maxdist=maxdist,type=type,subtype=subtype,situation=situation,role=role,priority=priority):
	#if AIGenericPower(tactic_id, char, object, flt, user, power, pct, time, energy, mindist, maxdist, type, subtype, situation, role, priority, level, fn, 'charging up for',onself=1):
		AddRole(char,'rally_used')
		m25event.SetTimer(5,'RallyUsedOff',char)
		return 1			
	#if AIGenericPower(tactic_id, char, object, flt, user, power, pct, time, energy, mindist, maxdist, type, subtype, situation, role, priority, level, fn, 'rallying',onself=1):
		
	
def TRallyFriend(tactic_id, char, object, flt, user, power, pct=50, time=3, energy=None, mindist=0, maxdist=m25distance.DISTANCE_MELEE, type=TARGET_ENEMY, subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM, level=0, fn='', carrier=-1):
	if HasRole(char, 'rally_used'):
		return
	if AIGenericPowerOnSelf(tactic_id, char, object, flt, user, power,time=time,energy=energy,mindist=mindist,maxdist=maxdist,type=type,subtype=subtype,situation=situation,role=role,priority=priority,carrier=carrier):
		AddRole(char,'rally_used')
		m25event.SetTimer(5,'RallyUsedOff',char)
		return 1
	
def AIRallyUsedOff(event):
	RemoveRole(event.object,'rally_used')		


# remedies

def TUseRemedy(tactic_id,char,string2,flt,user,power='m25ai Use Remedy',time=3,situation='lowerhealthpct 20',role='none',priority=TACTIC_PRIORITY_HIGH):
	#if not m25timestamp.Timestamp_ObjectAttrOutdated(char, '_hp_nouse_period', delta=10): return 0
	if ffx.FFX_ObjectGetAttr(char,'minionType') == 4: return 0 # illusions should not use remedies
	hp = js.Object_GetAttr(char,'heroPoints')
	if hp < 1:return 0
	if priority <= GetAIPriority(char): return 0
	health = js.Object_GetAttr(char,'health')
	maxhealth = js.Object_GetAttr(char,'maxHealth')
	if health < 1 or ((health/maxhealth) > 0.2): return 0 
	js.Object_SetAttr(char,'heroPoints',hp-1)
	js.Object_SetAttr(char,'health',js.Object_GetAttr(char,'maxHealth'))
	cshelper.normalize(char)
	#m25timestamp.Timestamp_ObjectAttrSync(char, '_hp_nouse_period')
	#AIDoPower(char, power=power, target=char, priority=priority, level=0, time=time, function='', description='use remedy')
	#js.Object_PlayEffect(char, 'effect_m25ai_powerup', '', FX_TRACK_OBJECT_POS)
	AIDoAnimation(char,anim='power',fx='effect_m25ai_powerup',time=time,priority=priority,description='using remedy')
	return 1
	
	
#old stuff to be removed when the transition is complete	
	
def AIGenericPower(tactic_id, char, object, flt, user, power, pct=None, time=3, energy=None, mindist=None, maxdist=None, type=TARGET_ENEMY, subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM, level=0, fn='', desc='attack on',onself=0,end_function=None, end_time=5, overpower=0, use_timer=0, debug=0):
	cached_power_values = m25getset.GetMissionVar(power+'_cachedvalues',default=None,nosave=1)
	if cached_power_values == None:
		energy,mindist,maxdist = AIReadActualValues(power, energy,mindist,maxdist)
		m25getset.SetMissionVar(power+'_cachedvalues',[ConvertEnergy(energy),DistanceConvert(mindist),DistanceConvert(maxdist)],nosave=1)
		cached_power_values = m25getset.GetMissionVar(power+'_cachedvalues',default=None,nosave=1)
	energy,mindist,maxdist = cached_power_values
	if pct == None:
		pct = GetPowerPct(power)
		#pct = 50
		
	if not m25string.IsFunction(power):
		subtype = AddToList(subtype, GetSpecialSubtype(char,power))
	
	#debug = 0
	if debug:
		m25report.Report(AI_DEBUG_CHANNEL,'---------------------------')
	if power in debugpowerlist:
		m25report.Report(AI_DEBUG_CHANNEL,'AIGenericPower ' + char + ' ' + power)
		debug = 1
	target = FindTacticTarget(tactic_id, char, mindist=mindist, maxdist=maxdist, energy=energy, pct=pct, priority=priority, role=role, situation=situation, reaction=type, target_trait=subtype, debug=debug)
	if debug:
		m25report.Report(AI_DEBUG_CHANNEL,'---------------------------')
	if target != None:
		if overpower and (not m25string.IsFunction(power)):
			level = CheckForOverpower(char,energy)
		if AngleTooHigh(char,target):
			AIGenericRandMove(tactic_id, char, object, flt, user, pct=100, mindist=0, maxdist=100, desc='random move on bad attack angle')
			return 1
		if onself:
			if m25string.IsFunction(power):
				AIDoGeneric(char,priority=priority,time=time,description=desc)
				#print m25string.GetLeftString(power,'(') + '(' + `char` + ',' + m25string.GetRightString(power,'(')
				exec m25string.GetLeftString(power,'(') + '(' + `char` + ',' + m25string.GetRightString(power,'(')
			else:
				AIDoPower(char,power,char,priority,level,time,fn,desc,end_function=end_function,end_function_time=end_time,use_timer=use_timer)
			return 1
		else:
			if m25string.IsFunction(power):
				AIDoGeneric(char,priority=priority,time=time,description=desc)
				#print m25string.GetLeftString(power,'(') + '(' + `target` + ',' + m25string.GetRightString(power,'(')
				exec m25string.GetLeftString(power,'(') + '(' + `target` + ',' + m25string.GetRightString(power,'(')
			else:
				AIDoPower(char,power,target,priority,level,time,fn,desc,end_function=end_function,end_function_time=end_time,use_timer=use_timer)
			return 1
	return 0

def AIGenericTravelToObject(tactic_id, char, object, flt, user, power, pct=20, time=5, energy=None, mindist=None, maxdist=None, type=TARGET_ENEMY, subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM, fn='', desc='travel away to', objectdist=m25distance.DISTANCE_SHORT, travel_function=''):
	energy,mindist,maxdist = AIReadActualValues(power, energy,mindist,maxdist)
	target = FindNearestOfType(char, reaction=type, maxdist=maxdist, subtype=subtype, default=None)
	if target == None:
		return 0
	if not AITacticCheck(tactic_id, char, energy, pct, priority, role, situation, target):
		return 0
	if (GetDistance(char,target) <= DistanceConvert(mindist)) and IsSubtype(target,subtype):
		#object = FindRandObjectInRange(target,mindist=objectdist,maxdist=maxdist,subtype=['not building','object'],default=None)
		#object = FindFleePoint(char,target,mindist=50,maxdist=maxdist,default=None)
		#if object != None:
		if m25string.IsFunction(power):
			object = FindRandObjectInRange(target,mindist=objectdist,maxdist=maxdist,subtype=['not building','object'],default=None)
			if object != None:
				AIDoGeneric(char,priority=priority,time=time,description=desc)
				exec m25string.GetLeftString(power,'(') + '(' + `object` + ',' + m25string.GetRightString(power,'(')				
				return 1
		else:
			object = FindFleePoint(char,target,mindist=90,maxdist=maxdist,default=None)
			if object != None:
				AIDoPower(char,power,object,priority,0,time,fn,desc)
				return 1
	return 0


def CheckForOverpower(char, energy):
	if js.Object_GetAttr(char,'energyPoints') >= energy+33 and (randint(1,8)==1):
		if AI_CHANNEL:
			m25report.Report(AI_CHANNEL, char + ' overpowers his/her/its attack!')
		return 1
	return 0	
	
#####################################################################################
#
# Tactics - FFX
# 
# Description: tactics that relate to DrMike's FFX addon
#
#####################################################################################

#------------------------------
# Absorber
#------------------------------

def TAbsorber(tactic_id, char, object, flt, user, power='OnAbsorb(target,char)', objecttype=['physical','not cloth','not flesh'], pct=50, time=1.5, energy=33, mindist=0, maxdist=50, type='enemy', subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM, desc='absorb'):
	#use a timed attribute to prevent the AI from absorbing too often
	if not m25timestamp.Timestamp_ObjectAttrOutdated(char, 'ai_has_absorbed', delta=20): return 0
	target = Target_GetCurrentTarget(char,type,subtype)
	distance = Target_GetDistance(char,type)
	if not AISupport_CheckConditions(char, target, distance, pct, priority, mindist, m25distance.DISTANCE_LONG, subtype, situation, role): 
		return 0
	object = None
	if IsSubtype(target, objecttype):
		object = target
	else:		
		object, distance = AI_GetNearestObject(char)
	if object and IsSubtype(object, objecttype):
		AISupport_DoFunctionTimed(char, power, object, priority, time, desc='absorb material against')
		m25timestamp.Timestamp_ObjectAttrSync(char, 'ai_has_absorbed')
		return 1
	return 0

def AIGen_absorber(char):
	return AIGen_absorbmat(char)

def AIGen_absorbmat(char):
	return [
		["TAbsorbMetal()",1],
		["TAbsorbStone()",1],
		["TAbsorbRubber()",1],
		["TAbsorbAir()",1],
		["TAbsorbWood()",1],
		["TAbsorbConcrete()",1],
		["TAbsorbIce()",1],
		["TAbsorbWater()",1],
		["TAbsorbCloth()",1],
		["TAbsorbAdamantium()",1],
		["TAbsorbSuperAlloy()",1],
		["TAbsorbDivineMetal()",1],
		["TAbsorbSand()",1],
		["TAbsorbAcid()",1],
		["TAbsorbGas()",1],
		["TAbsorbDiamond()",1],
		["TAbsorbAlienMineral()",1],
	]
	
def AIGen_absorbenergy(char):
	return [
		["TAbsorbFire()",1],
		["TAbsorbEnergy()",1],
		["TAbsorbElectricity()",1],
		["TAbsorbRadiation()",1],
		["TAbsorbSound()",1],
	]
	
def AIOpt_TAbsorber(char,tactic):
	return AIOpt_TAbsorbMetal(char,tactic)
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.OnAbsorb(target, char)',
		'objecttype': ['physical','not cloth', 'not flesh'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}

def AIOpt_TAbsorbMetal(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbMetal(target, char)',
		'objecttype': ['physical','metal'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}	
	
def AIOpt_TAbsorbStone(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbStone(target, char)',
		'objecttype': ['physical','stone','not liquid','not alien_mineral'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}	

def AIOpt_TAbsorbRubber(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbRubber(target, char)',
		'objecttype': ['physical','rubber or vehicle'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}	
	
def AIOpt_TAbsorbFire(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbFire(target, char)',
		'objecttype': ['physical','fire or flammable'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}
	
def AIOpt_TAbsorbEnergy(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbEnergy(target, char)',
		'objecttype': ['physical','energy','not electricity','not radiation','not sound'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}
	
	
def AIOpt_TAbsorbAir(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbAir(target, char)',
		'objecttype': ['physical','air','not gas','not sand','not acid'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}	
	
def AIOpt_TAbsorbWood(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbWood(target, char)',
		'objecttype': ['physical','wood'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}
	
def AIOpt_TAbsorbConcrete(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbConcrete(target, char)',
		'objecttype': ['physical','concrete','not metal','not adamantium','not diamond','not divine_metal','not super_alloy','not flesh'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}	
	
def AIOpt_TAbsorbIce(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbIce(target, char)',
		'objecttype': ['physical','frozen','ice'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}	
	
def AIOpt_TAbsorbCloth(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbCloth(target, char)',
		'objecttype': ['physical','cloth'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}		
	
	
def AIOpt_TAbsorbElectricity(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbElectrical(target, char)',
		'objecttype': ['physical','electrical'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}		
	
def AIOpt_TAbsorbRadiation(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbRadiation(target, char)',
		'objecttype': ['physical','radiation'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}	
	
def AIOpt_TAbsorbSound(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbSound(target, char)',
		'objecttype': ['physical','sound'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}		
	
def AIOpt_TAbsorbAdamantium(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbAdamantium(target, char)',
		'objecttype': ['physical','adamantium'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}	
	
def AIOpt_TAbsorbDivineMetal(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbDivineMetal(target, char)',
		'objecttype': ['physical','divine_metal'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}		
	
def AIOpt_TAbsorbSuperAlloy(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbSuperAlloy(target, char)',
		'objecttype': ['physical','super_alloy'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}		
	
def AIOpt_TAbsorbWater(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbWater(target, char)',
		'objecttype': ['physical','liquid','water_source'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}	
	
def AIOpt_TAbsorbSand(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbSand(target, char)',
		'objecttype': ['physical','sand'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}		
	
def AIOpt_TAbsorbAcid(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbAcid(target, char)',
		'objecttype': ['physical','acid'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}		
	
def AIOpt_TAbsorbGas(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbGas(target, char)',
		'objecttype': ['physical','gas'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}	
	
def AIOpt_TAbsorbDiamond(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbDiamond(target, char)',
		'objecttype': ['physical','diamond'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}	
	
def AIOpt_TAbsorbAlienMineral(char,tactic):
	return { 
		'exec': 'TAbsorber',
		'power': 'ffx.AbsorbAlienMineral(target, char)',
		'objecttype': ['physical','alien_mineral'],
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MELEE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}	
	
absorbmaterialNames=['flesh','metal','stone','rubber','fire','energy','air','wood','concrete','frozen','cloth','electricity','radiation','sound','adamantium','divine_metal','super_alloy','liquid','acid','sand','gas','diamond','alien_mineral']


#------------------------------
# Absorb Powers
#------------------------------

def AIOpt_TAbsorbPowers(char,tactic):
	return { 
		'exec': 'TAbsorbPowers',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_SHORT)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddSubtype(tactic, ["not robot", "not hitech"])',
	}	
	
# m25's note:  this could be cleaned up, but I'll leave it for the next version
def TAbsorbPowers(tactic_id, char, object, flt, user, pct=50, time=1, energy=33, mindist=0, maxdist=100, type='enemy', subtype=['not nulled','not stunned'], situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	if not m25timestamp.Timestamp_ObjectAttrOutdated(char, 'ai_has_absorbedpowers', delta=20): return 0
	target = Target_GetCurrentTarget(char,type,subtype)
	if not target: return 0
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, m25distance.DISTANCE_LONG, subtype, situation, role): 
		return 0
	temp=ffx.FFX_GetTemplate(char)
	root=ffx.GetRootName(temp)
	mat=ffx.GetSubtypeMaterial(target)
	template='%s_%s'%(root,absorbmaterialNames[int(mat)])
	if js.Template_Exists(template):
		AISupport_DoFunctionTimed(char, 'ffx.absorbPower_changeForm(%s,%s)'%(`target`,`char`), target, priority, time, desc='absorb powers against')
	else:
		template='%s (%s)'%(root,absorbmaterialNames[int(mat)])
		if ffx.Custom_Exists(template):
			AISupport_DoFunctionTimed(char, 'ffx.absorbPower_changeForm(%s,%s)'%(`target`,`char`), target, priority, time, desc='absorb powers against')
		else:
			AISupport_DoFunctionTimed(char, 'ffx.absorbPower(%s,%s)'%(`target`,`char`), target, priority, time, desc='absorb powers against')
	m25timestamp.Timestamp_ObjectAttrSync(char, 'ai_has_absorbedpowers')
	return 1

def AIGen_absorbpower(char):
	return ["TAbsorbPowers()",1]
	

#------------------------------
# FF Active Defense
#------------------------------


def TActiveDefenseFFRemote(tactic_id, char, object, flt, user, pct=50, time=1, energy=33, mindist=0, maxdist=200, type='friend', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	defense_time = ffx.getByTemplate(char,FFX_FFACTIVEDEFENSE_CUSTOM,3)
	power = ffx.getByTemplate(char,FFX_FFACTIVEDEFENSE_CUSTOM,2)
	epCost = ffx.getByTemplate(char,FFX_FFACTIVEDEFENSE_CUSTOM,7)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role, energy=epCost): 
		return 0
	AISupport_DoFunction(char, 'ffx.OnffActiveDefense(%s,%s)'%(`target`,`char`), target, priority, time, desc='remote ff active defense')
	AddRole(target,power)
	AddRole(target, 'active_defense')
	m25event.SetTimer(defense_time,'AIActiveDefenseOff',target,power)
	return 1

def AIOpt_TActiveDefenseFFRemote(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'subtype': 'Opt_AddSubtype(tactic,"not active_defense")',
	}

def TActiveDefenseFFSelf(tactic_id, char, object, flt, user, pct=50, time=1, energy=None, mindist=0, maxdist=200, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	if ffx.FFX_ObjectGetAttr(char,'ffactivedefenseMaxHP')==0:
		return 0
	target = Target_GetCurrentTarget(char,type,subtype)
	defense_time = ffx.getByTemplate(char,FFX_FFACTIVEDEFENSE_CUSTOM,3)
	power = ffx.getByTemplate(char,FFX_FFACTIVEDEFENSE_CUSTOM,2)
	epCost = ffx.getByTemplate(char,FFX_FFACTIVEDEFENSE_CUSTOM,7)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role, energy=epCost): 
		return 0
	if ffx.FFX_ObjectGetAttr(char,'ffactivedefenseSelf',0)==1:
		return 0
	turnTo(char,target)
	AISupport_DoFunction(char, 'ffx.OnffActiveDefense(%s,%s)'%(`char`,`char`), target, priority, time, desc='ff active defense')
	AddRole(char,power)
	AddRole(char,'active_defense')
	ffx.FFX_ObjectSetAttr(char,'ffactivedefenseSelf',1)
	m25event.SetTimer(defense_time,'AIActiveDefenseOff',char,power)
	return 1

def TDeflectMissiles(tactic_id, char, object, flt, user, pct=100, time=2, energy=None, mindist=70, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_HIGH):
	if not AITacticCheck(tactic_id, char, 0, pct, priority, role, situation, None):
		return 0
	if ffx.FFX_ObjectGetAttr(char,'ffactivedefenseSelf',0)==1:
		return 0
	if ffx.FFX_ObjectGetAttr(char,'ffactivedefenseMaxHP')==0:
		return 0
	objects = js.Mission_GetDynamicObjects()
	for obj in objects:
		if Object_GetTemplate(obj)=='projectile_standin' and obj[:5]=='mobj_':
			dist = m25distance.GetDistance(char,obj)
			if (dist >= DistanceConvert(mindist)) and (dist <= DistanceConvert(maxdist)):
				target = FindTarget(char, TARGET_ENEMY, mindist=70, maxdist=300, subtype=[subtype], default=None)
				defense_time = ffx.getByTemplate(char,FFX_FFACTIVEDEFENSE_CUSTOM,4)
				defense_time = defense_time+3.0#delay to stop AI from using this power to quickly and not use the animation correctly
				power = ffx.getByTemplate(char,FFX_FFACTIVEDEFENSE_CUSTOM,3)
				epCost = ffx.getByTemplate(char,FFX_FFACTIVEDEFENSE_CUSTOM,8)
				if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role, energy=epCost): 
					return 0
				AISupport_DoFunction(char, 'ffx.OnffActiveDefense(%s,%s)'%(`char`,`char`), target, priority, time, desc='ff deflecting missiles')
				AddRole(char,power)
				AddRole(char,'active_defense')
				ffx.FFX_ObjectSetAttr(char,'ffactivedefenseSelf',1)
				m25event.SetTimer(defense_time,'AIActiveDefenseOff',char,power)
				return 1

def AIOpt_TActiveDefenseFFSelf(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_VISIBLE)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_ALWAYS)',
		'time': 'Opt_Convert(tactic,"time",3)',
		'situation': 'Opt_AddSituation(tactic,"not active_defense")',
		}


#------------------------------
# Amplifier
#------------------------------
# def TAmplify(tactic_id, char, object, flt, user, pct=50, time=3, energy=33, mindist=0, maxdist=100, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
# 	AIGenericPower(tactic_id,char,object,flt,user,'ffx.amplify(char)',pct,time,energy,mindist,maxdist,type,subtype,situation,role,priority,desc='amplify')

def AIOpt_TAmplifier(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'power': 'ffx.amplify(target,char)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_SHORT)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddSubtype(tactic, ["not robot", "not hitech"])',
	}	
	
def AIGen_amplifier(char):
	return ["TAmplifier()",1]
	

#------------------------------
# Animate Dead
#------------------------------

def AIOpt_TAnimateDead(char,tactic):
	return { 
		'exec': 'AIGenericPowerOnSelf',
		'pct': 'Opt_ConvertPct(tactic,PRECALC_PCT_ALWAYS)',
		'mindist': 0,
		'maxdist': 150,
		'energy': 'Opt_GetEnergy(tactic)',
		'subtype': 'Opt_AddSubtype(tactic, "lowerhealthpct 25")',
	}


#------------------------------
# Anti Warzone
#------------------------------

def TAntiWarZone(tactic_id, char, object, flt, user, pct=50, time=2, energy=33, mindist=100, maxdist=300, type='enemy', subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_HIGH):
	if ffx.FFX_ObjectGetAttr(char, 'warzone') != 0: return 0
	target = Target_GetCurrentTarget(char)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.OnPeaceField("%s","%s")'%(char,char), target, priority, time, desc='anti-warzone')
	return 1

def AIOpt_TAntiWarZone(char, tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_HIGH)',
		'category': 'defense',
	}


#------------------------------
# Battle Computer
#------------------------------

def AIOpt_TStudyTarget(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'power': 'ffx.studyTarget5(target,char)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddSubtype(tactic, "not in_database")',
	}	
	
def AIGen_battlecomputer(char):
	return ["TStudyTarget()",1]


#------------------------------
# Bulk Teleporter
#------------------------------

def TBulkTeleporter(tactic_id, char, object, flt, user, pct=50, time=5, energy=33, mindist=0, maxdist=300, type='enemy', subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM, desc='bulk teleporter'):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	objects = skXMapInfo.MapInfo_GetObjects(char,tiDistance=maxdist,tfFilter=AIBulkTPFilter)
	if objects in [None,[]]: return 0
	object = objects[0]
	AISupport_DoFunction(char, 'ffx.OnBulkTP(%s,%s)'%(`object`,`char`), target, priority, time, desc='bulk teleport')
	return 1

def AIBulkTPFilter(object):
	objclass = js.Object_GetClass(object)
	if (objclass & ( ffx.FFX_CHARACTER |  js.OC_CONTROLLABLE |  js.OC_POLICE |  js.OC_CIVILIAN |  js.OC_MINION |  js.OC_VILLIAN | js.OC_BUILDING | js.OC_TERRAIN)):
		return 0
	return 1

def AIOpt_TBulkTeleporter(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
	}


#------------------------------
# Blazing Glory
#------------------------------

def TBlazingGlory(tactic_id, char, object, flt, user, pct=50, time=1.5, energy=0, mindist=0, maxdist=100, type='enemy', subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if ffx.FFX_ObjectGetAttr(char,'blazing') == 1:
		if (Target_GetDistance(char,type) > maxdist) or (js.Object_GetAttr(char,'health') <= 30):
			AISupport_DoFunctionTimed(char, 'ffx.OnBlazeOff(%s,%s)'%(`char`,`char`), target, priority, time, desc='blazing glory off')
			return 1
	else:
		if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
			return 0
		AISupport_DoFunctionTimed(char, 'ffx.OnBlazeOn(%s,%s)'%(`char`,`char`), target, priority, time, desc='blazing glory on')
		return 1

def AIOpt_TBlazingGlory(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_SHORT)',
		'energy': 'Opt_Convert(tactic,"energy",0)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,["higherhealth 30", "alone 100"])',
	}


	
#------------------------------
# Earth Control 
# note: also uses TTK
#------------------------------	
	
def AIOpt_TEarthSpike(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_SHORT)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddSubtype(tactic,"not flying")',
		'power': 'ffx.OnSpike(target,char)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}
	
def AIOpt_TEarthWall(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_MELEE)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_LOW)',
		'subtype': 'Opt_AddSubtype(tactic,"not flying")',
		'power': 'ffx.OnWall(target,char)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}

	
#------------------------------
# Energy Shield
#------------------------------

def TEnergyShield(tactic_id, char, object, flt, user, pct=50, time=3, energy=33, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_HIGH):
	if ffx.FFX_ObjectGetAttr(char, 'mshield') != 0: 
		return 0
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	#make sure the effect is down before we restart the shield, otherwise there is overlap
	if ffx.FFX_ObjectGetAttr(char,'fxHandle',-1)!=-1:
		js.Object_StopEffect(char,ffx.FFX_ObjectGetAttr(char,'fxHandle'))
	AISupport_DoFunction(char, 'ffx.OnShield(%s,%s)'%(`char`,`char`), target, priority, time, desc='energy shield vs.')
	return 1

def AIOpt_TEnergyShield(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_VISIBLE)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_HIGH)',
		'category': 'defense',
	}


#------------------------------
# Environmental Control
#------------------------------	

def AIOpt_TEnvControlGlobal(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_MELEE)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",50)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"alone 100")',
		'power': 'ffx.OnEnvSelf(target,char)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}
	
def AIOpt_TEnvControlLocal(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_MELEE)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",50)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"alone 100")',
		'power': 'ffx.OnEnvTarget(target,char)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}
	
#------------------------------
# Explosive
#------------------------------

def AIOpt_TExplosive(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_SHORT)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"alone 100")',
		'power': 'ffx.OnExplode(char,char)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}
	
#------------------------------
# Fire Control
#------------------------------

def AIOpt_TFireWall(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_SHORT)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'power': 'ffx.OnFireWall(target,char)',
		'time': 'Opt_Convert(tactic,"time",3)',
		'subtype': 'Opt_AddSubtype(tactic,"not nearenemy 100")',
	}
	

#------------------------------
# Free Spirit
#------------------------------

def TFreeSpirit(tactic_id, char, object, flt, user, pct=30, time=1, energy=33, mindist=0, maxdist=200, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.OnReleaseSpirit(%s,%s)'%(`char`,`char`), target, priority, time, desc='release free spirit')
	return 1

def AIOpt_TFreeSpirit(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_VISIBLE)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"alone")',
		'time': 'Opt_Convert(tactic,"time",3)',
	}

def AIGen_freespirit(char):
	return ["TFreeSpirit()",1]
#------------------------------
# Gaseous Form
#------------------------------	

def TGaseousForm(tactic_id, char, object, flt, user, pct=30, time=1, energy=33, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	if ST_displaced(char): return 0
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.OnGaseousForm(%s,%s)'%(`char`,`char`), target, priority, time, desc='turn into poisonous gas')
	AddGoal_MoveGaseous(char)
	return 1

def AIOpt_TGaseousForm(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"alone 100")',
		'time': 'Opt_Convert(tactic,"time",3)',
	}
	
#------------------------------
# Gravity Well
#------------------------------	

def TGravityWell(tactic_id, char, object, flt, user, pct=30, time=1, energy=33, mindist=0, maxdist=200, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.gravity_well(%s,%s)'%(`char`,`char`), target, priority, time, desc='gravity well vs.')
	return 1

def AIOpt_TGravityWell(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MEDIUM)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"alone 100")',
		'time': 'Opt_Convert(tactic,"time",3)',
	}
	

#------------------------------
# Group Teleport
#------------------------------	

def TAddAllies(tactic_id, char, object, flt, user, pct=30, time=1, energy=0, mindist=0, maxdist=200, type='friend', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.OnAddAllies(%s,%s)'%(`char`,`char`), target, priority, time, desc=' adding allies for group teleport')
	return 1

def AIOpt_TAddAllies(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MEDIUM)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"not tpallies")',
	}
	
def TAddFoes(tactic_id, char, object, flt, user, pct=30, time=1, energy=0, mindist=0, maxdist=200, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.OnAddFoes(%s,%s)'%(`char`,`char`), target, priority, time, desc=' adding foes for group teleport')
	return 1

def AIOpt_TAddFoes(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MEDIUM)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"not tpfoes")',
	}

def TRemoveAllies(tactic_id, char, object, flt, user, pct=30, time=1, energy=0, mindist=0, maxdist=200, type='friend', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.OnRemoveAllies(%s,%s)'%(`char`,`char`), target, priority, time, desc=' removing allies for group teleport')
	return 1

def AIOpt_TRemoveAllies(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MEDIUM)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"tpallies")',
	}

def TRemoveFoes(tactic_id, char, object, flt, user, pct=30, time=1, energy=0, mindist=0, maxdist=200, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.OnRemoveFoes(%s,%s)'%(`char`,`char`), target, priority, time, desc=' removing foes for group teleport')
	return 1

def AIOpt_TRemoveFoes(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_MEDIUM)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"tpfoes")',
	}

#------------------------------
# Human Top
#------------------------------	

def THumanTop(tactic_id, char, object, flt, user, pct=30, time=1, energy=33, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	if ST_displaced(char): return 0
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.OnSpin(%s,%s)'%(`char`,`char`), target, priority, time, desc='human top vs.')
	AddGoal_MoveGaseous(char)
	return 1

def AIOpt_THumanTop(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"alone 100")',
		'time': 'Opt_Convert(tactic,"time",3)',
	}
	
	
#------------------------------
# Ice Cloud
#------------------------------	

def TIceCloud(tactic_id, char, object, flt, user, pct=30, time=1, energy=33, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	if ST_displaced(char): return 0
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.OnIceCloud(%s,%s)'%(`char`,`char`), target, priority, time, desc='ice cloud vs.')
	AddGoal_MoveGaseous(char)
	return 1

def AIOpt_TIceCloud(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"alone 100")',
		'time': 'Opt_Convert(tactic,"time",3)',
		'subtype': 'Opt_AddSubtype(tactic,"not frozen")',
	}
	
	
#------------------------------
# Ice Control
#------------------------------

def AIOpt_TIceWall(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_SHORT)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_LOW)',
		'power': 'ffx.OnIceWall(target,char)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}	
	
def AIOpt_TIcePatch(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'power': 'ffx.OnIcePatch(target,char)',
		'time': 'Opt_Convert(tactic,"time",3)',
		'subtype': 'Opt_AddSubtype(tactic,"not flying")',
	}
	
	
#------------------------------
# Illusionist
#------------------------------	

def AIOpt_TIllusionist(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'power': 'ffx.OnIllusion(target,char)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
		'situation': 'Opt_AddSituation(tactic,"not illusion")',
	}	
	
	
#------------------------------
# Inner Being
#------------------------------		
	
def TInnerBeing(tactic_id, char, object, flt, user, pct=10, time=1.5, energy=0, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
 		return 0
 	for index in range(1,4):
 		if ffx.FFX_ObjectGetAttr(char, 'inner'+`index`) != 0:
 			AISupport_DoFunction(char, 'ffx.OnReleaseInner(%i,%s)'%(index,`char`), target, priority, time, desc='release inner being vs.')
 			return 1
 	return 0
 
def AIOpt_TInnerBeing(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}	 	
 		
	


#------------------------------
# Living Laser
#------------------------------	

def AIOpt_TLivingLaser(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'power': 'ffx.OnLaserTo(target,char)',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_SHORT)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}	

def AIOpt_TLivingLaserAway(char,tactic):
	return { 
		'exec': 'AIGenericTravelFunction',
		'power': 'ffx.OnLaserTo(object,char)',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_SHORT)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
		'category': 'flee',
	}		
	
	
#------------------------------
# Loner
#------------------------------		
	
def AIOpt_TLoner(char,tactic):
	return { 
		'exec': 'AIGenericFleeOpt',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_SHORT)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
		'type': TARGET_ALLY,
		'category': 'flee',
	}

	


		
#------------------------------
# Magnetic Control
#------------------------------	

def AIOpt_TMagneticAttract(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'power': 'ffx.OnAttract(target,char)',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_SHORT)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
		'subtype': 'Opt_AddSubtype(tactic,"metal or adamantium or super_alloy")',
	}	

def AIOpt_TMagneticRepel(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'power': 'ffx.OnRepel(target,char)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
		'subtype': 'Opt_AddSubtype(tactic,"metal or adamantium or super_alloy")',
	}
	
def AIOpt_TMagneticCustom(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'power': 'ffx.magnetic_disruption(target,char)',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_SHORT)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
		'subtype': 'Opt_AddSubtype(tactic,"metal or adamantium or super_alloy")',
	}	
	

#------------------------------
# Mass Control
#------------------------------		
	
def TMassControl(tactic_id, char, object, flt, user, pct=10, time=0.5, energy=0, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	dist = Target_GetDistance(char,type)
	if not AISupport_CheckConditions(char, target, dist, pct, priority, mindist, maxdist, subtype, situation, role): 
 		return 0
 	state = ffx.FFX_ObjectGetAttr(char,'massstate')
 	power = None
 	if (dist>200) and (state>=0): power = 'ffx.OnMassDown(%s,%s)'%(`char`,`char`)
 	elif (dist<200) and (dist>80):
 		if state>0: power = 'ffx.OnMassDown(%s,%s)'%(`char`,`char`)
 		elif state<0: power = 'ffx.OnMassUp(%s,%s)'%(`char`,`char`)
 		else: power = None
 	elif (dist<80) and (state <=0): power = 'ffx.OnMassUp(%s,%s)'%(`char`,`char`)
 	else: power = None
 	if power:
 		AISupport_DoFunctionTimed(char, power, target, priority, time, desc='mass control vs.')
 		return 1
 	else:
 		return 0
 
def AIOpt_TMassControl(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_WHOLEMAP)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_HIGH)',
		'time': 'Opt_Convert(tactic,"time",0.5)',
		'category': 'pre',
	}		
		

#------------------------------
# MegaBlast
#------------------------------

def TMegaBlast(tactic_id, char, object, flt, user, pct=30, time=1, energy=0, mindist=0, maxdist=200, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	function=ffx.getByTemplate(char,FFX_MIRACLE1_CUSTOM,2)
	AISupport_DoFunction(char, 'ffx.%s1(target,char)'%ffx.getByTemplate(char,FFX_MIRACLE1_CUSTOM,2), target, priority, time, desc='megablast vs.')
	return 1

def AIOpt_TMegaBlast(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}


#------------------------------
# Phasing
#------------------------------

def TPhaseOut(tactic_id, char, object, flt, user, pct=50, time=1, energy=0, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.OnPhaseOut(%s,%s)'%(`char`,`char`), target, priority, time, desc='phasing out')
	return 1

def TPhaseIn(tactic_id, char, object, flt, user, pct=50, time=1, energy=0, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM, form=None):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.OnNPCChange(%s,%s)'%(`char`,`form`), target, priority, time, desc='phasing in')
	return 1

def TPhaseDisrupt(tactic_id, char, object, flt, user, pct=50, time=1, energy=0, mindist=0, maxdist=100, type='enemy', subtype='robot or computer or hitech', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.onPhaseDisrupt(%s,%s)'%(`target`,`char`), target, priority, time, desc='disrupt electronics')
	return 1

#------------------------------
# Plasma Sculptor
#------------------------------	

def AIOpt_TPlasmaSculptorBarrier(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'power': 'ffx.OnMakeBarrier(target,char)',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_SHORT)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_LOW)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}	

def AIOpt_TPlasmaSculptorDrop(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTarget',
		'power': 'ffx.OnMakeDrop(target,char)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
		'subtype': 'Opt_AddSubtype(tactic,"not nearenemy 80")',
	}


#------------------------------
# Power Reserve
#------------------------------

def TPowerReserveStr(tactic_id, char, object, flt, user, pct=30, time=1, energy=66, mindist=0, maxdist=200, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.boostStr(%s,%s)'%(`target`,`char`), target, priority, time, desc='boost strength')
	return 1

def TPowerReserveAgl(tactic_id, char, object, flt, user, pct=30, time=1, energy=66, mindist=0, maxdist=200, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.boostAgl(%s,%s)'%(`target`,`char`), target, priority, time, desc='boost Agility')
	return 1

def TPowerReserveSpd(tactic_id, char, object, flt, user, pct=30, time=1, energy=66, mindist=0, maxdist=200, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.boostSpd(%s,%s)'%(`target`,`char`), target, priority, time, desc='boost speed')
	return 1

def TPowerReserveEnd(tactic_id, char, object, flt, user, pct=30, time=1, energy=66, mindist=0, maxdist=200, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.boostEnd(%s,%s)'%(`target`,`char`), target, priority, time, desc='boost Endurance')
	return 1

def TPowerReservePwr(tactic_id, char, object, flt, user, pct=30, time=1, energy=66, mindist=0, maxdist=200, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.boostPwr(%s,%s)'%(`target`,`char`), target, priority, time, desc='boost Power')
	return 1

def TPowerReserveRes(tactic_id, char, object, flt, user, pct=30, time=1, energy=66, mindist=0, maxdist=200, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.boostResist(%s,%s)'%(`target`,`char`), target, priority, time, desc='boost Resistance')
	return 1


def AIGen_powerReserve(char):
	return ["TPowerReserveStr()",1]

#------------------------------
# Puppet Master
#------------------------------			

def TPuppetMasterNature(tactic_id, char, object, flt, user, pct=50, time=3, energy=33, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	object = skXMapInfo.MapInfo_GetObjects(char,tiDistance='nearest',tfFilter=AINatureFilter)
	if m25distance.GetDistance(char,object) > m25distance.DISTANCE_LONG: return 0
	#if objects in [None,[]]: return 0
	#object = objects[0]
	AISupport_DoFunction(char, 'ffx.OnAnimateNature(%s,%s)'%(`object`,`char`), target, priority, time, desc='puppet master nature vs.')
	return 1

ai_nature_templates = []	
	
def AI_FillNatureTemplates():
	global ai_nature_templates
	if ai_nature_templates == []:
		for pair in ffx.natureTemplates:
			ai_nature_templates.append(pair[0])

def AINatureFilter(object):
	global ai_nature_templates
	if ai_nature_templates == []: AI_FillNatureTemplates()
	return js.Object_GetTemplate(object) in ai_nature_templates		
		
def AIOpt_TPuppetMasterNature(char,tactic):
	AI_FillNatureTemplates()
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}	

	
def TPuppetMasterUrban(tactic_id, char, object, flt, user, pct=50, time=3, energy=33, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	object = skXMapInfo.MapInfo_GetObjects(char,tiDistance='nearest',tfFilter=AIUrbanFilter)
	if m25distance.GetDistance(char,object) > m25distance.DISTANCE_LONG: return 0
	#if objects in [None,[]]: return 0
	#object = objects[0]
	AISupport_DoFunction(char, 'ffx.OnAnimateUrban(%s,%s)'%(`object`,`char`), target, priority, time, desc='puppet master urban vs.')
	return 1

ai_Urban_templates = []	
	
def AI_FillUrbanTemplates():
	global ai_Urban_templates
	if ai_Urban_templates == []:
		for pair in ffx.urbanTemplates:
			ai_Urban_templates.append(pair[0])

def AIUrbanFilter(object):
	global ai_Urban_templates
	if ai_Urban_templates == []: AI_FillUrbanTemplates()
	return js.Object_GetTemplate(object) in ai_Urban_templates		
		
def AIOpt_TPuppetMasterUrban(char,tactic):
	AI_FillUrbanTemplates()
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}		

	
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!111	
	
def TAstralForm(tactic_id, char, object, flt, user, pct=50, time=1.5, energy=33, mindist=100, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	return
	#AIGenericPower(tactic_id,char,object,flt,user,'ffx.OnReturnNPC(char)',pct,time,energy,mindist,maxdist,type,subtype,situation,role,priority,desc='astral return')
	
def TAstralProjection(tactic_id, char, object, flt, user, pct=50, time=1.5, energy=33, mindist=200, maxdist=500, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	return
	#AIGenericPower(tactic_id,char,object,flt,user,'ffx.OnReleaseAstral(char)',pct,time,energy,mindist,maxdist,type,subtype,situation,role,priority,desc='astral projection', onself=1)	
	

#------------------------------
# Sandstorm
#------------------------------	

def TSandstorm(tactic_id, char, object, flt, user, pct=30, time=1, energy=33, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	if ST_displaced(char): return 0
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.OnSandStorm(%s,%s)'%(`char`,`char`), target, priority, time, desc='sandstorm vs.')
	AddGoal_MoveGaseous(char)
	return 1

def AIOpt_TSandstorm(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"alone 100")',
		'time': 'Opt_Convert(tactic,"time",3)',
	}
	

#------------------------------
# Shapeshift
#------------------------------
	
def TShapeShift(tactic_id, char, object, flt, user, pct=40, time=5, energy=33, mindist=20, maxdist=500, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM, form=None):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	if form == None:
		allforms = GetForms(char)
		if allforms == ['default']: return 0
		if len(allforms) > 0:
			tempform = randint(0,len(allforms)-1)
			form = allforms[tempform]
	if (form != None) and (form != ''):
		shift = 1
		#if (subtype == 'none') and (situation=='none'):
		#	if not ShiftMatchTarget(char,target, form): shift = 0
		if shift:
			AISupport_DoFunctionTimed(char, 'ffx.OnNPCChange(%s,%s)'%(`char`,`form`), target, priority, time, desc='shapeshift '+`form`)
			return 1
			#AIDoGeneric(char,priority=priority,time=time,description='shapeshift '+ `form`)
			#ffx.OnNPCChange(char,form)
	return 0

def AIOpt_TShapeShift(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_WHOLEMAP)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_LOW)',
		'time': 'Opt_Convert(tactic,"time",5)',
		'category': 'pre',
	}		
		
def GetForms(char):
	row = ffx.FFX_ObjectGetAttr(char, 'shapeRow')
	entry = ffx.FFX_SHAPESHIFT_CUSTOM[int(row)]
	forms = []
	#temp = m25template.GetComplexTemplate(char)
	temp = js.Object_GetTemplate(char)
	if temp != entry[0]:
		forms.append(entry[0])
	for form in entry[4:]:
		if temp != form and form != '':
			forms.append(form)
	return forms
	
def ShiftMatchTarget(char, target, newform):
	st = GetAITargetTypes(newform, TARGET_ENEMY, default=None)
	#print 'Shiftmatch ' + newform + ' ' + `st`
	if st == None: return 1
	weight = Target_SubtypeWeight(target, st, char, None)
	#print 'SHIFTMATCHTARGET %s'%(`weight`)
	if randint(1,100) <= weight*10: return 1
	return 0
	
	
#------------------------------
# Summon
#------------------------------	
	
def TSummon(tactic_id, char, object, flt, user, pct=20, time=3, energy=50, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM, creature=None):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	if creature == None:
		if creature == None:
			creature = randint(1,3)
			summonses = int(ffx.FFX_ObjectGetAttr(char,'summonses'))
			if ((summonses&2)==0) & (creature == 2):
				creature = 1
			if ((summonses&4)==0) & (creature==3):
				creature = 1
	AISupport_DoFunctionTimed(char, 'ffx.OnSummon%d("%s","%s")'%(creature,target,char), target, priority, time, desc='summon vs.')
	return 1

def AIOpt_TSummon(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",67)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}
		

#------------------------------
# Super Leap
#------------------------------	
		
def AIOpt_TSuperLeap(char,tactic):
	return { 
		'exec': 'AIGenericFunctionOnTargetTimed',
		'power': 'ffx.OnLeapTo(target,char)',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_SHORT)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",0)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",1.5)',
		'situation': 'Opt_AddSituation(tactic,"not flying")',
		'category': 'move',
	}	

def AIOpt_TSuperLeapAway(char,tactic):
	return { 
		'exec': 'AIGenericTravelFunctionTimed',
		'power': 'ffx.OnLeapTo(target,char)',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_SHORT)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",0)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",2)',
		'category': 'flee',
	}		

#------------------------------
# Telekinesis
#------------------------------		
	
def TTK(tactic_id, char, object, flt, user, pct=50, time=1, energy=33, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	if len(ffx.getTiggots(char)) > 0:
		# check if the object will hit the user
		targetPos = js.Get_ObjectPos(target)
		myPos = js.Get_ObjectPos(char)
		tigs = ffx.getTiggots(char)
		for tig in tigs:
			tigPos = js.Get_ObjectPos(tig)
			dx, dy = tigPos[0]-targetPos[0], tigPos[1]-targetPos[1]
			dist = math.sqrt(dx*dx+dy*dy)
			dx, dy = dx/dist, dy/dist
			userOffset = (tigPos[0]-myPos[0])*dy-(tigPos[1]-myPos[1])*dx
			if (userOffset > -40) and (userOffset<40):
				AISupport_DoMove(char,target, 'move',0, (myPos[0]+100*dy,myPos[1]-100*dx,myPos[2]), myPos, priority, 1, 'move away from tiggots')
				return 0
		AISupport_DoFunction(char, 'ffx.OnThrowTo("%s","%s")'%(target,char), target, priority, time, desc='ttk throw')
		return 1
	else:
		bestObj=ffx.GetTKTarget(target,char)
		if bestObj != '':
			AISupport_DoFunction(char, 'ffx.OnLift("%s","%s")'%(bestObj,char), target, priority, time, desc='ttk lift')
			return 1
	return 0

def AIOpt_TTK(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",33)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}
								

def TTKGroupFlight(tactic_id, char, object, flt, user, pct=60, time=1, energy=0, mindist=0, maxdist=150, type='friend', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.OnTkGroupFlight(%s,%s)'%(`char`,`target`), target, priority, time, desc='ttk flight')
	return 1

def TTKGroupLand(tactic_id, char, object, flt, user, pct=80, time=1, energy=0, mindist=0, maxdist=300, type='friend', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0
	AISupport_DoFunction(char, 'ffx.OnTkLand(%s,%s)'%(`char`,`target`), target, priority, time, desc='ttk throw')
	return 1


#------------------------------
# Time Manipulation
#------------------------------	
	
def TTimeSlow(tactic_id, char, object, flt, user, pct=50, time=2, energy=50, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	try:
		if Mission_GetAttr('ffqtimeflow'): return 0
	except:
		return 0
	
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0		
	AISupport_DoFunction(char, 'ffx.OnFFqTimeSlow("%s","%s")'%(char,char), target, priority, time, desc='slow time')
	return 1

	
def AIOpt_TTimeSlow(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",66)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}

def TTimeTravel(tactic_id, char, object, flt, user, pct=50, time=2, energy=50, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	if ffx.FFX_ObjectGetAttr(char, 'ffqtimetravelTacticClone'): return 0  #already have a time clone?
	if char[0:5] == 'mobj_': return 0  #I am a time clone!
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0		
	AISupport_DoFunction(char, 'ffx.OnFFqTimeTravel("%s","%s")'%(char,char), target, priority, time, desc='time travel')
	return 1

	
def AIOpt_TTimeTravel(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",66)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}
	
	
#------------------------------
# Utility Belt
#------------------------------		
	


def AIGen_trickshot(char):
	return ['TTrickShot()',1]
	
def AIOpt_TTrickShot(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",0)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",2)',
		'category': 'post',
	}

#Note: it seems you can't have both trickshot and utility belt on the same character!
def TTrickShot_Init(char, items=None):
	utils=['ffx_arrow_bomb','ffx_arrow_smoke','ffx_arrow_net','ffx_arrow_punch','ffx_arrow_boomerang','ffx_arrow_tranq']
	if items:
		for x in items:
			if x in utils:
				ffx.equipffx_arrow(char, x)
		return
	
	#choose three at random
	for i in range(1,3):
		item = utils[randint(0,len(utils)-1)]
		utils.remove(item)
		ffx.equipffx_arrow(char,item)
	#remove any and all utility items on the map
	for i in range(0,6):
		obj='%s_u%d'%(char,i)
		if Object_Exists(obj): Object_Destroy(obj)

def TTrickShot(tactic_id, char, object, flt, user, pct=50, time=2, energy=0, mindist=0, maxdist=300, type='enemy', subytpe=None, situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM, items=None):
	if js.Object_Exists('%s_u%d'%(char,0)):
		TTrickShot_Init(char, items)
		return 0
	else:
		#tactics to use trick arrows
		if ffx.FFX_ObjectGetAttr(char,'cffx_arrow_bomb'):
			if randint(1,100) < 25:
				return AIGenericFunctionOnTargetTimed(tactic_id, char, object, flt, user, power='ffx.onThrowffx_arrow_bomb(target,char)', target=None, pct=100, time=2, energy=0, mindist=m25distance.DISTANCE_SHORT, maxdist=m25distance.DISTANCE_LONG, type=TARGET_ENEMY, subtype='not nearattackerally 100', situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM, desc='trickshot bomb arrow')
		if ffx.FFX_ObjectGetAttr(char,'cffx_arrow_smoke'):
			if randint(1,100) < 25:
				return AIGenericFunctionOnTargetTimed(tactic_id, char, object, flt, user, power='ffx.onThrowffx_arrow_smoke(target,char)', pct=100, time=2, energy=0, mindist=m25distance.DISTANCE_MELEE, maxdist=m25distance.DISTANCE_LONG, priority=TACTIC_PRIORITY_MEDIUM, subtype=['not nearattackerally 100','not blind'], desc='trickshot smoke arrow')
		if ffx.FFX_ObjectGetAttr(char,'cffx_arrow_net'):
			if randint(1,100) < 25:
				return AIGenericFunctionOnTargetTimed(tactic_id, char, object, flt, user, power='ffx.onThrowffx_arrow_net(target,char)', pct=100, time=2, energy=0, mindist=m25distance.DISTANCE_MELEE, maxdist=m25distance.DISTANCE_LONG, priority=TACTIC_PRIORITY_MEDIUM, subtype='not staticked', desc='trickshot net arrow')
		if ffx.FFX_ObjectGetAttr(char,'cffx_arrow_punch'):
			if randint(1,100) < 25:
				return AIGenericFunctionOnTargetTimed(tactic_id, char, object, flt, user, power='ffx.onThrowffx_arrow_punch(target,char)', pct=100, time=2, energy=0, mindist=m25distance.DISTANCE_MELEE, maxdist=m25distance.DISTANCE_LONG, priority=TACTIC_PRIORITY_MEDIUM, desc='trickshot punch arrow')
		if ffx.FFX_ObjectGetAttr(char,'cffx_arrow_boomerang'):
			if randint(1,100) < 25:
				return AIGenericFunctionOnTargetTimed(tactic_id, char, object, flt, user, power='ffx.onThrowffx_arrow_boomerang(target,char)', pct=100, time=2, energy=0, mindist=m25distance.DISTANCE_MELEE, maxdist=m25distance.DISTANCE_LONG, priority=TACTIC_PRIORITY_MEDIUM, desc='trickshot boomerang arrow')
		if ffx.FFX_ObjectGetAttr(char,'cffx_arrow_tranq'):
			if randint(1,100) < 25:
				return AIGenericFunctionOnTargetTimed(tactic_id, char, object, flt, user, power='ffx.onThrowffx_arrow_tranq(target,char)', pct=100, time=2, energy=0, mindist=m25distance.DISTANCE_MELEE, maxdist=m25distance.DISTANCE_LONG, priority=TACTIC_PRIORITY_MEDIUM, desc='trickshot tranq arrow')
	return 0		


def AIGen_utilitybelt(char):
	return ['TUtilityBelt()',1]

def AIOpt_TUtilityBelt(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",0)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
		'category': 'post',
	}

def TUtilityBelt_Init(char, items=None):
	utils=['bomb','mine','medpack','tracker','laser','cuffs']
	#if the user has supplied the items, use them
	if items:
		for x in items:
			if x in utils:
				template='ffx_util_%s'%(item)
				exec 'ffx.equip%s(char)'%(template)
		return
	#otherwise choose three at random
	for i in range(1,3):
		item = utils[randint(0,len(utils)-1)]
		utils.remove(item)
		template='ffx_util_%s'%(item)
		exec 'ffx.equip%s(char)'%(template)
	#remove any and all utility items on the map
	for i in range(0,6):
		obj='%s_u%d'%(char,i)
		if Object_Exists(obj): Object_Destroy(obj)
	
	
def TUtilityBelt(tactic_id, char, object, flt, user, pct=50, time=2, energy=0, mindist=0, maxdist=300, type='enemy', subytpe=None, situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM, items=None):
	if js.Object_Exists('%s_u%d'%(char,0)):
		TUtilityBelt_Init(char, items)
		return 0
	else:
		#heal thyself!
		if InSituation(char,'irradiated or acid_burnt'):
			return AIGenericFunctionOnTargetTimed(tactic_id, char, object, flt, user, power='ffx.onMedPack(char,char)', pct=100, time=2, energy=0, mindist=0, maxdist=m25distance.DISTANCE_WHOLEMAP, priority=TACTIC_PRIORITY_HIGH, desc='utility medpack on self')
		#check for mines
		if ffx.FFX_ObjectGetAttr(char,'mdrop'):
			#character has dropped mines, check to see if anyone is in range
			fire = 1
			for i in range (0,4):
				name='%sm%d'%(char,i)
				if Object_Exists(name):
					x,y,z = js.Get_ObjectPos(name)
					persons = AI_GetGrid(x, y, distance=2)
					if (persons==[]) or (filter(lambda x,c=char: m25team.AreAllies(x,c),persons) != []):
						fire = 0
			if fire:
				AISupport_DoFunctionTimed(char, 'ffx.OnDetonate(char,char)', None, priority, time, desc='detonate mines')
				return 1
		#tactics to use utility belt items
		if ffx.FFX_ObjectGetAttr(char,'smokebombs'):
			if randint(1,100) < 25:
				return AIGenericFunctionOnTargetTimed(tactic_id, char, object, flt, user, power='ffx.onThrowSmokeBomb(target,char)', target=None, pct=100, time=2, energy=0, mindist=m25distance.DISTANCE_SHORT, maxdist=m25distance.DISTANCE_LONG, type=TARGET_ENEMY, subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM, desc='utility smoke bomb')
		if ffx.FFX_ObjectGetAttr(char,'lasers'):
			if randint(1,100) < 25:
				return AIGenericFunctionOnTargetTimed(tactic_id, char, object, flt, user, power='ffx.onLaserTorch(target,char)', pct=100, time=2, energy=0, mindist=m25distance.DISTANCE_MELEE, maxdist=m25distance.DISTANCE_MEDIUM, priority=TACTIC_PRIORITY_MEDIUM, desc='utility laser torch')
		if ffx.FFX_ObjectGetAttr(char,'cuffs'):
			if randint(1,100) < 25:
				return AIGenericFunctionOnTargetTimed(tactic_id, char, object, flt, user, power='ffx.onHandCuff(target,char)', pct=100, time=2, energy=0, mindist=0, maxdist=m25distance.DISTANCE_MELEE, priority=TACTIC_PRIORITY_MEDIUM, desc='utility handcuffs')
		if ffx.FFX_ObjectGetAttr(char,'meds'):
			if randint(1,100) < 25:
				return AIGenericFunctionOnTargetTimed(tactic_id, char, object, flt, user, power='ffx.onMedPack(target,char)', pct=100, time=2, energy=0, mindist=0, maxdist=m25distance.DISTANCE_LONG, priority=TACTIC_PRIORITY_HIGH, type=TARGET_ALLY, subtype='acid_burnt or irradiated', desc='utility medpack')
		if ffx.FFX_ObjectGetAttr(char,'scans'):
			if randint(1,100) < 25:
				return AIGenericFunctionOnTargetTimed(tactic_id, char, object, flt, user, power='ffx.OnUtilScan(target,char)', pct=100, time=2, energy=0, mindist=0, maxdist=m25distance.DISTANCE_VISIBLE, priority=TACTIC_PRIORITY_MEDIUM, desc='utility scan')
		if ffx.FFX_ObjectGetAttr(char,'mines'):
			if randint(1,100) < 25:
				return AIGenericFunctionOnTargetTimed(tactic_id, char, object, flt, user, power='ffx.OnDropMine(target,char)', pct=100, time=2, energy=0, mindist=0, maxdist=m25distance.DISTANCE_SHORT, priority=TACTIC_PRIORITY_MEDIUM, desc='utility limpett mine')
	return 0		
		

			

#------------------------------
# Weather Control
#------------------------------				
		
def TWeatherControlHail(tactic_id, char, object, flt, user, pct=50, time=2, energy=50, mindist=0, maxdist=300, type='enemy', subtype=None, situation=None, role=None, priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0		
	if ffx.FFX_ObjectGetAttr(char,'storm') == 0:
		AISupport_DoFunction(char, 'ffx.OnStorm("%s","%s")'%(char,char), target, priority, time, desc='summon storm')
		return 1
	elif ffx.FFX_ObjectGetAttr(char,'storm') == 1:
		AISupport_DoFunction(char, 'ffx.OnHailStorm("%s","%s")'%(char,char), target, priority, time, desc='summon hail')
	else: return 0
	
def AIOpt_TWeatherControlHail(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",66)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"alone 100")',
		'time': 'Opt_Convert(tactic,"time",3)',
	}	
	
def TWeatherControlLightning(tactic_id, char, object, flt, user, pct=50, time=2, energy=50, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0		
	if ffx.FFX_ObjectGetAttr(char,'storm') == 0:
		AISupport_DoFunction(char, 'ffx.OnStorm("%s","%s")'%(char,char), target, priority, time, desc='summon storm')
		return 1
	elif ffx.FFX_ObjectGetAttr(char,'storm') == 1:
		AISupport_DoFunction(char, 'ffx.OnLightningStorm("%s","%s")'%(char,char), target, priority, time, desc='summon lightning')
	else: return 0
	
	
def AIOpt_TWeatherControlLightning(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",66)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"alone 100")',
		'time': 'Opt_Convert(tactic,"time",3)',
	}		

def TWeatherControlRain(tactic_id, char, object, flt, user, pct=50, time=2, energy=50, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_MEDIUM):
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0		
	if ffx.FFX_ObjectGetAttr(char,'storm') == 0:
		AISupport_DoFunction(char, 'ffx.OnStorm("%s","%s")'%(char,char), target, priority, time, desc='summon storm')
		return 1
	elif ffx.FFX_ObjectGetAttr(char,'storm') == 1:
		AISupport_DoFunction(char, 'ffx.OnRainStorm("%s","%s")'%(char,char), target, priority, time, desc='summon rain')
	else: return 0
	
	
def AIOpt_TWeatherControlRain(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",66)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"alone 100")',
		'time': 'Opt_Convert(tactic,"time",3)',
	}		


#def AIGen_weathercontrol(char):
#	return [["TWeatherControlHail()",1],["TWeatherControlLightning()",1]]

	
	
#------------------------------
# Temporary Invulnerability
#------------------------------			
		
def TTempInvulnerability(tactic_id, char, object, flt, user, pct=50, time=3, energy=80, mindist=0, maxdist=300, type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_HIGH):
	if ffx.FFX_ObjectGetAttr(char, 'ffqtempinvuln-on') > 0: return 0
	target = Target_GetCurrentTarget(char,type,subtype)
	if not AISupport_CheckConditions(char, target, Target_GetDistance(char,type), pct, priority, mindist, maxdist, subtype, situation, role): 
		return 0		
	AISupport_DoFunction(char, 'ffx.OnFFqTempInvulnOn("%s","%s")'%(char,char), target, priority, time, desc='temporary invulnerability')
	return 1

def AIOpt_TTempInvulnerability(char,tactic):
	return { 
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'energy': 'Opt_Convert(tactic,"energy",66)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'time': 'Opt_Convert(tactic,"time",3)',
	}
		
	

			
#####################################################################################
#
# Targeting
# 
# Description: find a target of a given type for a given character
#
#####################################################################################

#-------------------------------------------------------------------------------------
#
# Find a valid target for a tactic
#
#-------------------------------------------------------------------------------------

def FindTacticTarget(tactic_id=0,char='',mindist=0,maxdist='visible',energy=0,pct=20,priority=TACTIC_PRIORITY_LOW,role='none',situation='none',reaction=TEAM_ENEMIES,target_trait='none',target=None,default=None,debug=0):
	#start = ff.clock()
	mindist = m25distance.DistanceConvert(mindist)
	maxdist = m25distance.DistanceConvert(maxdist)
	#print 'findtactictarget ' + `maxdist`
	#mdist = m25distance.DISTANCE_LONG
	#if maxdist > mdist:
	#	mdist = maxdist
	#mark1 = ff.clock()
	#print 'findtactictarget distance convert ' + `mark1-start`
	#target= FindTarget(char, reaction=reaction, mindist=0, maxdist=mdist, default=None)
	target = Target_GetCurrentTarget(char, reaction)
	#mark2 = ff.clock()
	#print 'findtactictarget find target ' + `mark2-mark1`
	if target == None:
		if debug:
			m25report.Report(AI_DEBUG_CHANNEL,'findtactictarget ' + char  + ' no target')
		return default
	if not AITacticCheck(tactic_id, char, energy=energy, pct=pct, priority=priority, role=role, situation=situation,target=target,debug=debug):
		return
	#mark3 = ff.clock()
	#print 'findtactictarget tactic check ' + `mark3-mark2`
	dist = GetDistance(char,target)
	maxdist = maxdist + AdjustTargetDistance(target)
	#mark4 = ff.clock()
	#print 'findtactictarget adjust distance ' + `mark4-mark3`
	if (dist >= mindist) and (dist <= maxdist):
		if IsSubtype(target,target_trait,char,debug=debug):
			#mark5 = ff.clock()
			#print 'findtactictarget subtype check ' + `mark5-mark4`
			if debug:
				m25report.Report(AI_DEBUG_CHANNEL,'findtactictarget ' + char + ' found target ' + target)
			return target
	if debug:
		m25report.Report(AI_DEBUG_CHANNEL,'findtactictarget ' + char + ' bad range or subtype')
	return default

def AdjustTargetDistance(target):
	if target == None: return 0
	try:
		if js.Object_GetClass(target) & js.OC_BUILDING:
			return 150
		if js.Object_GetClass(target) == 167908581:  #tanks
			return 50
	except:
		return 0
	return 0
	
#-------------------------------------------------------------------------------------
#
# Find Target
#
#-------------------------------------------------------------------------------------	


def FindTarget(object, reaction=TARGET_ENEMY, mindist=0, maxdist='visible', subtype=[], default=[], nocache=0, force=0):
	#times = ff.time()
	#return default #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	#start = ff.clock()
	newreaction = Target_CheckForAlteredReaction(object, reaction)
	#if newreaction != reaction: print object, reaction, newreaction
	#mark1 = ff.clock()
	#print 'FT altered ' + `mark1-start`
	if subtype == []:  
		subtype = GetAITargetTypes(object,reaction,default=[])
	#mark2 = ff.clock()
	#print 'FT types ' + `mark2-mark1`
	# try cached
	target = None
	#if not force:
 	#	target = Target_FindCached(object, newreaction,mindist,maxdist)
	#DMC added Target!=None below - the cached values dont respect maxdist, so sometimes
	#the caching only occurs on low maxdist calls and high maxdist ones never get calculated again
	#mark3 = ff.clock()
	#print 'FT cache ' + `mark3-mark2`
 	#if target != '__recalc' and target!=None:
 	#	return target
 	# try target list
 	target = Target_CheckList(object, newreaction, mindist, maxdist, subtype)
 	#mark4 = ff.clock()
	#print 'FT checklist ' + `mark4-mark3`
 	if target[0] != None:
 		if not nocache:
 			Target_Cache(object, reaction, target)  #!!! cache them under the original reaction
 		return target[0]
	# default to checking everything!
	target = Target_CheckAll(object, newreaction, mindist, maxdist, subtype)
	#mark5 = ff.clock()
	#print 'FT all ' + `mark5-mark4`
	if not nocache:
		Target_Cache(object, reaction, target) #!!! cache them under the original reaction
	#timee=ff.time()
	#end = ff.clock()
	#print 'FT total ' + `end-start`
	if target != None:
		#print 'target find time (all) %s : %f'%(object,(timee-times))
		return target[0]
	#print 'target find time (none found) %s : %f'%(object,(timee-times))
	return default



#-------------------------------------------------------------------------------------
#
# Find Cached Target
#
#-------------------------------------------------------------------------------------	

ai_target_cache = {TARGET_ALLY:{}, TARGET_ENEMY:{}, TARGET_NEUTRAL:{}, TARGET_ANY:{}}

def Target_GetDistance(char, reaction=TARGET_ENEMY):
	global ai_target_cache
	if ai_target_cache.has_key(reaction) and ai_target_cache[reaction].has_key(char):
		return ai_target_cache[reaction][char]['distance']
	return 100000

def Target_GetCurrentTarget(object, reaction=TARGET_ENEMY, subtype=None):
	global ai_target_cache
	if reaction == TARGET_ANY:
		data = Target_CheckAll(object, reaction, mindist=0, maxdist=m25distance.DISTANCE_VISIBLE, subtype=subtype)
		Target_Cache(object,reaction,data)
	if ai_target_cache.has_key(reaction) and ai_target_cache[reaction].has_key(object):
		return ai_target_cache[reaction][object]['target']
	return None

	
def Target_FindCached(object, reaction,mindist,maxdist):
	if m25timestamp.Timestamp_ObjectAttrOutdated(object, TARGET_CACHED+reaction, delta=TARGET_CACHED_TIMER):
		# cached value has expired
		Target_RemoveCache(object, reaction)
		return '__recalc'
	else:
		#print 'Target using cached value ' + object
		mind = DistanceConvert(mindist,object)
		maxd = DistanceConvert(maxdist,object)
		t = GetObjectVar(object, TARGET_CACHED+reaction,default=None,nosave=1)
		if Target_Valid(object,t):
			return t
		return None
		
def Target_Cache(object, reaction, target):
	global ai_target_cache
	m25report.Report(TARGET_CHANNEL,`object` + ' ' + `reaction` + ' ' + `target`)
	if target[0] != None:
		ai_target_cache[reaction][object] = {'target': target[0], 'distance':Target_AdjustTargetDistance(target[0],target[1])}
	else:
		ai_target_cache[reaction][object] = {'target':None, 'distance':10000}


def Target_AdjustTargetDistance(target, distance):
	if target == None: return distance
	try:
		if js.Object_GetClass(target) & js.OC_BUILDING:
			distance = distance - 150
		if js.Object_GetClass(target) == 167908581:  #tanks
			distance = distance - 50
		if distance < 0: distance = 10
		return distance
	except:
		return distance
	return distance

			
def Target_RemoveCache(object, reaction):
	old_target = GetObjectVar(object, 	TARGET_CACHED+reaction, default=None,nosave=1)
	if old_target != None:
		RemoveTargettedBy(object, reaction, old_target)
	
def IsInCombat(object):
	target = GetObjectVar(object, TARGET_CACHED+TARGET_ENEMY,default=None,nosave=1)
	if target == None:
		return 0
	return m25distance.GetDistance(object,target) <= m25distance.DISTANCE_LONG
	
	
def AddTargettedBy(char, reaction, target):
	if target == None: return
	targetted_by_list = GetObjectVar(target, 'targetted_by_' + reaction, default = None,nosave=1)
	if targetted_by_list == None:
		targetted_by_list = [char]
	else:
		targetted_by_list.append(char)
	#print target, targetted_by_list
	SetObjectVar(target, 'targetted_by_' + reaction, targetted_by_list,nosave=1)
	
def RemoveTargettedBy(char, reaction, target):
	if target == None: return
	targetted_by_list = GetObjectVar(target, 'targetted_by_' + reaction, default = None,nosave=1)
	if targetted_by_list == None:
		return
	if char in targetted_by_list:
		try:
			targetted_by_list.remove(char)
		except:
			pass
	#print target, targetted_by_list
	SetObjectVar(target, 'targetted_by_' + reaction, targetted_by_list,nosave=1)
		
def GetTargettedByList(target, reaction):
	if target == None: return None
	return GetObjectVar(target, 'targetted_by_' + reaction, default = None,nosave=1)
	
def GetNumTargettedBy(target, reaction):	
	if target == None: return 0
	tbl = GetObjectVar(target, 'targetted_by_' + reaction, default = None,nosave=1)
	if tbl != None:
		return len(tbl)
	return 0
	
def AIIsBeingTargetted(target, reaction=TARGET_ENEMY):
	print 'aiisbeingtargetted',target, reaction, GetObjectVar(target, 'targetted_by_' + reaction, default = [],nosave=1)
	if target == None: return 0
	if GetObjectVar(target, 'targetted_by_' + reaction, default = [],nosave=1) == []: return 0
	return 1
	
	
def GetTargettedByPosition(char, reaction, target):
	if target == None: return 0
	tbl = GetObjectVar(target, 'targetted_by_' + reaction, default = None,nosave=1)
	if tbl != None:
		num = 0
		for t in tbl:
			if t == char:
				#print 'Targetted by position ' + `num`
				return num
			num = num +1
	return 0
	
	
#-------------------------------------------------------------------------------------
#
# Find target from targeting list
#
#-------------------------------------------------------------------------------------			
	
ai_specialtargetlist = {TARGET_ENEMY:{}, TARGET_ALLY:{}, TARGET_NEUTRAL:{}, TARGET_ANY:{}}

def Target_CheckList(object, reaction, mindist, maxdist, subtype):
	tlist = Target_GetTargetList(object, reaction)
	if tlist == None:
		print 'Target_Checklist: target list empty!'
		return [None,10000]
	highest = TargetList_FindHighest(object, tlist)
	num = highest
	while num >= 0:
		sublist = TargetList_FindNum(object, tlist, num)
		if sublist != []:
			target = skXMapInfo.MapInfo_GetNearest(object,taObjects = sublist,tsFilter=TargetList_Filter)
			#if Target_IsValid(target, object): 
			return [target, m25distance.GetDistance(object,target)]
		num = num - 1
	return [None, 10000]

def TargetList_Filter(object):
	return js.Object_Exists(object) and js.Object_IsAlive(object)	
	
def TargetList_FindHighest(char, tlist):
	num = 0
	priority = -1
	for t in tlist:
		val = t[1]	
		if val > priority:
			priority = val
		num = num + 1
	return priority
			
def TargetList_FindNum(char, tlist, num):
	current = []
	for t in tlist:
		val = t[1]
		if val == num:
			if type(t[0]) == types.ListType:
				for char in t[0]:
					current.append(char)
			else:
				current.append(t[0])
	return current
		

def Target_GetTargetList(object, reaction):
	global ai_specialtargetlist
	if ai_specialtargetlist[reaction].has_key(object):
		return ai_specialtargetlist[reaction][object]
	return []


def Target_AddTarget(char, object, reaction, priority=TARGET_PRIORITY_LOW):
	tlist = Target_GetTargetList(char,reaction)
	newlist = []
	if tlist == None:
		if type(object) == types.ListType:
			for o in object:
				newlist.append([o,priority])
		else:
			newlist = [[object,priority]]
	else:
		newlist = tlist[0:]
		if type(object) == types.ListType:
			for o in object:
				newlist.append([o,priority])
		else:
			newlist.append([object,priority])
	Target_SetTargetList(char,reaction,newlist)
	newtarget = Target_CheckList(char,reaction,0,m25distance.DISTANCE_VISIBLE,None)
	m25report.Report(TARGET_CHANNEL, '%s targetting %s with reaction %s'%(char,`newtarget`,reaction))
	if newtarget[0] != None:
		Target_Cache(char,reaction,newtarget)
	
def Target_SetTargetList(object, reaction, tlist):
	global ai_specialtargetlist
	ai_specialtargetlist[reaction][object] = tlist

def Target_ClearTargets(object, reaction):
	global ai_specialtargetlist
	ai_specialtargetlist[reaction][object] = []

	
#-------------------------------------------------------------------------------------
#
# Find target from all dynamic and static objects
#
#-------------------------------------------------------------------------------------		

target_teams = {}

def Target_CheckAll(object, reaction=TARGET_ENEMY, mindist=0, maxdist=m25distance.DISTANCE_VISIBLE, subtype=None):
	if object == None: 
		m25report.Report(ERROR_CHANNEL, 'Target_CheckAll called for non-existent object.')
		return [None,10000]
	if (m25team.GetTeam(object) == m25team.TEAM_NONCOMBATANT) or (js.Object_GetClass(object) & js.OC_CIVILIAN) : return [None,10000]
	#if not ff.MapInfo.has_key(object): return [None,10000]
	
#	if js.Object_GetPrimaryState(object) == js.PCSTATE_ENRAGED: reaction=TARGET_ANY
	
	if reaction == TARGET_ENEMY:
		target = [None, 10000]
		#try the nearest enemy first
		if Target_NeedFullCheck(object, subtype):
			m25report.Report(TARGET_CHANNEL,'%s needs full check'%(object))
			target = Target_FindValidEnemy(object, subtype)
			m25report.Report(TARGET_CHANNEL,'%s targetting %s after full check.'%(object,repr(target[0])))
		else:
			target = AI_GetNearestEnemy(object)
			#target = ff.MapInfo[object]['nearestEnemy']
		#if no target is close, try a farther target
		if target[0] == None:
			if maxdist > m25distance.DISTANCE_VISIBLE:
				target = Target_FindValidEnemy(object,subtype, distance=maxdist)
				#target = ff.MapInfo[object]['nearestEnemyElsewhere']
		else:			
			#if the nearest enemy is not the best for some reason, do a more lengthy search for a better target
			if not Target_Filter(target[0]):
				target = Target_FindValidEnemy(object,subtype)
			#if the results of the search aren't good, go back to the nearest enemy
			if not Target_Filter(target[0]):
				if AI_GetNearestEnemy(object)[0] != None:
				#if ff.MapInfo[object]['nearestEnemy'][0] != None:
					target = AI_GetNearestEnemy(object)
					#target = ff.MapInfo[object]['nearestEnemy']
		if Target_IsValid(target[0], object): return target
		else: return [None, 10000]
	elif reaction == TARGET_ALLY:
		return AI_GetNearestAlly(object)
		#return ff.MapInfo[object]['nearestAlly']
	elif reaction == TARGET_ANY:
		tlist = AddToList(AI_GetNearDynamic(object), AI_GetNearObjects(object))
		target = skXMapInfo.MapInfo_GetNearest(object, taObjects=tlist, tsFilter=lambda x, subtype=subtype, char=object: IsSubtype(x,subtype,char))
		if target != None:
			return [target, skXAngles.fSkXGetDistance (js.Get_ObjectPos(object), js.Get_ObjectPos(target), "xyz")]
		else:
			return [None, 10000]
	else:
		return [None, 10000]

		
def Target_TryCheckAgain(object, target):
	if target == None: return 1
	return (not Target_Filter(target)) or AreAllies(object,target) or AreNeutral(object,target) or IsAIDisabled(target)
			
def Target_NeedFullCheck(object, subtype):
	if ((subtype != None) and ('building' in subtype)) or AI_GetAttacksBuildings(object):
		return 1
	if ((subtype!=None) and ('object' in subtype)) or AI_GetAttacksObjects(object):
		return 1
	
def Target_FindValidEnemy(object, subtype, distance=m25distance.DISTANCE_VISIBLE):
	if distance > m25distance.DISTANCE_VISIBLE:
		tlist = AI_TupleToList(filter(lambda x, char=object: m25team.AreEnemies(x,char), js.Mission_GetDynamicObjects()))
	else:
		tlist = AI_GetNearEnemies(object)
	#tlist = skXMapInfo.MapInfo_GetNearbyEnemies(object)
	if ((subtype != None) and ('building' in subtype)) or AI_GetAttacksBuildings(object):
		tlist = AddToList(tlist,AI_GetNearBuildings(object))
	if ((subtype!=None) and ('object' in subtype)) or AI_GetAttacksObjects(object):
		tlist = AddToList(tlist, AI_GetNearObjects(object))
	if ((subtype!=None) and ('civilian' in subtype)) or AI_GetAttacksCivilians(object):
		tlist = AddToList(tlist, AI_GetNearCivilians(object))
	target = skXMapInfo.MapInfo_GetNearest(object, taObjects=tlist, tsFilter=lambda x, char=object: Target_Filter(x,char))
	if target != None:
		return [target, skXAngles.fSkXGetDistance (js.Get_ObjectPos(object), js.Get_ObjectPos(target), "xyz")]
	else:
		return [None, 10000]

def AI_TupleToList(t):
	newlist = []
	for e in t:
		newlist.append(e)
	return newlist
			
def Target_Filter(char, fighter=None):
	#print char, fighter
	if char == None: return 0
	if not js.Object_Exists(char): return 0
	if not js.Object_IsAlive(char): return 0
	if not js.Object_GetAttr(char,'physical'): return 0
	if IsAIDisabled(char) and ST_char(char):
		if ffx.FFX_ObjectGetAttr(char,'ffactivedefense',0)==0:
			return 0
	if AI_GetIgnore(char) or (fighter and AI_GetIgnore(fighter, char)): return 0
	#if AIIsBeingTargetted(char): return 0
	#if not (js.Object_GetClass(char) & ( ffx.FFX_CHARACTER |  js.OC_CONTROLLABLE |  js.OC_POLICE |  js.OC_CIVILIAN |  js.OC_MINION |  js.OC_VILLIAN)):
	#	return 0
	if js.Object_GetSecondaryStates(char) & (js.SCSTATE_INVISIBLE |js.SCSTATE_IRRADIATED | js.SCSTATE_ACID_BURNT):
		return 0
	if js.Object_GetPrimaryState(char) in [js.PCSTATE_FROZEN,js.PCSTATE_INTEROGATE,js.PCSTATE_BLANK,js.PCSTATE_DIMENSION_SHIFT,js.PCSTATE_PANICKED,js.PCSTATE_STUNNED,js.PCSTATE_STATIC,js.PCSTATE_TOAD,js.PCSTATE_OBJECT,js.PCSTATE_POLYMORPHED,js.PCSTATE_HYPNOTISED]:
		return 0
	if (ffx.FFX_ObjectGetAttr(char,'ffx_sleep',0)==1) | (ffx.FFX_ObjectGetAttr(char,'ffs_knockedout',0)==1):
		return 0
	if char[:3] == 'ffx': return 0
	if js.Object_GetClass(char) & js.OC_TERRAIN: return 0
	#if len(GetTargettedByList(target, TARGET_ENEMY)) > 0: return 0
	return 1

def Target_IsValid(char, fighter=None):
	if char == None: return 0
	if not js.Object_Exists(char): return 0
	if not js.Object_IsAlive(char): return 0
	if not js.Object_GetAttr(char,'physical'): return 0
	if (fighter and AI_GetIgnore(fighter, char)) or (AI_GetIgnore(char)): return 0
	#if char[:3] == 'ffx': return 0
	if IsAIDisabled(char) and ST_char(char):
		if ffx.FFX_ObjectGetAttr(char,'ffactivedefense',0)==0:
			return 0
	if (m25team.GetTeam(char)==0) and (ST_char(char)): return 0
	return 1
 	
# def AI_GetNearBuildings(char,distance=None):
# 	objects = skXMapInfo.MapInfo_GetObjects(js.Get_ObjectPos(char),distance)
# 	if not objects: return []
# 	return_list = []
# 	for o in objects:
# 		if js.Object_GetClass(o) & js.OC_BUILDING:
# 			return_list.append(o)
# 	#print return_list
# 	return return_list
# 	
# def AI_GetNearObjects(char):
# 	objects = skXMapInfo.MapInfo_GetObjects(js.Get_ObjectPos(char))
# 	if not objects: return []
# 	#print objects
# 	return objects
 	
def Target_CheckAll2(object, reaction, mindist, maxdist, subtype): 	
	#print 'Target_CheckAll2 ' + object + ' ' + `maxdist`
	global valid_characters
	tlist = []
	all = []
	try:
		#all = valid_characters
		if m25distance.DistanceConvert(maxdist) > 400:
			all = Mission_GetDynamicObjects()
		else:
			all = skXMapInfo.MapInfo_GetFighters(object)
		if not all: all = []
	except:
		all = Mission_GetDynamicObjects()
	check_objects = 0
# 	if ('object' in subtype) or ('building' in subtype):
# 	#all = Mission_GetDynamicObjects()
# 		check_objects = 1
# 		all = Mission_GetObjects()
		
	#val = Timestamp_GetMissionCachedAttr(`GetTeam(object)` +'_'+ `reaction` + '_cache',delta=1)
	val = None
	if val == None or check_objects:
		for c in all:
			if GetTeam(c) == GetTeam(object):
				if Target_AgainstOwnTeam(c):
					if TeamsReactionOfType(object,c,Target_FlipReaction(reaction)):
						tlist.append(c)
				else:
					if TeamsReactionOfType(object, c, reaction):
						tlist.append(c)
			else:
				if TeamsReactionOfType(object, c, reaction):
					tlist.append(c)
	else: 
		tlist = val
	
	if reaction == TARGET_ENEMY:
		#print 'reaction enemy'	
		if ('building' in subtype) or AI_GetAttacksBuildings(object):
			tlist = AddToList(tlist,AI_GetNearBuildings(object))
		#print AI_GetAttacksObjects(object)
		#if AI_GetAttacksObjects(object):
		#	print 'attacks objects'
		if ('object' in subtype) or AI_GetAttacksObjects(object):
			tlist = AddToList(tlist, AI_GetNearObjects(object))	
		
	#if not check_objects:
	#	Timestamp_SetMissionCachedAttr(`GetTeam(object)` +'_'+ `reaction` + '_cache', tlist)
	#print object + '(' + reaction + ') -> ' + `tlist`
	if len(tlist) == 1:
		return tlist[0]
 	target = Target_GetBestMatch(object, tlist, mindist, maxdist, subtype)
 	#print target
 	return target

def Target_AgainstOwnTeam(object):
	if Object_Exists(object):
		if js.Object_GetPrimaryState(object) == js.PCSTATE_HYPNOTISED:
			return 1
		if js.Object_GetPrimaryState(object) == js.PCSTATE_POLYMORPHED:
			return 1
		if js.Object_GetSecondaryStates(object) & js.SCSTATE_POSSESSED:
			return 1
		if js.Object_GetSecondaryStates(object) & js.SCSTATE_ZOMBIE:
			return 1
	return 0
 	
def Target_UpdateTeams(event):
	global valid_characters
	valid_characters = []
	chars = js.Mission_GetDynamicObjects()
	for c in chars:
		if js.Object_GetAttr(c,'physical') and ((Object_GetClass(c) & ffx.FFX_CHARACTER) or (Object_GetClass(c) & 167777508)):
			valid_characters.append(c)
	js.Event_CancelSinks('Target_UpdateTeams')
	m25event.SetTimer(1,'Target_UpdateTeams')
		
	
#-------------------------------------------------------------------------------------
#
# Get best target match from a list
#
#-------------------------------------------------------------------------------------	
	
def Target_GetBestMatch(object, tlist, mindist, maxdist, subtype):
	#m25report.Report(TARGET_CHANNEL,object + ' get best match ' + `tlist` + ' with subtype ' + `subtype`)
	mind = DistanceConvert(mindist,object)
	maxd = DistanceConvert(maxdist,object)
	current = None
	curdist = m25distance.DISTANCE_OUTOFRANGE
	curmatch = 0
	size = len(tlist)
	for t in tlist:
		match = 0
		if (object != t) and Target_Valid(object,t):
			#print t + ' valid'
			dist = GetDistance(object,t)
			if dist >= mind and dist <= maxd:
				#print t + ' in range'
				match = Target_SubtypeWeight(t,subtype,object,current,size) + Target_DistanceWeight(dist) + Target_DecoyWeight(t)
				#print 'Target_GetBestMatch ' + object + ' ' + t + ' ' + `match` + ' : ' + `dist` + ' (' + `maxdist` +')'
				if curmatch < match:
					current = t
					curmatch = match
					curdist = dist
				if curmatch == match:
					if dist < curdist:
						current = t
						curdist = dist
						curmatch = match
	return current
	
def Target_SubtypeWeight(char, subtype, target, current,size=0):
	if size > 4: return 0  #when the number of enemies is large, don't worry so much about choosing the 'best' target
	#start = ff.clock()
	num = 0
	if type(subtype) == types.ListType:
		for s in subtype:
			sub = s
			weight = 1
			if type(s) == types.ListType:
				weight = s[1]
				sub = s[0]
			if (IsSubtype(char,sub,target) or Target_CheckSpecialWeight(char,sub,current)):
				num = num + weight
	else:
		num = IsSubtype(char,subtype,target)
	#print 'Target_Subtypeweight ' + target + ' ' + char + ' = ' + `num`
	#end = ff.clock()
	#print 'Target_SubtypeWeight ' + `end-start`
	return num

def Target_CheckSpecialWeight(char,subtype,current):
	if subtype == 'fastest':
		if current==None:
			return 1
		return ST_fasterthan(current,char)
	if subtype == 'slowest':
		if current==None:
			return 1
		return ST_slowerthan(current,char)
	if subtype == 'strongest':
		if current==None:
			return 1
		return ST_strongerthan(current,char)
	if subtype == 'weakest':
		if current==None:
			return 1
		return ST_weakerthan(current,char)
	if subtype == 'highesthealth':
		if current==None:
			return 1
		return ST_higherhealththan(current,char)
	if subtype == 'lowesthealth':
		if current==None:
			return 1
		return ST_lowerhealththan(current,char)
	if subtype == 'highestenergy':
		if current==None:
			return 1
		return ST_higherenergythan(current,char)
	if subtype == 'lowestenergy':
		if current==None:
			return 1
		return ST_lowerenergythan(current,char)
	return 0
	
    
def Target_DistanceWeight(dist):
	if dist <= m25distance.DISTANCE_LONG:
		return 5
	return 0
	
def Target_DecoyWeight(char):
	if GetObjectVar(char,'_decoy',default=0,integer=1):
		#print 'decoy found'
		return 2
	return 0
	
def Target_Valid(char,target):
	#print 'Target_Valid checking ' + `target`
	if target == None:
		return 0
	if Object_Exists(target) and Object_IsAlive(target):
		if IsAIDisabled(target) and not ST_object(target): return 0
		if ST_decoy(target) and TeamsReactionOfType(char, target, TEAM_ALLIES):
			return 0
		if (not (js.Object_GetSecondaryStates(target) & js.SCSTATE_INVISIBLE)) and js.Object_GetAttr(target,'physical'):
			if not (js.Object_GetPrimaryState(target) in [js.PCSTATE_DIMENSION_SHIFT,js.PCSTATE_INTEROGATE,js.PCSTATE_TOAD,js.PCSTATE_OBJECT,js.PCSTATE_POLYMORPHED]):
				return 1
	return 0

	
#-------------------------------------------------------------------------------------
#
# Altering reactions to targets
#
#-------------------------------------------------------------------------------------	

def Target_CheckForAlteredReaction(object, reaction):
	pstate = js.Object_GetPrimaryState(object)
### updated so code checks to see if the char is hypnotized by the hypnosis state swap and if not then swap reactions - 01/24/08 - TMX
	if ffx.FFX_ObjectGetAttr(object,'hypnosistohypnosis',0)!=1:
		if pstate in [js.PCSTATE_HYPNOTISED,js.PCSTATE_POLYMORPHED]:
			return Target_FlipReaction(reaction)
	else:
		if pstate == js.PCSTATE_POLYMORPHED:
			return Target_FlipReaction(reaction)
	if pstate == js.PCSTATE_ENRAGED:
		i = randint(1,2)
		if i==1:
			return Target_FlipReaction(reaction)
		else:
			return reaction
### Took these lines out so the Zombies attack the Master's Enemies! - 02/08/07 - TMX
	sstate = js.Object_GetSecondaryStates(object)
#	if sstate & js.SCSTATE_ZOMBIE:
#		print object + ' is possessed or is a zombie - flipping target type'
#		ffx.FFX_ObjectSetAttr(object,'isZombie',1)
#		return Target_FlipReaction(reaction)
#	if sstate & js.SCSTATE_POSSESSED:
#		print object + ' is possessed - flipping target type'
#		ffx.FFX_ObjectSetAttr(object,'isPossessed',1)
#		return Target_FlipReaction(reaction)
	return reaction
			
def Target_FlipReaction(reaction):
	if reaction == TARGET_ALLY:
		return TARGET_ENEMY
	elif reaction == TARGET_ENEMY:
		return TARGET_ALLY
	else:
		return reaction
		
				

	

#####################################################################################
#
# Throwing
# 
# Description: a special tactic that allows characters to throw cars (and other objects) for fun and profit
#
# Special Note:  this makes use of a goal (Goal_Throw) to get the desired behaviour
#
#####################################################################################

def TThrow(tactic_id, char, string2, flt, user,pct=70,time=AI_STANDARD_MOVE_TIME,mindist=0,maxdist=m25distance.DISTANCE_LONG,type='enemy', subtype='none', situation='none', role='none', priority=TACTIC_PRIORITY_LOW+1, objdist=m25distance.DISTANCE_WHOLEMAP,mode='move', pickup_time=5, throw_time=5):
	if HasRole(char, 'throwing'):  #character already throwing something
		return
	#print 'trying throw'
	#target = FindTacticTarget(tactic_id, char, mindist=mindist, maxdist=maxdist, energy=0, pct=pct, priority=priority, role=role, situation=situation, reaction=type, target_trait=subtype)
	target = Target_GetCurrentTarget(char)
	distance = Target_GetDistance(char)
	if not AISupport_CheckConditions(char, target, distance, pct, priority, mindist, maxdist, subtype, situation, role):
		return 0
	#print `target`
	if target == None:
		return 0
	if not AIIsFlying(target) and (distance < m25distance.DISTANCE_SHORT):
		return 0
	object, dist = AI_GetNearestThrowable(char)
	#object = skXMapInfo.MapInfo_GetObjects(js.Get_ObjectPos(char), 'nearest', tfFilter=ff.ThrowFilter)
	#object = FindRandObjectInRange(char,mindist=0,maxdist=objdist,subtype=['throwable','object','not being_thrown','not terrain'],default=None)
	#object = FindNearestObjectOfType(char,subtype=['throwable','object','not being_thrown','not terrain'],default=None)
	#print `object`
	if object != None:
		#dist = m25distance.GetDistance(char,object)
		if DistanceCompare(dist,objdist,'<='):
			try:
				zdist = js.Get_ObjectPos(object)[2] - js.Get_ObjectPos(char)[2]
			except:
				return 0 #object was destroyed while we were looking at it
			if zdist > 40: #too high to get at easily
				return
			SetObjectVar(char,'aipickup_time',pickup_time)
			SetObjectVar(char,'aithrow_time',throw_time)
			AddRole(char,'throwing')
			AddRole(object,'being_thrown')
			m25timestamp.Timestamp_ObjectAttrSync(char, '_bad_throw_timer')
			AddGoal_Throw(char,object,mode,time,priority)
			Trigger_Move(char,'move',js.Get_ObjectPos(object),'',0,0)
			AIDoGeneric(char,priority=TACTIC_PRIORITY_HIGH,time=0.1,function='',description='preparing to throw')
			return 1

def ThrowFilter(obj):
	if not (js.Object_GetClass(obj) & ( ffx.FFX_CHARACTER |  js.OC_CONTROLLABLE |  js.OC_POLICE |  js.OC_CIVILIAN |  js.OC_MINION |  js.OC_VILLIAN)):
		if js.Object_GetAttr(obj,'throwable') and (not js.Object_GetClass(obj) & js.OC_TERRAIN):
			return not HasRole(obj,'being_thrown')
	return 0			
			
def ThrowPickup(char,obj):
	SetAIPriority(char,TACTIC_PRIORITY_CRITICAL)
	SetPerformingTactic(char)
	SetObjectVar(obj,'ai_pickingup',char)
	AddRole(obj,'being_thrown')
	SetObjectVar(char,'ai_thowobject',obj)
	SetObjectVar(char,'ai_pickupsuccess',0)
	Trigger_Pickup(char,obj,'ThrowPickupSuccess')
	pickup_time = GetObjectVar(char,'aipickup_time',default=5,nosave=1)
	#give the character two seconds to get into position before the pickup
	m25event.SetTimer(pickup_time+2,'ThrowPickupFailed',char,obj)
	Report(AI_CHANNEL,char + ' picking up ' + obj)		

def ThrowPickupSuccess(event):
	obj = event.object
	char = GetObjectVar(obj,'ai_pickingup')
	Report(AI_CHANNEL,char + ' pickup success on ' + obj)
	SetObjectVar(char,'ai_pickupsuccess',1)
	ThrowFindTarget(char,obj)
	
def ThrowPickupFailed(event):
	char, obj = event.object, event.string
	if GetObjectVar(char,'ai_pickupsuccess',default=0,integer=1) == 0:
		Report(AI_CHANNEL,char + ' pickup failed on ' + obj)
		ThrowCancel(char,obj)

def ThrowCancel(char, obj):
	RemoveRole(obj,'being_thrown')
	RemoveRole(char,'throwing')
	#cshelper.stun(char) #make them drop any object they might be carrying (otherwise it just vanishes)	
	m25event.SetTimer(0.1,'AITacticDone',char,user=TACTIC_PRIORITY_CRITICAL)
	m25event.SetTimer(0.1,'ThrowUnstun',char)

def ThrowCancelEvent(event):
	char = event.object
	obj = event.string
	ThrowCancel(char, obj)    
    
def ThrowUnstun(event):
    char = event.object
    #if cshelper.stunned(char):
     #   Object_SetPrimaryState(char, js.PCSTATE_NORMAL, 0) 
        
def ThrowFindTarget(char,obj):
	target=FindTarget(char,maxdist=m25distance.DISTANCE_WHOLEMAP,default=None)
	Report(AI_CHANNEL,char + ' finding target ' + `target`)
	if target == None:
		ThrowCancel(char,obj)
	target_pos = js.Get_ObjectPos(target)
	if m25distance.GetDistance(char,target) > 1000:
		ThrowCancel(char,obj) #too far away
	elif m25distance.GetDistance(char,target) > 300:
		Trigger_Move(char,'move',target_pos)
		m25event.SetTimer(1,'ThrowMoveToTarget',char,target)
	else:
		SetObjectVar(char,'ai_throwtargetpos',target_pos)
		m25event.SetTimer(0.2, 'ThrowTrackTarget',char,target)

def ThrowMoveToTarget(event):
	char = event.object
	target = event.string
	try:
		target_pos = js.Get_ObjectPos(target)
	except:
		ThrowCancel(char,target)
	Report(AI_CHANNEL,char + ' moving to target ' + `target`)
	pos = m25distance.GetDistance(char,target)
	if m25distance.GetDistance(char,target) > 1000:
		ThrowCancel(char,target) #too far away
	elif m25distance.GetDistance(char,target) > 300:
		Trigger_Move(char,'move',target_pos)
		m25event.SetTimer(1,'ThrowMoveToTarget',char,target)
	else:
		SetObjectVar(char,'ai_throwtargetpos',target_pos)
		m25event.SetTimer(0.05, 'ThrowTrackTarget',char,target)
		
def ThrowTrackTarget(event):
	char = event.object
	target = event.string
	throw_time = GetObjectVar(char,'aithrow_time',default=5,nosave=1)
	#predict = (throw_time/2)/0.05
	predict = 6
	Report(AI_CHANNEL,char + ' tracking target ' + `target`)
	old_pos = GetObjectVar(char, 'ai_throwtargetpos')
	target_pos = js.Get_ObjectPos(target) 
	travel_vector = ((target_pos[0] - old_pos[0])*predict, (target_pos[1]-old_pos[1])*predict,(target_pos[2]-old_pos[2])*predict) #predict throw_time seconds travel
	throw_pos = Vector_Add(target_pos,travel_vector)
	
	#throw_pos = ff.MapInfo[target]["projectedPos"]
	
	Trigger_Throw(char,(throw_pos[0],throw_pos[1],throw_pos[2]+20))
	Report(AI_CHANNEL,char  + ' throwing at ' + target)	
	throw_time = GetObjectVar(char,'aithrow_time',default=5,nosave=1)
	m25event.SetTimer(throw_time, 'ThrowCancelEvent', char, GetObjectVar(char, 'ai_thowobject',default='none'))


#helper functions

def GetCharData(char):
	cdata = {}
	if js.Object_GetAttr(char,'complex') == 0:
		return {}
	cdata = chardata.GetCharacterData(char)
	return cdata


	
def pownum(event):
	print 'object = ' + `event.object`
	print 'string = ' + `event.string`
	print 'user = ' + `event.user`
	print 'float = ' + `event.float`
	Event_RegisterSink(21,'powerhit',event.object,event.string,event.float,1,1)
	
def powerhit(event):
	print 'power hit'	
	print 'object = ' + `event.object`
	print 'string = ' + `event.string`
	print 'user = ' + `event.user`
	print 'float = ' + `event.float`
	
def powerpos(powername):
	power = datfiles.Campaign_ReadPowers()
	if datfiles._PowPos.has_key(powername):
		print powername + ' ' + `datfiles._PowPos[powername]`
	else:
		print powername + ' not found'
		
def trigpower(powername):
	js.Trigger_Power('hero_0','hero_1',powername,'pownum')
	js.Trigger_Move('hero_0','move',js.Get_ObjectPos('hero_0'),'',10,1)
	#js.AI_NullController('hero_0')
	powerpos(powername)


def FindThrowableObjects(event):
	#times = ff.time()
	#print 'm25ai.FindThrowableObjects()'
	global m25throwableobjects
	objects = M25GetObjects()
	m25throwableobjects = []
	for obj in objects:
		if Object_GetAttr(obj,'throwable') and not ST_building(obj) and not ST_char(obj) and (js.Object_GetTemplate(obj) not in ['building_rubble']):
			m25throwableobjects.append(obj)
	m25event.SetTimer(30,'FindThrowableObjects')
	#timee=ff.time()
	#print 'FindThrowableObjects time %f'%(timee-times)
	
def InWatchMode():
	return js.Mission_TestAttr('_in_watch_mode')
	#return m25getset.GetMissionVar('_in_watch_mode',default=0)
	
def SetWatchMode(mode=1):
	js.Mission_SetAttr('_in_watch_mode',mode)
	#m25getset.SetMissionVar('_in_watch_mode',mode)
	
def TransferAI(tochar='', fromchar=''):
	print 'TRANSFERAI %s to %s (watch=%f)'%(fromchar, tochar, AIInWatch(fromchar))
	SetupAI(tochar)
	SetTeam(tochar,GetTeam(fromchar))
	SetBrawler(tochar,GetBrawler(fromchar))
	if (not ffx.FFX_HasPortrait(tochar)) or AIInWatch(fromchar): 
		AIEnable(tochar)
		AIEnableWatch(tochar)
	else: 
		AIDisableCustom(tochar)
	#if InWatchMode(): AIEnableCustom(char2)
	AISetGoals(tochar, AIGetGoals(fromchar))
	if CharGetGroupName(fromchar) != None:
		AddToGroup(CharGetGroupName(fromchar),tochar)
		Group_PromoteLeader(CharGetGroupName(tochar))
	if ffx.FFX_HasPortrait(tochar) and m25config.M25_PLAYERAITOGGLE:
		if AIInWatch(tochar):
			js.Mission_CustomAction('stop ai', '', tochar, 'PlayerDisableAI', 1000, 1)
		else:
			js.Mission_CustomAction('start ai', '', tochar, 'PlayerEnableAI', 1000, 1)

def SetFlailing(char):
	print 'SetFlailing ' + char
	m25getset.SetObjectVar(char,'ai_flailing',1,nosave=1)		


stcodedict = {}
		
def generate_st_switch():
	global stcodedict
	stlist = []

	curfunction = ''
	curcode = []
		
	modpath = datfiles.GetModPath()
	fname = open(modpath+'\Missions\Scripts\m25ai.py','r')
	line = fname.readline()
	while line != '':
		tag = m25string.GetLeftString(line,' ')
		leftover = m25string.GetRightString(line,' ')
		if tag == 'def':
			
			#add subtype code to a dictionary for later processing and optimization
			if (curcode != []) and (curfunction != ''):
				stcodedict[curfunction] = curcode
				curcode = []
				curfunction = ''
		
			#print leftover
			#print leftover[0:3]
			if leftover[0:3] == "ST_":
				stlist.append(m25string.GetLeftString(leftover,'('))
				curfunction = m25string.GetLeftString(leftover,'(')
				curcode = []
		else: #not a function header
			l2 = string.replace(line, 'return ', 'codeval = ')
			line3 = string.lstrip(l2)
			if line3 != '' and (line[0] != '#'):
				curcode.append(l2[1:])
		line = fname.readline()
	fname.close()



	
	
	
def AICountFarPowers(char):
	count = 0
	cdata = chardata.GetCharacterData(char)
	for powdata in cdata['powers'].keys():
		power = cdata['powers'][powdata]
		if power['PowerType'] in [datfiles.PT_RANGED,datfiles.PT_DIRECT] and (not AIFriendPower(power)):
			count = count + 1
	return count
	
def AICountNearPowers(char):
	count = 0
	cdata = chardata.GetCharacterData(char)
	for powdata in cdata['powers'].keys():
		power = cdata['powers'][powdata]
		if power['PowerType'] in [datfiles.PT_MELEE,datfiles.PT_AREA] and (not AIFriendPower(power)):
			count = count + 1
	return count
	
def AIFriendPower(power):
	friendspecial = [datfiles.PT_SPECIAL_BUOYANCY,
		datfiles.PT_SPECIAL_CEREBRAL_NORMALISATION,
		datfiles.PT_SPECIAL_ENERGIZE,
		datfiles.PT_SPECIAL_ENERGIZE_AREA,
		datfiles.PT_SPECIAL_EP_TRANSFER,
		datfiles.PT_SPECIAL_LIFE_GIFT,
		datfiles.PT_SPECIAL_NORMALISE,
		datfiles.PT_SPECIAL_RALLY]
	if power['SpecialType'] in friendspecial:
		return 1
	return 0
		
def AIIsMelee(char):
	return not AIIsRanged(char)	
	
def AIIsRanged(char):
	if chardata.GetCharacterData(char)['strength'] >= 5:
		return 0
	return AICountFarPowers(char) >= AICountNearPowers(char)	
	
def ST_isranged(char,target=None,param=None,param2=None):
	return (not chardata.GetCharacterData(char)['strength'] >= 5) and (AICountFarPowers(char) >= AICountNearPowers(char))
	
def ST_ismelee(char, target=None, param=None, param2=None):
	return (chardata.GetCharacterData(char)['strength'] > 5) or (AICountFarPowers(char) < AICountNearPowers(char))

def AIIsFlying(char, default=0):
	#Ep. 2007-03-31: MLOG is now mandatory
	return mlogreader.MLOG_IsFlying(char)
	#if M25_USE_MLOG:
	#	return mlogreader.MLOG_IsFlying(char)
	#else:
	#	return m25getset.GetObjectVar(char,'ai_flying',default=default,integer=1)

def AIIsFlailing(char, default=0):
	#Ep. 2007-03-31: MLOG is now mandatory
	return mlogreader.MLOG_IsFlailing(char)
	#if M25_USE_MLOG:
	#	return mlogreader.MLOG_IsFlailing(char)
	#else:
	#	return m25getset.GetObjectVar(char,'ai_flailing',default=default,integer=1) 	

def AIIsLanding(char, default=0):
	#Ep. 2007-03-31: MLOG is now mandatory
	return mlogreader.MLOG_IsLanding(char)
	#if M25_USE_MLOG:
	#	return mlogreader.MLOG_IsLanding(char)
	#else:
	#	return default #don't know the state so return the default			
	
def AI_SubtypeFilter(char, subtype):
	return IsSubtype(char,subtype)

	
#####################################################################################
#
# Object Information
# 
# Description: sort active characters so targetting is made easier
#
#####################################################################################

#-------------------------------------------------------------------------------------
#
# 
#
#-------------------------------------------------------------------------------------

_ai_mapinfo = {}
_ai_grid = {}
_ai_grid_static = {}

def AI_GetMapInfo(char, key, default=None):
	global _ai_mapinfo
	try:
		return _ai_mapinfo[char][key]
	except:
		return default

def AI_MapInfoAvailable(char, key):
	global _ai_mapinfo
	return _ai_mapinfo.has_key(char) and _ai_mapinfo[char].has_key(key)
		
def AI_GetNearestEnemy(char, default=None):
	if not AI_MapInfoAvailable(char,'enemy'): AI_GetDynamicMapInfo(char)
	return AI_GetMapInfo(char, 'enemy', default)
	
def AI_GetNearestAlly(char, default=None):
	if not AI_MapInfoAvailable(char,'ally'): AI_GetDynamicMapInfo(char)
	return AI_GetMapInfo(char, 'ally', default)
		
def AI_GetNearestCivilian(char, default=None):
	if not AI_MapInfoAvailable(char,'civilian'): AI_GetDynamicMapInfo(char)
	return AI_GetMapInfo(char, 'civilian', default)	
	
def AI_GetNearEnemies(char, default=[]):
	if not AI_MapInfoAvailable(char,'near_enemies'): AI_GetDynamicMapInfo(char)
	return AI_GetMapInfo(char, 'near_enemies', default)	
	
def AI_GetNearAllies(char, default=[]):
	if not AI_MapInfoAvailable(char,'near_allies'): AI_GetDynamicMapInfo(char)
	return AI_GetMapInfo(char, 'near_allies', default)	
	
def AI_GetNearDynamic(char, default=[]):
	if not AI_MapInfoAvailable(char,'near_dynamic'): AI_GetDynamicMapInfo(char)
	return AI_GetMapInfo(char, 'near_dynamic', default)		
	
def AI_GetNearCivilians(char, default=[]):
	if not AI_MapInfoAvailable(char,'near_civilians'): AI_GetDynamicMapInfo(char)
	return AI_GetMapInfo(char, 'near_civilians', default)	
	
def AI_GetNearObjects(char, default=[]):
	if not AI_MapInfoAvailable(char,'near_static'): AI_GetStaticMapInfo(char)
	return AI_GetMapInfo(char, 'near_static', default)		

def AI_GetNearBuildings(char, default=[]):
	if not AI_MapInfoAvailable(char,'near_buildings'): AI_GetStaticMapInfo(char)
	return AI_GetMapInfo(char, 'near_buildings', default)
	
def AI_GetNearestThrowable(char, default=None):
	if not AI_MapInfoAvailable(char,'throwable'): AI_GetStaticMapInfo(char)
	return AI_GetMapInfo(char, 'throwable', default)
	
def AI_GetNearestObject(char, default=None):
	if not AI_MapInfoAvailable(char,'object'): AI_GetStaticMapInfo(char)
	return AI_GetMapInfo(char, 'object', default)
	
def AI_GetNearestBuilding(char, default=None):
	if not AI_MapInfoAvailable(char,'building'): AI_GetStaticMapInfo(char)
	return AI_GetMapInfo(char, 'building', default)	
	
def AI_AddToGrid(char, x, y):
	global _ai_grid
	x, y = int(x/100), int(y/100)
	if _ai_grid.has_key(x):
		if _ai_grid[x].has_key(y):
			_ai_grid[x][y].append(char)
		else:
			_ai_grid[x][y] = [char]
	else:
		_ai_grid[x] = {y:[char]}
	
def AI_AddToGridStatic(char, x, y):
	global _ai_grid_static
	x, y = int(x/100), int(y/100)
	if _ai_grid_static.has_key(x):
		if _ai_grid_static[x].has_key(y):
			_ai_grid_static[x][y].append(char)
		else:
			_ai_grid_static[x][y] = [char]
	else:
		_ai_grid_static[x] = {y:[char]}
			
def AI_GetGrid(x, y, distance=4, char=None):
	global _ai_grid
	x, y = int(x/100), int(y/100)
	objects = []
	for i in range(x-distance, x+distance):
		for j in range(y-distance, y+distance):
			if _ai_grid.has_key(i) and _ai_grid[i].has_key(j):
				for c in _ai_grid[i][j]:
					objects.append(c)
	if char and (char in objects): 
		objects.remove(char)
		if not objects: return []
	return objects
			
def AI_GetGridStatic(x, y, distance=4, char=None):
	global _ai_grid
	x, y = int(x/100), int(y/100)
	objects = []
	for i in range(x-distance, x+distance):
		for j in range(y-distance, y+distance):
			if _ai_grid_static.has_key(i) and _ai_grid_static[i].has_key(j):
				for c in _ai_grid_static[i][j]:
					objects.append(c)
	if char and (char in objects): 
		objects.remove(char)
		if not objects: return []
	return objects	
	
def AI_SetupMapInfo():
	global _ai_mapinfo
	_ai_mapinfo = {}
	m25event.SetTimer(1,'AI_UpdateMapInfoDynamic')
	m25event.SetTimer(1,'AI_UpdateMapInfoStatic')

def AI_UpdateFilter(char):
	return ST_char(char) and js.Object_IsAlive(char)
	
def AI_UpdateMapInfoDynamic(event=None):
	if m25missionover.GetMissionOver(): return
	global _ai_mapinfo
	global _ai_grid
	_ai_grid = {}
	start = ff.clock()
	chars = filter(AI_UpdateFilter, Mission_GetDynamicObjects())
	for c in chars:
		_ai_mapinfo[c] = {}
		_ai_mapinfo[c]['pos'] = js.Get_ObjectPos(c)
		AI_AddToGrid(c, _ai_mapinfo[c]['pos'][0],_ai_mapinfo[c]['pos'][1])
	m25report.ReportTime('AI_UpdateMapInfoDynamic', (ff.clock()-start))
	m25event.SetTimer(1,'AI_UpdateMapInfoDynamic')

def AI_GetDynamicMapInfo(char):
	global _ai_mapinfo
	c = char
	closest_enemy = None
	closest_enemy_dist = m25distance.DISTANCE_INFINITE
	closest_ally = None
	closest_ally_dist = m25distance.DISTANCE_INFINITE
	closest_civ = None
	closest_civ_dist = m25distance.DISTANCE_INFINITE
	near_enemies = []
	near_allies = []
	near_dynamic = []
	near_civilians = []
	if not _ai_mapinfo.has_key(c):
		_ai_mapinfo[c] = {}
		_ai_mapinfo[c]['pos'] = js.Get_ObjectPos(c)
	for t in AI_GetGrid(_ai_mapinfo[c]['pos'][0],_ai_mapinfo[c]['pos'][1],char=c):
		reaction = m25team.Team_GetReaction(c,t)
		new_dist = skXAngles.fSkXGetDistance (_ai_mapinfo[c]['pos'], _ai_mapinfo[t]['pos'], "xyz")
		close = new_dist < m25distance.DISTANCE_VISIBLE
		if close: near_dynamic.append(t)
		if reaction == m25team.TEAM_ENEMIES:
			if close: near_enemies.append(t)
			if (new_dist < closest_enemy_dist): 
				closest_enemy = t
				closest_enemy_dist = new_dist
		elif reaction == m25team.TEAM_ALLIES:
			if close: near_allies.append(t)
			if (new_dist < closest_ally_dist): 
				closest_ally = t
				closest_ally_dist = new_dist
		elif ST_civilian(t):
			if close: near_civilians.append(t)
			if (new_dist < closest_civ_dist):
				closest_civ = t
				closest_civ_dist = new_dist
		else:
			pass
	_ai_mapinfo[c]['enemy'] = [closest_enemy, closest_enemy_dist]
	_ai_mapinfo[c]['ally'] = [closest_ally, closest_ally_dist]
	_ai_mapinfo[c]['civilian'] = [closest_civ, closest_civ_dist]
	_ai_mapinfo[c]['near_enemies'] = near_enemies
	_ai_mapinfo[c]['near_allies'] = near_allies
	_ai_mapinfo[c]['near_dynamic'] = near_dynamic	
	_ai_mapinfo[c]['near_civilians'] = near_civilians
	return _ai_mapinfo[c]
	
	
def AI_UpdateMapInfoStatic(event=None):
	if m25missionover.GetMissionOver(): return
	global _ai_mapinfo
	global _ai_grid_static
	start = ff.clock()
	_ai_grid_static = {}
	chars = js.Mission_GetObjects()
	for c in chars:
		if not _ai_mapinfo.has_key(c):_ai_mapinfo[c] = {}
		_ai_mapinfo[c]['pos'] = js.Get_ObjectPos(c)
		if not ST_char(c):
			AI_AddToGridStatic(c, _ai_mapinfo[c]['pos'][0],_ai_mapinfo[c]['pos'][1])
	m25report.ReportTime('AI_UpdateMapInfoStatic', (ff.clock()-start))
	m25event.SetTimer(5,'AI_UpdateMapInfoStatic')
	
def AI_GetStaticMapInfo(char):
	c = char
	closest_object = None
	closest_object_dist = m25distance.DISTANCE_INFINITE
	closest_throwable = None
	closest_throwable_dist = m25distance.DISTANCE_INFINITE
	closest_building = None
	closest_building_dist = m25distance.DISTANCE_INFINITE
	near_static = []
	near_buildings = []
	if not _ai_mapinfo.has_key(c):
		_ai_mapinfo[c] = {}
		_ai_mapinfo[c]['pos'] = js.Get_ObjectPos(c)
	for t in AI_GetGridStatic(_ai_mapinfo[c]['pos'][0],_ai_mapinfo[c]['pos'][1]):
		if not js.Object_Exists(t): continue
		if js.Object_GetClass(t) & js.OC_TERRAIN: continue
		new_dist = skXAngles.fSkXGetDistance (_ai_mapinfo[c]['pos'], _ai_mapinfo[t]['pos'], "xyz")
		close = new_dist < m25distance.DISTANCE_VISIBLE
		if close: near_static.append(t)
		if (new_dist < closest_object_dist): 
			closest_object = t
			closest_object_dist = new_dist
		if js.Object_GetAttr(t,'throwable') and (new_dist < closest_throwable_dist): 
			closest_throwable = t
			closest_throwable_dist = new_dist
		if (js.Object_GetClass(t) & js.OC_BUILDING) and (new_dist < closest_building_dist):
			closest_building = t
			closest_building_dist = new_dist
			if close: near_buildings.append(t)
	_ai_mapinfo[c]['object'] = [closest_object, closest_object_dist]
	_ai_mapinfo[c]['throwable'] = [closest_throwable, closest_throwable_dist]
	_ai_mapinfo[c]['near_static'] = near_static
	_ai_mapinfo[c]['near_buildings'] = near_buildings
	_ai_mapinfo[c]['building'] = [closest_building, closest_building_dist]
	return _ai_mapinfo[c]
	
def AI_UpdateDynamicObject():
	global _ai_mapinfo
	for c in filter(AI_UpdateFilter,js.Mission_GetDynamicObjects()):
		AI_GetStaticMapInfo(c)
